#!/usr/bin/env Rscript
#
# ITI project
#
# Copyright Maxime Garcia, Ghislain Bidaut, Claire Rioualen, Quentin Da Costa
# (2009-2017)
#
# maxime.u.garcia@gmail.com
# ghislain.bidaut@inserm.fr
# claire.rioualen@inserm.fr
# quentin.da-costa@inserm.fr
#
# http://iti.sourceforge.net
#
# This software is a computer program whose purpose is to superimpose
# interactome over several gene expression datasets to find
# discriminative subnetworks predicting metastatic relapse in breast
# cancer.
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty and the software's author, the holder of the
# economic rights, and the successive licensors have only limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading, using, modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean that it is complicated to manipulate, and that also
# therefore means that it is reserved for developers and experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and, more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

# Author:
# Quentin DA COSTA  quentin.da-costa@inserm.fr

# Example of use:
# ~/workspace/iti/pipeline/R/computePValues.R real-subnet-scores-nwa.txt scores_dist1-normal.txt scores_dist2-normal.txt scores_dist3-normal.txt ~/workspace/output_data/

options(warn = -1)

## Packages ##

# "?packageName" in R console for more info (after loading the package with library() function)
# "?functionName" for more info about a function.
library(MASS)


## Functions ##

# Takes the real subnetwork score file, the 3 score files of the 3 random distributions
# and the output directory.
computePValues = function(realSnwScoreFile, randomSnwScoreFile, dir){
    if(is.na(dir)){
        scoresDir = ""
        listsDir  = ""
    }
    else{
        scoresDir = paste(dir, "scores/", sep = "")
        listsDir  = paste(dir, "lists/",  sep = "")
    }

    # Parsing the list of kept subnetworks.
    con          = file(paste(listsDir, "subnets-kept.txt", sep = ""), open = "r")
    subnetIDList = readLines(con)
    close(con)
    # Removing the comment lines.
    subnetIDList = grep("^[^#]", subnetIDList, value = TRUE)

    # Retrieve scores in a data.frame.
    realScoresDf = read.table(realSnwScoreFile, header = TRUE, sep = "\t", check.names = FALSE, stringsAsFactors = FALSE)

    # Retrieve labels of the analysed datasets.
    realHeader = colnames(realScoresDf)
    cols       = grep("score-", realHeader)
    labels     = realHeader[cols]

    randomScoresDf  = read.table(randomSnwScoreFile, header = TRUE, sep = "\t", check.names = FALSE, stringsAsFactors = FALSE)

    # Retrieve the header with distribution informations.
    distHead = colnames(randomScoresDf)[1]
    distHead = unlist(strsplit(distHead, "\\", fixed = TRUE))[2]
    # Retrieve the distribution name used to select subnetworks.
    dist     = unlist(strsplit(distHead, "-"))[2]

    randomHeader   = colnames(randomScoresDf)

    # Apply getPValues() to each subnetwork ID and retrieve the list of vectors (c(currentSubnetID, pValues)).
    pvalues = lapply(subnetIDList, getPValues, realScoresDf = realScoresDf, realHeader = realHeader, cols = cols, randomScoresDf = randomScoresDf, randomHeader = randomHeader, dist = dist)

    headers   = sub("score", paste("p_value-", distHead, sep = ""), labels, fixed = TRUE)
    colNames = c("Network", headers)

    # Writes p-value file.
    write.table(do.call("rbind", pvalues), paste(listsDir, "p-values_", distHead, "-nwa.txt", sep = ""), sep = "\t", quote = FALSE, row.names = FALSE, col.names = colNames)
}

# Takes a subnetwork ID, the data.frame with real subnetwork scores, the corresponding header and dataset labels.
# Returns the vector with the ID of the current subnetwork and its p-values.
getPValues = function(subnetID, realScoresDf, realHeader, cols, randomScoresDf, randomHeader, dist){
    lineNb = grep(paste(subnetID, "$", sep = ""), realScoresDf[, 1])
    row    = unlist(realScoresDf[lineNb,])
    ds     = realHeader[cols]
    scores = row[cols]

    # Apply calcPVal() to each dataset and corresponding current subnetwork score
    # and retrieve the vector containing the n (= number of distributions) p-values computed.
    subnetDatasetPVals = apply(cbind(ds, scores), 1, calcPVal, randomScoresDf = randomScoresDf, randomHeader = randomHeader, dist = dist)

    return(c(subnetID, subnetDatasetPVals))
}

# Calculates the p-value of the current subnetwork for the current dataset and the current distribution.
# Takes a vector with the dataset label and the corresponding subnetwork score, a random subnetwork score data.frame,
# the corresponding header and the current distribution.
# Returns the p-value.
calcPVal = function(dataset_score, randomScoresDf, randomHeader, dist){
    if(dataset_score[[2]] == "undef"){
        return("undef")
    }

    # Retrieve the column number of the current dataset.
    colNb = grep(paste(dataset_score[[1]], "$", sep = ""), randomHeader)
    # Retrieve the dataset scores in a vector without "undef" score.
    dataVect = as.numeric(subset(randomScoresDf, randomScoresDf[, colNb] != "undef", colNb)[[1]])

    if(dist == "gmd"){
        library(nor1mix)
        # Compute the fitting of the univariate gaussian mixture distribution from the Expectation-Maximization algorithm.

        norMixObj = norMixEM(dataVect, m=2, tol=1e-6)

        # Cumulative Distribution Function for P[X > x].
        pVal = pnorMix(as.numeric(dataset_score[[2]]), norMixObj, lower.tail = FALSE)
    }

    else if(dist == "normal"){
        # Maximum-likelihood fitting of dist Distribution.
        normMlf = fitdistr(dataVect, dist)
        # Retrieve needed parameters.
        mean = normMlf$estimate[["mean"]]
        sd   = normMlf$estimate[["sd"]]

        pVal = pnorm(as.numeric(dataset_score[[2]]), mean, sd, lower.tail = FALSE)
    }

    # Idem.
    else if(dist == "gamma"){
        gamMlf = fitdistr(dataVect, dist, lower = -Inf)
        shape  = gamMlf$estimate[["shape"]]
        rate   = gamMlf$estimate[["rate"]]

        pVal = pgamma(as.numeric(dataset_score[[2]]), shape, rate, lower.tail = FALSE)
    }

    return(pVal)
}


## Main program ##

args = commandArgs(TRUE)
computePValues(args[1], args[2], args[3])
if(!is.null(warnings())){ warnings()}
