#!/usr/bin/env Rscript
#
# ITI project
#
# Copyright Maxime Garcia, Ghislain Bidaut, Claire Rioualen, Quentin Da Costa
# (2009-2017)
#
# maxime.u.garcia@gmail.com
# ghislain.bidaut@inserm.fr
# claire.rioualen@inserm.fr
# quentin.da-costa@inserm.fr
#
# http://iti.sourceforge.net
#
# This software is a computer program whose purpose is to superimpose
# interactome over several gene expression datasets to find
# discriminative subnetworks predicting metastatic relapse in breast
# cancer.
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty and the software's author, the holder of the
# economic rights, and the successive licensors have only limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading, using, modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean that it is complicated to manipulate, and that also
# therefore means that it is reserved for developers and experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and, more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

# Author:
# Quentin DA COSTA  quentin.da-costa@inserm.fr

# Example of use:
# ~/workspace/iti/pipeline/R/computeThresholds.R ~/workspace/output_data/random1-subnets/random1-subnet-scores-nwa.txt normal 1 ~/workspace/output_data/

options( warn = -1 )

## Packages ##

# "?packageName" in R console for more info (after loading the package with library() function)
# "?functionName" for more info about a function.
library(MASS)


## Functions ##

# Writes the random subnetwork score file with distrivution informations in the header
# and the corresponding threshold file.
# Takes the random subnetwork score file, the distribution ("gamma", "normal" or "gmd"),
# the distribution number and the output directory.
computeThresholds = function(randomSubnetScoreFile, dist, distNb, dir){
    if(is.na(dir)){
        scoresDir     = ""
        thresholdsDir = ""
        picturesDir   = ""
    }
    else{
        scoresDir     = paste(dir, "scores/",     sep = "")
        thresholdsDir = paste(dir, "thresholds/", sep = "")
        picturesDir   = paste(dir, "pictures/",   sep = "")
    }

    # Retrieve scores in a data.frame.
    randomScoresDf = read.table(randomSubnetScoreFile, header = TRUE, sep = "\t", check.names = FALSE, stringsAsFactors = FALSE)
    # print(dim(randomScoresDf))

    # Retrieve the labels of the analysed datasets.
    header = colnames(randomScoresDf)
    longLabels = header[2:length(header)]

    if(length(longLabels) == 0){
        stop("The header of random subnetwork score files doesn't contain dataset names!", call. = FALSE)
    }

    colNames = c(paste("Network\\", "dist", distNb, "-", dist, sep = ""), header[2:length(header)])
    # print(colNames)

    if(dist == "gamma"){
        # Remove negative scores.
        for (colNb in 2:length(header)){
            randomScoresDf[randomScoresDf[colNb] <= 0, colNb] = "undef"
        }
    }

    # Write the data.frame in the specified file.
    write.table(randomScoresDf, paste(scoresDir, "scores_dist", distNb, "-", dist, "-nwa.txt", sep = ""), sep = "\t", quote = FALSE, row.names = FALSE, col.names = colNames)

    # All possible combinations in an array.
    pVals = expand.grid(c(5, 1), -1:-10)

    # Apply computeDatasetThresholds function to each dataset which name is label and retrieve an array list.
    thresholds = lapply(longLabels, computeDatasetThresholds, header = header, randomScoresDf = randomScoresDf, pVals = pVals, dist = dist, distNb = distNb, picturesDir = picturesDir)

    if(dist == "gmd"){
        colNames = c("# Dataset", "P-Value", "Threshold")
    }
    else if(dist == "normal"){
        colNames = c("# Dataset", "P-Value", "Threshold", "Mean", "Standard_Deviation")
    }
    else if(dist == "gamma"){
        colNames = c("# Dataset", "P-Value", "Threshold", "Shape", "Rate")
    }
    # Write thresholds in the specified file.
    write.table(do.call("rbind", thresholds), paste(thresholdsDir, "thresholds_dist", distNb, "-", dist, ".txt", sep = ""), sep = "\t", quote = FALSE, row.names = FALSE, col.names = colNames)
}

# Takes the dataset name, the vector with column names, the score data.frame, the p-values,
# the distribution, the distribution number and pictures/ directory.
# Returns the (transposed) array containing vectors containing: dataset_label, p-value, threshold.
computeDatasetThresholds = function(longLabel, header, randomScoresDf, pVals, dist, distNb, picturesDir){

    # Retrieve the column number of the dataset which name is label
    colNb = grep(paste(longLabel, "$", sep = ""), header)
    # Retrieve the dataset scores in a vector without "undef" score.
    dataVect = as.numeric(subset(randomScoresDf, randomScoresDf[, colNb] != "undef", colNb)[[1]])
    label = sub("score-", "", longLabel, fixed = TRUE)
    pictName = paste(picturesDir, "pict-", label, "RandDist", distNb, sep = "")

    if(dist == "gmd"){
        library(nor1mix)

       # Compute the fitting of the univariate gaussian mixture distribution from the Expectation-Maximization algorithm.
        norMixObj = norMixEM(dataVect, m=2, tol=1e-6)

        # Generating a pdf and a png imagefile with the histogram of the dataset network scores
        # and the fitted corresponding density curve.

	pdf(paste(pictName, "-fittedGM.pdf", sep = ""))
       	hist(dataVect, probability = TRUE, breaks = 100, col = "blue", main = paste(label, "random distribution", distNb, "\nand fitted Gaussian Mixture density curve"), xlab = "Subnetwork score")
       	curve(dnorMix(x, norMixObj), add = TRUE, col = "red", lwd = 3)
        legend("topright", legend = c("GM density function", "Null distribution"), lwd = 3, col = c("red", "blue"))
        trash = dev.off()
	
	if(capabilities('X11')) {
	        png(paste(pictName, "-fittedGM.png", sep = ""), width=1024, height=1024, pointsize=24)
        	hist(dataVect, probability = TRUE, breaks = 100, col = "blue", main = paste(label, "random distribution", distNb, "\nand fitted Gaussian Mixture density curve"), xlab = "Subnetwork score")
        	curve(dnorMix(x, norMixObj), add = TRUE, col = "red", lwd = 3)
        	legend("topright", legend = c("GM density function", "Null distribution"), lwd = 3, col = c("red", "blue"))
        	trash = dev.off()
	}
			
        # Apply computeThreshold function adapted to the current distribution to each p-value and return the (transposed) array

        # containing each vector (dataset_label, p-value, threshold) returned by this function.
        return(t(apply(pVals, 1, computeThresholdWithGMDist, norMixObj = norMixObj, label = label)))}
	

    else if(dist == "normal"){
        # Maximum-likelihood fitting of dist Distribution.
        normMlf = fitdistr(dataVect, dist)
        # Retrieve needed parameters.
        mean    = normMlf$estimate[["mean"]]
        sd      = normMlf$estimate[["sd"]]

        pdf(paste(pictName, "-fittedNormal.pdf", sep = ""))
       	hist(dataVect, probability = TRUE, breaks = 100, col = "blue", main = paste(label, "random distribution", distNb, "\nand fitted Normal density curve"), xlab = "Subnetwork score")
       	curve(dnorm(x, mean, sd), add = TRUE, col = "red", lwd = 3)
       	legend("topright", legend = c("Normal density function", "Null distribution"), lwd = 3, col = c("red", "blue"))
       	trash = dev.off()

	if(capabilities('X11')) {
		
		png(paste(pictName, "-fittedNormal.png", sep = ""), width=1024, height=1024, pointsize=24)
        	hist(dataVect, probability = TRUE, breaks = 100, col = "blue", main = paste(label, "random distribution", distNb, "\nand fitted Normal density curve"), xlab = "Subnetwork score")
       		curve(dnorm(x, mean, sd), add = TRUE, col = "red", lwd = 3)
       		legend("topright", legend = c("Normal density function", "Null distribution"), lwd = 3, col = c("red", "blue"))
       		trash = dev.off()
	}
	
        return(t(apply(pVals, 1, computeThresholdWithNormalDist, mean = mean, sd = sd, label = label)))
    }

    # Idem.
    else if(dist == "gamma"){
        gamMlf = fitdistr(dataVect, dist, lower = -Inf)
        shape  = gamMlf$estimate[["shape"]]
        rate   = gamMlf$estimate[["rate"]]

	pdf(paste(pictName, "-fittedGamma.pdf", sep = ""))
       	hist(dataVect, probability = TRUE, breaks = 100, col = "blue", main = paste(label, "random distribution", distNb, "\nand fitted Gamma density curve"), xlab = "Subnetwork score")
       	curve(dgamma(x, shape, rate), add = TRUE, col = "red", lwd = 3)
       	legend("topright", legend = c("Gamma density function", "Null distribution"), lwd = 3, col = c("red", "blue"))
      	trash = dev.off()

	if(capabilities('X11')) {        
		png(paste(pictName, "-fittedGamma.png", sep = ""), width=1024, height=1024, pointsize=24)
        	hist(dataVect, probability = TRUE, breaks = 100, col = "blue", main = paste(label, "random distribution", distNb, "\nand fitted Gamma density curve"), xlab = "Subnetwork score")
        	curve(dgamma(x, shape, rate), add = TRUE, col = "red", lwd = 3)
        	legend("topright", legend = c("Gamma density function", "Null distribution"), lwd = 3, col = c("red", "blue"))
        	trash = dev.off()
	}
	
	
        return(t(apply(pVals, 1, computeThresholdWithGammaDist, shape = shape, rate = rate, label = label)))
    }

    else{
        stop("Invalid distribution: \"normal\", \"gmd\" or \"gamma\" required", call. = FALSE)
    }
}

# Takes the p-value, the norMix object and the dataset name.
# Returns a vector in the following format: c(dataset_label, p-value, threshold).
computeThresholdWithGMDist = function(pVal, norMixObj, label){
    pVal = as.numeric(pVal[1])*10^as.numeric(pVal[2])
    # Quantile function (= inverse Cumulative Distribution Function) for P[X > x].
    threshold = qnorMix(pVal, norMixObj, lower.tail = FALSE)
    return(c(label, sprintf("%.0e", pVal), threshold))
}

# Takes the p-value, the mean, the standard deviation and the dataset name.
# Returns a vector in the following format: c(dataset_label, p-value, threshold, mean, sd).
computeThresholdWithNormalDist = function(pVal, mean, sd, label){
    pVal = as.numeric(pVal[1])*10^as.numeric(pVal[2])
    threshold = qnorm(pVal, mean, sd, lower.tail = FALSE)
    return(c(label, sprintf("%.0e", pVal), threshold, mean, sd))
}

# Takes the p-value, the shape, the rate and the dataset name.
# Returns a vector in the following format: c(dataset_label, p-value, threshold, shape, rate).
computeThresholdWithGammaDist = function(pVal, shape, rate, label){
    pVal = as.numeric(pVal[1])*10^as.numeric(pVal[2])
    threshold = qgamma(pVal, shape, rate, lower.tail = FALSE)
    return(c(label, sprintf("%.0e", pVal), threshold, shape, rate))
}


## Main program ##

args = commandArgs(TRUE)

# args[1] = random_subnetwork_score_file_name
# args[2] = distribution_name
# args[3] = distribution_number
# args[4] = output_directory_name
computeThresholds(args[1], args[2], args[3], args[4])

if(!is.null(warnings())){ warnings()}

# Debugging code
#file =
#dist = "gmd"
#nb =
#dirname = "/Users/bidaut/data/out-Desmedt-test/"
#computeThresholds(file, dist, nb, dirname)
