#!/usr/bin/env Rscript
#
# ITI project
#
# Copyright Maxime Garcia, Ghislain Bidaut, Claire Rioualen, Quentin Da Costa
# (2009-2017)
#
# maxime.u.garcia@gmail.com
# ghislain.bidaut@inserm.fr
# claire.rioualen@inserm.fr
# quentin.da-costa@inserm.fr
#
# http://iti.sourceforge.net
#
# This software is a computer program whose purpose is to superimpose
# interactome over several gene expression datasets to find
# discriminative subnetworks predicting metastatic relapse in breast
# cancer.
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty and the software's author, the holder of the
# economic rights, and the successive licensors have only limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading, using, modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean that it is complicated to manipulate, and that also
# therefore means that it is reserved for developers and experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and, more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

# Example of use:
# ~/workspace/iti/R/plotSubnetworkScoreHist.R ~/workspace/output_data/real-subnet-annot.txt ~/workspace/output_data/


## Functions ##

# Takes the network attribute file and the output directory.
plotDistribution = function(subnetScoreFile, dir) {
    if(is.na(dir)){
        picturesDir = ""
    }
    else{
        picturesDir = paste(dir, "pictures/", sep = "")
    }

    # Retrieve scores in a data.frame.
    scoresDf = read.table(subnetScoreFile, header = TRUE, sep = "\t", check.names = FALSE, stringsAsFactors = FALSE)

    # Retrieve the labels of the analysed datasets.
    header = colnames(scoresDf)
    longLabels = header[2:length(header)]

    if(length(longLabels) == 0){
        stop("The header of subnetwork score file doesn't contain dataset names!", call. = FALSE)
    }

    # Apply plotHist function to each dataset.
    trash = lapply(longLabels, plotHist, header = header, scoresDf = scoresDf, picturesDir = picturesDir)
}

# Takes the dataset name, the vector with column names and the score data.frame.
plotHist = function(longLabel, header, scoresDf, picturesDir) {
    # Retrieve the column number of the dataset which name is label
    colNb = grep(longLabel, header)
    # Retrieve the dataset scores in a vector ("undef" scores removed).
    dataVect = as.numeric(subset(scoresDf, scoresDf[, colNb] != "undef", colNb)[[1]])
    label = sub("score-", "", longLabel, fixed = TRUE)

    # Generating a pdf and a png imagefile with the histogram of the dataset network scores.
    pictName = paste(picturesDir, "pict-", label, "RealDist", sep = "")
    
    options(bitmapType='cairo')
    pdf(paste(pictName, ".pdf", sep = ""))
    hist(dataVect, probability = TRUE, breaks = 100, col = "red", main = paste(label, "detected\nsubnetwork scores distribution"), xlab = "Subnetwork score")
    legend("topright", legend = c("Scores distribution"), lwd = 3, col = c("red"))
    trash = dev.off()
    if(capabilities('X11')) {
    
    png(paste(pictName, ".png", sep = ""), width=1024, height=1024, pointsize=24)
    hist(dataVect, probability = TRUE, breaks = 100, col = "red", main = paste(label, "detected\nsubnetwork scores distribution"), xlab = "Subnetwork score")
    legend("topright", legend = c("Scores distribution"), lwd = 3, col = c("red"))
    trash = dev.off()
    }
}


## Main program ##

args = commandArgs(TRUE)

# args[1] = real_subnetwork_score_file_name
# args[2] = output_directory_name
plotDistribution(args[1], args[2])
