#!/usr/bin/perl
#
# HTS-Net project
#
# Copyright Ghislain Bidaut, Claire Rioualen, Quentin Da Costa
# (2009-2017)
#
# ghislain.bidaut@inserm.fr
# claire.rioualen@inserm.fr
# quentin.da-costa@inserm.fr
#
# http://htsnet.marseille.inserm.fr
#
# This software is a computer program whose purpose is to superimpose
# interactome over several gene expression datasets to find
# discriminative subnetworks predicting metastatic relapse in breast
# cancer.
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty and the software's author, the holder of the
# economic rights, and the successive licensors have only limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading, using, modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean that it is complicated to manipulate, and that also
# therefore means that it is reserved for developers and experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and, more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


## Modules ##

use strict;
use warnings;

use Getopt::Long;
use Data::Dumper;


## Arguments management ##

my ($help,              $d_output,          $d_htsnet,
    $f_regulomeNetwork, $f_interactomeNetwork, $f_totalEda,
    $f_seedList,        $f_geneInfo,
    $taxonID,           $pearsonCorrTh,
    $improvRate,        $maxOvl,            $useAbsVal,
    $useAnti,           $gamma, $normal,
    $gaussianMixture,   @a_distributions,   $f_gene2go,
    @a_f_interestGenes,
    @a_pValsInt,        @a_pValsReg,
    $noAbs,             $d_graphviz,
    @a_studies, $d_int, $d_reg,
    $project, $f_desc,  $d_metaSubnets,
    $connectivityTh,    $snwMerging, $f_exp);

my $snwDepth    = 5;
my $nrand       = 200;

GetOptions(
    'help|usage' => \$help,

    'outputDirs|od=s'      => \$d_output,
    'htsnetPath|htsnet=s'        => \$d_htsnet,
    
    'regulomeNetworkFile|rnwf=s'    => \$f_regulomeNetwork,
    'interactomeNetworkFile|inwf=s' => \$f_interactomeNetwork,

    'edgeAttFile|edaf=s'    => \$f_totalEda,
    'seedListFile|slf=s'    => \$f_seedList,
    'geneInfoFile|gif=s'    => \$f_geneInfo,
    
    'pcth|th=f'     => \$pearsonCorrTh,
    'rate|r=f'      => \$improvRate,
    'ov|mo=f'       => \$maxOvl,
    'nrand=i'       => \$nrand,
    'depth=i'       => \$snwDepth,
    'taxon=i'       => \$taxonID,

    'abs'   => \$useAbsVal,
    'anti'  => \$useAnti,
    'noabs' => \$noAbs,

    'gamma|g'              => \$gamma,
    'normal'               => \$normal,
    'gaussianMixture|gmd'  => \$gaussianMixture,
    'distributions|d=s{,}' => \@a_distributions,

    'gene2goFile|gof=s'          => \$f_gene2go,
    'interestGeneFiles|igf=s{,}' => \@a_f_interestGenes,
    'pValInt=s'                  => \@a_pValsInt,
    'pValReg=s'                  => \@a_pValsReg,
    'snwMerging|merge'           => \$snwMerging,

    'connectivityThreshold|conTh=i' => \$connectivityTh,

    'graphvizDir|gvd=s'        => \$d_graphviz,
    'intDir|id=s'              => \$d_int,
    'regDir|rd=s'              => \$d_reg,
    'metaSubnetsDir|md=s'      => \$d_metaSubnets,
    'projectName|project=s'    => \$project,
    'descriptionFiles|df=s'    => \$f_desc,

    'iFile|if=s'    => \$f_exp,
);

if (defined $gamma) {
    @a_distributions = ("gamma", "gamma", "gamma");
}

if (defined $gaussianMixture) {
    @a_distributions = ("gmd", "gmd", "gmd");
}

if (defined $normal || scalar(@a_distributions) != 3) {
    @a_distributions = ("normal", "normal", "normal");
}

if(!defined $f_totalEda || !defined $f_geneInfo ||
   !defined $taxonID    || !defined $d_htsnet ||
   !defined $improvRate    || !defined $pearsonCorrTh || !defined $f_gene2go ||
   !defined $f_regulomeNetwork   || !defined $f_interactomeNetwork  || !scalar @a_distributions  ||
   !scalar @a_pValsInt || !scalar @a_pValsReg || !defined $maxOvl  || !defined $d_metaSubnets  ||
   !defined $connectivityTh || !defined $d_int  || !defined $d_reg ||
   !defined $d_graphviz || !defined $project || !defined $f_exp ){
    die "\nMissing argument(s)\n";
}


## Main program ##

my $anti;
if(defined $useAnti){
    $anti = " -anti";
    $noAbs = 1;
}
else {
    $anti = "";
}
my $abs;
if(defined $noAbs){
    $abs = "";
}
else{
    $abs = " -abs";
}

$project =~ s/[^\w\-]+/_/g;
unless(defined $d_output){
    $d_output = $project;
}
$d_output =~ s/\/+$//;
$d_output =~ s/[^\w\-]+/_/g;

system("rm -rf $d_output");
mkdir $d_output;

my %h_d_int_reg;
$h_d_int_reg{'int'} = $d_int;
$h_d_int_reg{'reg'} = $d_reg;

for my $nwt(keys %h_d_int_reg){
    my $d_res = $h_d_int_reg{$nwt};
    mkdir "$d_output/$d_res";
    $d_res =~ s/\/+$//;

    my $f_network;
    my $pval;
    if($nwt eq 'int') {
        $f_network   = $f_interactomeNetwork;
	$pval = "@a_pValsInt";
    }
    elsif($nwt eq 'reg') {
        $f_network   = $f_regulomeNetwork;
	$pval = "@a_pValsReg";
    }

    my $command = "perl -I $d_htsnet/pipeline $d_htsnet/pipeline/detectSubnetworks.pl -od $d_output/$d_res -if $f_exp -nwf $f_network -edaf $f_totalEda -r $improvRate -th $pearsonCorrTh -depth $snwDepth -nrand $nrand$abs$anti -gif  $f_geneInfo -nwt $nwt";
    
    $command .= " -slf $f_seedList"  if(defined $f_seedList);
    system($command) == 0 || die "Pipeline ended at subnetwork identification step.\n";
    #print("$command\n");

    $command = "perl -I $d_htsnet/pipeline $d_htsnet/pipeline/subnetworksStatisticalValidation.pl -htsnet $d_htsnet -od $d_output/$d_res --distributions @a_distributions -s $d_output";
    system($command) == 0 || die "Pipeline ended at statistical validation step.\n";
    #print("$command\n");

    # Extract p-value
    if($pval eq "auto" || $pval eq "Auto") {
	$command = "perl -I /home/bidaut/bin/hts-net-1.0/pipeline /home/bidaut/bin/hts-net-1.0/pipeline/extractAutomatedPValues.pl -cf $d_output/$d_res/thresholds/*count* -o $d_output/$d_res/pValue.txt";
	system($command) == 0 || die "Pipeline ended at p-value calculation step.\n";
	#print("$command\n");
	$command = "perl -I $d_htsnet/pipeline $d_htsnet/pipeline/intersectSubnetworks-getGO-computePVal.pl --pValFile $d_output/$d_res/pValue.txt --consensus 1 -mo $maxOvl -htsnet $d_htsnet -od $d_output/$d_res -s $d_output -nwf $f_network -gof $f_gene2go -taxon $taxonID";
	$command .= " -merge" if(defined $snwMerging);
	$command .= " -igf @a_f_interestGenes" if(scalar @a_f_interestGenes);
	
	system($command) == 0 || die "Pipeline ended at intersectSubnetworks-getGO-computePVal step.\n";
	#print("$command\n");
	
    }
    else {
	
	$command = "perl -I $d_htsnet/pipeline $d_htsnet/pipeline/intersectSubnetworks-getGO-computePVal.pl -pVal $pval --consensus 1 -mo $maxOvl -htsnet $d_htsnet -od $d_output/$d_res -s $d_output -nwf $f_network -gof $f_gene2go -taxon $taxonID";
	$command .= " -merge" if(defined $snwMerging);
	$command .= " -igf @a_f_interestGenes" if(scalar @a_f_interestGenes);
	
	system($command) == 0 || die "Pipeline ended at intersectSubnetworks-getGO-computePVal step.\n";
	#print("$command\n");
    } 
}


my $command = "perl -I $d_htsnet/pipeline $d_htsnet/pipeline/integrateInteractomeRegulomeSubnetworks.pl --intDir $d_output/$d_int --regDir $d_output/$d_reg -od $d_output/$d_metaSubnets --connectivityThreshold $connectivityTh -eda $f_totalEda -gif $f_geneInfo -taxon $taxonID  -gof $f_gene2go -rnwf $f_regulomeNetwork -inwf $f_interactomeNetwork";

#print "$command\n";
system($command) == 0 || die "Pipeline ended at interactome regulome subnetwork integration step.\n";

$command = "perl -I $d_htsnet/pipeline $d_htsnet/pipeline/generateHTMLReport.pl -od Web_Report-$d_output -graph $d_graphviz -studies $d_output -id $d_int -rd $d_reg -md $d_metaSubnets -project \'$project\' -go -htsnet $d_htsnet -eda $f_totalEda -gif $f_geneInfo -if $f_exp";
$command .= " -df $f_desc" if(defined $f_desc);

#print "$command\n";
system($command) == 0 || die "Pipeline ended at generateHTMLReport step.\n";

system("zip -rq $d_output $d_output") == 0 || die "system zip failed\n";
system("zip -rq Web_Report-$d_output Web_Report-$d_output") == 0 || die "system zip failed\n";
system("rm -rf $d_output Web_Report-$d_output") == 0 || die "system rm failed\n";
