#!/usr/bin/perl
#
# ITI project
#
# Copyright Maxime Garcia, Ghislain Bidaut, Claire Rioualen, Quentin Da Costa
# (2009-2014)
#
# maxime.u.garcia@gmail.com
# ghislain.bidaut@inserm.fr
# claire.rioualen@inserm.fr
# quentin.da-costa@inserm.fr
#
# http://iti.sourceforge.net
#
# This software is a computer program whose purpose is to superimpose
# interactome over several gene expression datasets to find
# discriminative subnetworks predicting metastatic relapse in breast
# cancer.
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty and the software's author, the holder of the
# economic rights, and the successive licensors have only limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading, using, modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean that it is complicated to manipulate, and that also
# therefore means that it is reserved for developers and experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and, more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod

=head1 Name
consolidateNetwork.pl

=head1 Description
Parses and formats networks stored in submited network directory.

=head1 Authors
Ghislain Bidaut		ghislain.bidaut@inserm.fr
Quentin Da Costa	quentin.da-costa@inserm.fr
Maxime Garcia		maxime.u.garcia@gmail.com
Claire Rioualen		claire.rioualen@inserm.fr

=head1 Usage
perl consolidateNetwork.pl -od ~/workspace/network-data/ -nwd /home/data/iti/stable/formatted-input/networks/human/2013-05

=cut


## Modules ##

use strict;
use warnings;

use Getopt::Long;

use Network;
use utils;


## Arguments management ##

my ($d_output, $d_network);
# nwd example: /home/data/iti/stable/formatted-input/interactomes/mouse/11-2013
GetOptions('outputDir|od=s'   => \$d_output,
		   'networkDir|nwd=s' => \$d_network);

$d_network =~ s/\/$//;

## Main program ##

print "\n********* Read Network Directory\n";
# \%rh_network(geneID => \@neighbor_genes);
# \%rh_networkEdas("$gene1 ($interaction_type) $gene2" => \%($networkID => \@names_of_databases_containing_interaction))
my ($rh_network,     $rh_networkEdas,
	$rh_interactome, $rh_bidirectionalReg, $rh_regulome,
	$ra_nodes,       $ra_TFs)
    = Network::readNetworkDir($d_network);

print "********* Write network informations\n";
unless(-d $d_output){
	mkdir "$d_output";
}

# Writes edge attributes (names of databases containing each interaction) of total network
# in the following Cytoscape compatible format:
# col1: gene1 (interaction_type) gene2	col2: database_list	col3 (if pd interaction): interaction_direction_like_geneA->geneB.
open(OUT, ">$d_output/edge_attributes-eda.txt") || die "failed to open output file: $!";
print OUT "Interaction\tDirection\tSources\n";
for my $interaction(keys %$rh_networkEdas){
	my $h = $$rh_networkEdas{$interaction};
	my @a_formatted;
	for my $networkID(keys %$h){
		my $ra_sources = $$h{$networkID};
		if (scalar(@$ra_sources) != 0){
			push(@a_formatted, "$networkID\:\:" . join(';', @$ra_sources));
		}
		else{
			push(@a_formatted, $networkID);
		}
	}

	print OUT "$interaction";

	if(scalar @$ra_TFs and scalar(grep(/pd/, $interaction))){
		my @a_interaction = split(/\ \(pd\)\ /, $interaction);
		my $gi1 = $a_interaction[0];
		my $gi2 = $a_interaction[1];

		if(	utils::in_array($gi1, $ra_TFs) and utils::in_array($gi2, $$rh_regulome{$gi1}) &&
			utils::in_array($gi2, $ra_TFs) and utils::in_array($gi1, $$rh_regulome{$gi2})){
			print OUT "\t" . "<>";
		}elsif(utils::in_array($gi1, $ra_TFs) and utils::in_array($gi2, $$rh_regulome{$gi1})){
			print OUT "\t" . ">";
		}elsif(utils::in_array($gi2, $ra_TFs) and utils::in_array($gi1, $$rh_regulome{$gi2})){
			print OUT "\t" . "<";
		}
	}
	elsif(scalar(grep(/pp/, $interaction))){
		print OUT "\t" . "--";
	}

	print OUT "\t" . join('||', @a_formatted) . "\n";
}
close OUT;

(scalar(keys %$rh_network))     ? utils::writeHash($rh_network, "$d_output/network.txt"):
print "No network data found.\n";
(scalar(keys %$rh_interactome)) ? utils::writeHash($rh_interactome, "$d_output/interactome.txt"):
print "No interactome data found.\n";

utils::writeHash($rh_bidirectionalReg, "$d_output/regulome-bidirectional_interactions.txt") if(scalar(keys %$rh_bidirectionalReg));
(scalar(keys %$rh_regulome)) ? utils::writeHash($rh_regulome, "$d_output/regulome-oriented_interactions.txt"):
print "No regulome data found.\n";

(scalar @$ra_nodes) ? utils::writeArray($ra_nodes, "$d_output/nodes.txt"):
print "No network data found.\n";
(scalar @$ra_TFs)   ? utils::writeArray($ra_TFs, "$d_output/TFs.txt"):
print "No TFs found.\n";

print "\n$0 done.\n\n";
