#!/usr/bin/perl
#
# ITI project
#
# Copyright Maxime Garcia, Ghislain Bidaut, Claire Rioualen, Quentin Da Costa
# (2009-2017)
#
# maxime.u.garcia@gmail.com
# ghislain.bidaut@inserm.fr
# claire.rioualen@inserm.fr
# quentin.da-costa@inserm.fr
#
# http://iti.sourceforge.net
#
# This software is a computer program whose purpose is to superimpose
# interactome over several gene expression datasets to find
# discriminative subnetworks predicting metastatic relapse in breast
# cancer.
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty and the software's author, the holder of the
# economic rights, and the successive licensors have only limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading, using, modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean that it is complicated to manipulate, and that also
# therefore means that it is reserved for developers and experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and, more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package HTS;

use strict;
use warnings;
use Utils;
use Data::Dumper;

=pod

=head2 readInputDataDirLabels()

Reads dataset names from input directory.

Returns:
\%($dataset_label => \%())

=cut

sub readInputDataDirLabels {
    my $inputDataDirectory = shift;

    my @a_directoryContent = sort(Utils::readInputDir($inputDataDirectory));
    my %h_datasets;

    for my $file (@a_directoryContent) {

        if ($file =~ m/\.txt\.gz$/) {

            my %h_dataset;
            my $studyName = $file;
            $studyName =~ s/\.txt\.gz$//;

            # In order to keep the architecture of the hash returned by readInputDataDir.
            $h_datasets{$studyName} = \%h_dataset;
        }
    }
    print "\nNumber of datasets read = " . (scalar keys %h_datasets) . "\n";

    if (scalar(keys %h_datasets) == 0) {
        die "No data found in $inputDataDirectory/\n";
    }

    return \%h_datasets;
}

=pod

=head2 readInputDataDir()

Reads data from input directory.

If f_exp is provided, data read from the file. Otherwise, read from directory.

Returns datasets in hash:
\%($dataset_label => \%('cond'          => \%($sample_ID => $condition),
                        'rancond'       => \%($sample_ID => $condition),
                        'collapseddata' => \%($gene_ID => \%($sample_ID => $expression)),
                        'expe'          => \@sample_list))

=cut

sub readInputDataDir {
    my $d_inputData   = shift;
    my $f_exp             = shift;    # For Mobyle version usage
    
    my @a_inputDirectoryContent;
    if($d_inputData ne ""){
        if($d_inputData !~ m/\/$/)   {$d_inputData   .= "/";}   
        @a_inputDirectoryContent = sort(Utils::readInputDir($d_inputData));
    }
    else{
        @a_inputDirectoryContent = ($f_exp);
    }
    
    my %h_datasets;

    for my $file (@a_inputDirectoryContent) {  # Read all txt files in dir
        next if ($file !~ m/\.txt$/);
        my %h_dataSet;

        my $studyName = Utils::getName($file);
        print "\nStudy name                         = $studyName\n";

        my($rh_data, $ra_expeList);

	($rh_data, $ra_expeList) = Utils::readHTSData("$d_inputData$file");

	# Case of a unique condition: a score is directly provided

	#print "Unique Condition Found - Score will be directly used\n";
	my %h_conditions;
	$h_conditions{$$ra_expeList[0]} = 'NoConds';
	$h_dataSet{'cond'}      = \%h_conditions;
	$h_dataSet{'expe'}      = $ra_expeList;
	$h_dataSet{'rancond'}   = \%h_conditions;
	$h_dataSet{'noHeatmap'} = 1;
        
	# No probe conversion needed - No probeID to geneID conversion to perform\
	
        $h_dataSet{'collapseddata'} = $rh_data;
        $h_datasets{$studyName} = \%h_dataSet;
    }

    if (scalar keys %h_datasets == 0) {
        die "No data found in $d_inputData/\n";
    }

    #print "Number of datasets read = " . scalar(keys(%h_datasets)) . "\n\n";

    return \%h_datasets;
}



=pod

=head2 projectsInputData()

Performs projection of input data on clinical condition vector.

Returns node attribute hash containing gene scores:
\%($gene_ID => \%('Correlations' => \%($dataset_label => $gene_score)))

=cut

sub projectsInputData {
    my $rh_datasets = shift;
    my $useAbsVal   = shift;
    my $useAnti     = shift;

    # \%rh_datasets($dataset_label => \%('cond'          => \%($sample_ID => $condition),
    #                                    'rancond'       => \%($sample_ID => $condition),
    #                                    'collapseddata' => \%($gene_ID => \%($sample_ID => $expression)),
    #                                    'expe'          => \@sample_list))

    my @a_sampleNbs;
    foreach my $label (keys %$rh_datasets) {
        my $ra_samples = $$rh_datasets{$label}{'expe'};
        push(@a_sampleNbs, scalar @$ra_samples);
    }

    my $maxSampleNb = Utils::max(@a_sampleNbs);

    #print "Maxmimum number of samples = $maxSampleNb\n";

    my %h_nodeAtts;
    foreach my $label (keys %$rh_datasets) {
        my $rh_data             = $$rh_datasets{$label}{'collapseddata'};
        my $ra_samples          = $$rh_datasets{$label}{'expe'};
        my $rh_randomConditions = $$rh_datasets{$label}{'rancond'};
        my $sampleNb            = scalar @$ra_samples;
        my @a_geneList          = keys %$rh_data;

        #print "Current Dataset = $label ($sampleNb samples)\n";

        foreach my $geneID (keys %$rh_data) {
            # print "gi = $geneID\n";

            my($correlation, $randCorrelation);
            my(@a_dataVect);

            foreach my $sample (@$ra_samples) {
                push(@a_dataVect, $$rh_data{$geneID}{$sample});
            }

            # Monoconditional Case
            if(scalar keys %{$$rh_datasets{$label}{'cond'}}) {
                my $rh_conditions = $$rh_datasets{$label}{'cond'};

                # Previously computed score
                if(scalar(keys %$rh_conditions) == 1) {    # NoConds
                    $correlation  = $$rh_data{$geneID}{$$ra_samples[0]};
                }
                elsif(defined $$rh_datasets{$label}{'MultiValues'}) {
                    my @a_condVect = ();
                    for my $sample (@$ra_samples) {
                        push(@a_condVect, $$rh_conditions{$sample});
                    }

                    $correlation  = Utils::multiconditionalDifferentialExpression(\@a_dataVect, \@a_condVect);
                    $correlation *= sqrt($sampleNb) / sqrt($maxSampleNb);

                }
                # Compute normal correlation
                else {

                    my @a_condVect = ();
                    foreach my $sample (@$ra_samples) {
                        push(@a_condVect, $$rh_conditions{$sample});
                    }
                    $correlation  = Utils::pearsonCorrelation(\@a_dataVect, \@a_condVect);
                    $correlation *= sqrt($sampleNb) / sqrt($maxSampleNb);

                }
            }
            # Multiconditional Case
            else {

                my $rh_conditions = $$rh_datasets{$label}{'mcond'};
                my $rh_conditionsWeight = $$rh_datasets{$label}{'mcondweight'};
                $correlation = 0;

                foreach my $condType(keys %$rh_conditionsWeight) {
                    my @a_condVect = ();
                    for my $sample (@$ra_samples) {
                        push(@a_condVect, $$rh_conditions{$sample}{$condType});
                    }

                    $correlation += $$rh_conditionsWeight{$condType} * Utils::pearsonCorrelation(\@a_dataVect, \@a_condVect);

                }
            }

            $h_nodeAtts{$geneID}{'Correlations'}{$label} = $correlation;

            if(defined $useAbsVal) {
                $correlation = abs($correlation);
            }
            elsif(defined $useAnti)  {
                $correlation = -$correlation;
            }

            # Storing the gene score in the dataset hash.
            # \%rh_datasets($dataset_label => \%('Gene_scores' => \%($gene_ID => \%("correlation-$dataset_label" => $gene_score))))
            $$rh_datasets{$label}{'Gene_scores'}{$geneID}{"correlation-$label"} = $correlation;


            # 2) Compute randomized correlation

            # Monoconditional Case
            if(scalar keys %{$$rh_datasets{$label}{'cond'}}) {
                my $rh_conditions = $$rh_datasets{$label}{'cond'};

                if(scalar(keys %$rh_conditions) == 1) {    # NoConds
                    my $randGene     = $a_geneList[int(rand(scalar keys %$rh_data))];
                    $randCorrelation = $$rh_data{$randGene}{$$ra_samples[0]};
                }
                elsif(defined $$rh_datasets{$label}{'MultiValues'}) {
                    my @a_randConds;
                    for my $sample (@$ra_samples) {
                        push(@a_randConds, $$rh_randomConditions{$sample});
                    }

                    $randCorrelation = Utils::multiconditionalDifferentialExpression(\@a_dataVect, \@a_randConds);
                    $randCorrelation *= sqrt($sampleNb) / sqrt($maxSampleNb);
                }
                # Compute normal correlation
                else {
                    my @a_randConds;
                    foreach my $sample (@$ra_samples) {
                        push(@a_randConds, $$rh_randomConditions{$sample});
                    }
                    $randCorrelation  = Utils::pearsonCorrelation(\@a_dataVect, \@a_randConds);
                    $randCorrelation *= sqrt($sampleNb) / sqrt($maxSampleNb);
                }
            }
            # Multiconditional Case
            else {
                my $rh_conditions = $$rh_datasets{$label}{'mcondrandom'};
                my $rh_conditionsWeight = $$rh_datasets{$label}{'mcondweightrandom'};
                $randCorrelation = 0;

                foreach my $condType(keys %$rh_conditionsWeight) {
                    my @a_condVect = ();
                    for my $sample (@$ra_samples) {push(@a_condVect, $$rh_conditions{$sample}{$condType});}

                    $randCorrelation += $$rh_conditionsWeight{$condType} * Utils::pearsonCorrelation(\@a_dataVect, \@a_condVect);
                }

            }

            if(defined $useAbsVal) {
                $randCorrelation = abs($randCorrelation);
            }
            elsif(defined $useAnti)  {
                $randCorrelation = -$randCorrelation;
            }

            $$rh_datasets{$label}{'Gene_scores'}{$geneID}{"randProjectedInputData-$label"} = $randCorrelation;
        }
    }

    return \%h_nodeAtts, $maxSampleNb;
}




=pod

=head2 annotationOverlap()

Computes the overlap between inputData data gene IDs and gene_ID-probe_IDs matching hash.

=cut

sub annotationOverlap {
    my ($rh_gene2probe, $rh_collapsedData) = (shift, shift);
    
    my $overlap = 0;
    foreach my $geneID (keys %$rh_collapsedData) {
        if (defined $$rh_gene2probe{$geneID}) {$overlap++;}
    }

    return $overlap;
}


=pod

=head2 randomizeConditions()

Shuffles conditions.

Returns randomized conditions hash:
\%($sample_ID => $condition)

=cut

sub randomizeConditions {
    my $ra_expeList = shift;
    my $rh_conditions = shift;
    # \%rh_conditions($sample_ID => $condition)

    my @randomExpe;
    my %h_randomConds;

    foreach my $exp (@$ra_expeList) {
        push(@randomExpe, $exp);
        $h_randomConds{$exp} = $$rh_conditions{$exp};
    }

    my $n = 100 * (scalar @$ra_expeList);

    while ($n > 1) {
        my $k1 = rand() * (scalar @$ra_expeList);
        my $k2 = rand() * (scalar @$ra_expeList);
        $n--;

        my $tmp = $randomExpe[$k1];
        $randomExpe[$k1] = $randomExpe[$k2];
        $randomExpe[$k2] = $tmp;

        $tmp                             = $h_randomConds{$randomExpe[$k1]};
        $h_randomConds{$randomExpe[$k1]} = $h_randomConds{$randomExpe[$k2]};
        $h_randomConds{$randomExpe[$k2]} = $tmp;
    }

    return \%h_randomConds;
}


=pod

=head2 calculateAverageExpressionOverCondition()

Calculates average expression of each gene.

Returns:
\%($gene_ID => $averageExpression)

=cut

sub calculateAverageExpressionOverCondition {
    my $rh_datasets = shift;
    my $dataset     = shift;
    # \%rh_datasets($dataset_label => \%('collapseddata' => \%($gene_ID => \%($sample_ID => $expression)),
    #                                    'expe'          => \@sample_list))

    my $rh_averageExpr;

    my $rh_data       = $$rh_datasets{$dataset}{'collapseddata'};
    my $ra_experiment = $$rh_datasets{$dataset}{'expe'};

    foreach my $geneID (keys %$rh_data) {

        my @av;
        for (my $i = 0; $i < scalar @$ra_experiment; $i++) {
            if (defined $$rh_data{$geneID}{$$ra_experiment[$i]}) {
                push(@av, $$rh_data{$geneID}{$$ra_experiment[$i]});
            }
        }

        my $average = 0;
        foreach (@av) {$average += $_;}

        if (scalar @av > 0) {
            $average /= scalar @av;
            $$rh_averageExpr{$geneID} = $average;
        }
    }

    return $rh_averageExpr;
}


1;
