#!/usr/bin/perl
#
# ITI project
#
# Copyright Maxime Garcia, Ghislain Bidaut, Claire Rioualen, Quentin Da Costa
# (2009-2017)
#
# maxime.u.garcia@gmail.com
# ghislain.bidaut@inserm.fr
# claire.rioualen@inserm.fr
# quentin.da-costa@inserm.fr
#
# http://iti.sourceforge.net
#
# This software is a computer program whose purpose is to superimpose
# interactome over several gene expression datasets to find
# discriminative subnetworks predicting metastatic relapse in breast
# cancer.
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty and the software's author, the holder of the
# economic rights, and the successive licensors have only limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading, using, modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean that it is complicated to manipulate, and that also
# therefore means that it is reserved for developers and experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and, more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package Metasubnet;

use strict;
use warnings;
use Utils;

use Data::Dumper;

# - Create an objet "metasubnet": $subnetID => {
#    'type' => $network                # network type of the seed subnet
#    'Interactors' => @subnetIDs       # interacting subnets
#    'Connectors' => @geneIDs          # list of genes common between seed & interactors
#    'Int_nodes' => @geneIDs           # list of not common genes from int subnets
#    'Reg_nodes' => @geneIDs}          # list of not common genes from reg subnets

sub metasubnets {
    my($rh_int_eda, $rh_int_nwa, $rh_int_noa, $rh_reg_eda, $rh_reg_nwa, $rh_reg_noa,
       $th, $rh_tot_eda, $totalGeneNb, $rh_GOccur, $rh_gene2GOs) = @_;

    my %h_meta_nwa = ();
    my %h_meta_noa = ();
    my %h_meta_eda = ();
    my %h_subnets = ();

    foreach my $int_subnetID (keys %{$rh_int_nwa}) {

        foreach my $reg_subnetID (keys %{$rh_reg_nwa}) {

            my @ints = ();
            my @regs = ();

            my @connections = @{overlapCount($int_subnetID, $reg_subnetID, $rh_int_nwa, $rh_reg_nwa)};


            if(scalar(@connections)>=$th){

                # Copy Interactome Info

                # INT subnets
                $h_meta_nwa{$int_subnetID}{'Type'} = "int";

                # INT nodes
                my @int_nodes = keys(%{$$rh_int_nwa{$int_subnetID}{'Nodes'}});

                foreach my $node (@int_nodes){

                    if (!Utils::in_array($node, \@ints) && !Utils::in_array($node, \@connections)){push(@ints, $node);}

                    $h_meta_nwa{$reg_subnetID}{'Nodes'}{$node} = 1;
                    $h_meta_nwa{$int_subnetID}{'Nodes'}{$node} = 1;


                    $h_meta_noa{$node}{'Type'}{'int'} = 1;
                    $h_meta_noa{$node}{'Gene_symbol'} = $$rh_int_noa{$node}{'Gene_symbol'};
                    $h_meta_noa{$node}{'Correlations'} = $$rh_int_noa{$node}{'Correlations'};
                    $h_meta_noa{$node}{'Highlights'} = $$rh_int_noa{$node}{'Highlights'};
                    $h_meta_noa{$node}{'Best_subnetwork_average_score'} = $$rh_int_noa{$node}{'Best_subnetwork_average_score'};

                    if(! Utils::in_array("meta-$reg_subnetID", \@{$h_meta_noa{$node}{'Subnets'}})){
                        push(@{$h_meta_noa{$node}{'Subnets'}}, "meta-$reg_subnetID");
                    }
                    if(! Utils::in_array("meta-$int_subnetID", \@{$h_meta_noa{$node}{'Subnets'}})){
                        push(@{$h_meta_noa{$node}{'Subnets'}}, "meta-$int_subnetID");
                    }
                }

                # Copy Regulome Info

                # REG subnets
                $h_meta_nwa{$reg_subnetID}{'Type'} = "reg";

                # REG nodes
                my @reg_nodes = keys(%{$$rh_reg_nwa{$reg_subnetID}{'Nodes'}});

                foreach my $node (@reg_nodes){

                    if (!Utils::in_array($node, \@regs) && !Utils::in_array($node, \@connections)){push(@regs, $node);}

                    $h_meta_nwa{$reg_subnetID}{'Nodes'}{$node} = 1;
                    $h_meta_nwa{$int_subnetID}{'Nodes'}{$node} = 1;

                    $h_meta_noa{$node}{'Type'}{'reg'} = 1;
                    $h_meta_noa{$node}{'Gene_symbol'} = $$rh_reg_noa{$node}{'Gene_symbol'};
                    $h_meta_noa{$node}{'Correlations'} = $$rh_reg_noa{$node}{'Correlations'};

                    $h_meta_noa{$node}{'Highlights'} = $$rh_reg_noa{$node}{'Highlights'};
                    $h_meta_noa{$node}{'Best_subnetwork_average_score'} = $$rh_reg_noa{$node}{'Best_subnetwork_average_score'};

                    if(! Utils::in_array("meta-$reg_subnetID", \@{$h_meta_noa{$node}{'Subnets'}})){
                        push(@{$h_meta_noa{$node}{'Subnets'}}, "meta-$reg_subnetID");
                    }
                    if(! Utils::in_array("meta-$int_subnetID", \@{$h_meta_noa{$node}{'Subnets'}})){
                        push(@{$h_meta_noa{$node}{'Subnets'}}, "meta-$int_subnetID");
                    }
                }

                # Add Metasubnets info

                # interactors
                if(! Utils::in_array($reg_subnetID, $h_meta_nwa{$int_subnetID}{'Interactors'})) { push(@{$h_meta_nwa{$int_subnetID}{'Interactors'}}, $reg_subnetID); }
                if(! Utils::in_array($int_subnetID, $h_meta_nwa{$reg_subnetID}{'Interactors'})){push(@{$h_meta_nwa{$reg_subnetID}{'Interactors'}}, $int_subnetID); }
                if(! Utils::in_array($reg_subnetID, $h_meta_nwa{$reg_subnetID}{'Interactors'})){push(@{$h_meta_nwa{$reg_subnetID}{'Interactors'}}, $reg_subnetID); }
                if(! Utils::in_array($int_subnetID, $h_meta_nwa{$int_subnetID}{'Interactors'})){push(@{$h_meta_nwa{$int_subnetID}{'Interactors'}}, $int_subnetID); }


                # connectors
                foreach my $c (@connections ){
                    if (! Utils::in_array($c, $h_meta_nwa{$int_subnetID}{'Connectors'})){push(@{$h_meta_nwa{$int_subnetID}{'Connectors'}}, $c); }
                    if (! Utils::in_array($c, $h_meta_nwa{$reg_subnetID}{'Connectors'})){push(@{$h_meta_nwa{$reg_subnetID}{'Connectors'}}, $c); }
                    $h_meta_noa{$c}{'Connector'} = 1;
                }


                # Store subnetwork attributes in a single hash (not metasubnets)
                @{$h_subnets{$int_subnetID}}{keys %{$$rh_int_nwa{$int_subnetID}}} = values %{$$rh_int_nwa{$int_subnetID}};
                @{$h_subnets{$reg_subnetID}}{keys %{$$rh_reg_nwa{$reg_subnetID}}} = values %{$$rh_reg_nwa{$reg_subnetID}};

                foreach my $n (@ints){
                    if (! Utils::in_array($n, $h_meta_nwa{$int_subnetID}{'Int_nodes'})){push(@{$h_meta_nwa{$int_subnetID}{'Int_nodes'}}, $n);}
                    if (! Utils::in_array($n, $h_meta_nwa{$reg_subnetID}{'Int_nodes'})){push(@{$h_meta_nwa{$reg_subnetID}{'Int_nodes'}}, $n);}
                }
                foreach my $n (@regs){
                    if (! Utils::in_array($n, $h_meta_nwa{$int_subnetID}{'Reg_nodes'})){push(@{$h_meta_nwa{$int_subnetID}{'Reg_nodes'}}, $n);}
                    if (! Utils::in_array($n, $h_meta_nwa{$reg_subnetID}{'Reg_nodes'})){push(@{$h_meta_nwa{$reg_subnetID}{'Reg_nodes'}}, $n);}
                }
            }
        }
    }

    for my $int(keys %$rh_tot_eda){
        my @a_int = split(/-/, $int);

        for my $snwID(keys %h_meta_nwa){
            next unless(defined $h_meta_nwa{$snwID}{'Nodes'}{$a_int[0]} and defined $h_meta_nwa{$snwID}{'Nodes'}{$a_int[1]});

            push(@{$h_meta_nwa{$snwID}{'Edges'}}, $int);
            my @a_type = keys %{$$rh_tot_eda{$int}};
            for my $type(@a_type){
                $h_meta_eda{$int}{$type} = $$rh_tot_eda{$int}{$type};

                if($type eq 'pd'){
                    $h_meta_noa{$a_int[0]}{'TF'} = 1 if(scalar(grep(/>/, $$rh_tot_eda{$int}{$type}{'Direction'})));
                    $h_meta_noa{$a_int[1]}{'TF'} = 1 if(scalar(grep(/</, $$rh_tot_eda{$int}{$type}{'Direction'})));
                }
            }
        }
    }

    foreach my $n (keys %h_meta_noa){
        $h_meta_noa{$n}{'Gene_frequency'} = scalar @{$h_meta_noa{$n}{'Subnets'}} ;

        my $int_degree = my $reg_degree = 0;
        if (defined $$rh_int_noa{$n}{'Gene_degree'}) {$int_degree = $$rh_int_noa{$n}{'Gene_degree'};}
        if (defined $$rh_reg_noa{$n}{'Gene_degree'}) {$reg_degree = $$rh_reg_noa{$n}{'Gene_degree'};}

        my $degree = $int_degree + $reg_degree;
        $h_meta_noa{$n}{'Gene_degree'} = $degree;
    }

    foreach my $k (keys %h_meta_nwa){
        $h_meta_nwa{"meta-$k"} = delete $h_meta_nwa{$k};
    }

    # Combine GO counts
    my %h_metaGONwas;
    my %h_metaGOnoas;
    for my $meta_subnetID(keys %h_meta_nwa){

        my @a_snwGenes = keys %{$h_meta_nwa{$meta_subnetID}{'Nodes'}};
        for my $geneID(@a_snwGenes){

            for my $GOID(keys %{$$rh_gene2GOs{$geneID}}){

                $h_metaGONwas{$meta_subnetID}{$GOID}{'Subnet_GO_occurrence'}++;
                $h_metaGONwas{$meta_subnetID}{$GOID}{'Subnet_gene_number'} = scalar(@a_snwGenes);
                $h_metaGONwas{$meta_subnetID}{$GOID}{'Total_GO_occurrence'} = $$rh_GOccur{$GOID};
                $h_metaGONwas{$meta_subnetID}{$GOID}{'Annotated_gene_number'} = $totalGeneNb;

                $h_metaGOnoas{$geneID} = $$rh_gene2GOs{$geneID};

            }
        }
    }

    # Updating GO statistics
    for my $meta_subnetID(keys %h_meta_nwa){

        for my $GOID(keys %{$h_metaGONwas{$meta_subnetID}}){
            #print "sngo=$h_metaGONwas{$meta_subnetID}{$GOID}{'Subnet_GO_occurrence'}  totgo=$h_metaGONwas{$meta_subnetID}{$GOID}{'Total_GO_occurrence'}  sngn=$h_metaGONwas{$meta_subnetID}{$GOID}{'Subnet_gene_number'} annogn= $h_metaGONwas{$meta_subnetID}{$GOID}{'Annotated_gene_number'}\n";

            my $hyperPVal = Utils::hypergeometricTest($h_metaGONwas{$meta_subnetID}{$GOID}{'Subnet_GO_occurrence'},
                                                      $h_metaGONwas{$meta_subnetID}{$GOID}{'Total_GO_occurrence'},
                                                      $h_metaGONwas{$meta_subnetID}{$GOID}{'Subnet_gene_number'},
                                                      $h_metaGONwas{$meta_subnetID}{$GOID}{'Annotated_gene_number'});

            my $corrPVal = Utils::min(1.0, $hyperPVal *  $h_metaGONwas{$meta_subnetID}{$GOID}{'Annotated_gene_number'});
            my $enrichRatio = log(($h_metaGONwas{$meta_subnetID}{$GOID}{'Subnet_GO_occurrence'} /
                                   $h_metaGONwas{$meta_subnetID}{$GOID}{'Subnet_gene_number'}) /
                                  ($h_metaGONwas{$meta_subnetID}{$GOID}{'Total_GO_occurrence'}  /
                                   $h_metaGONwas{$meta_subnetID}{$GOID}{'Annotated_gene_number'})) / log(2);
            $h_metaGONwas{$meta_subnetID}{$GOID}{'Hypergeometric_test'} = sprintf("%.2e", $hyperPVal);
            $h_metaGONwas{$meta_subnetID}{$GOID}{'Corrected_p-value'}   = sprintf("%.2e", $corrPVal);
            $h_metaGONwas{$meta_subnetID}{$GOID}{'Enrichment_ratio'}    = sprintf("%.3f", $enrichRatio);
        }
    }

    return (\%h_meta_nwa, \%h_meta_noa, \%h_meta_eda, \%h_subnets, \%h_metaGONwas, \%h_metaGOnoas);
}

# writes metasubnets in files
sub writeMetasubnets {
    my $outputDir = shift;
    my $rh_meta_nwa = shift;
    my $rh_meta_noa = shift;
    my $rh_meta_eda = shift;
    my $rh_subnets = shift;

    # writing metasubnets file
    open(META, ">$outputDir/metasubnets.txt") || die "failed to open output file: $!";
    print META "# ID\tSize\tInteractors\tConnectors\tType\tInt_nodes\tReg_nodes\n";

    my @a_networkList = keys %$rh_meta_nwa;
        
    foreach my $id(@a_networkList) {

        my $size = scalar(@{$$rh_meta_nwa{$id}{'Interactors'}});
        my $inter = join(";", @{$$rh_meta_nwa{$id}{'Interactors'}});
        my $connect = join(";", @{$$rh_meta_nwa{$id}{'Connectors'}});
        my $type = $$rh_meta_nwa{$id}{'Type'};
        my $int_nodes = join(";", @{$$rh_meta_nwa{$id}{'Int_nodes'}});
        my $reg_nodes = join(";", @{$$rh_meta_nwa{$id}{'Reg_nodes'}});

        print META $id."\t".$size."\t".$inter."\t".$connect."\t".$type."\t".$int_nodes."\t".$reg_nodes."\n";
    }
    close(META);

    # write metasubnets nodes file
    open(NODES, ">$outputDir/meta_nodes.txt") || die "failed to open output file: $!";
    print NODES "# ID\tSymbol\tTF\tFrequency\tGene_degree\tBest subnet score\tSubnets";
    
    my @a_geneList = keys %$rh_meta_noa;
    
    my @a_datasetList = keys(%{$$rh_meta_noa{$a_geneList[0]}{'Correlations'}});
    
    my @ar=();
    foreach my $dataset(@a_datasetList) {
        push(@ar, "Correlation:$dataset");
    }
    print NODES "\t".join(";", @ar);

    print NODES "\tHighlights" if(defined $$rh_meta_noa{$a_geneList[0]}{'Highlights'});
    
    print NODES "\n";
    
    foreach my $id(@a_geneList) {

        my $symbol = $$rh_meta_noa{$id}{'Gene_symbol'};

        my $TF = "-";
        if(defined $$rh_meta_noa{$id}{'TF'}) {  $TF = "TF"; }

        my $freq = $$rh_meta_noa{$id}{'Gene_frequency'};
        my $best = $$rh_meta_noa{$id}{'Best_subnetwork_average_score'};

        my $subnets = ";";
        if(defined $$rh_meta_noa{$id}{'Subnets'}) {
            $subnets = join(";", @{$$rh_meta_noa{$id}{'Subnets'}});
        }

        my $degree = $$rh_meta_noa{$id}{'Gene_degree'};

        print NODES "$id\t$symbol\t$TF\t$freq\t$degree\t$best\t$subnets";

        my @ar=();
        foreach my $dataset(@a_datasetList) {
            push(@ar, "$dataset:$$rh_meta_noa{$id}{'Correlations'}{$dataset}");
        }
        print NODES "\t".join(";", @ar);

        # write highlights gene list
        if(defined $$rh_meta_noa{$id}{'Highlights'}){
            my $highlights = "";
            foreach my $k (keys %{$$rh_meta_noa{$id}{'Highlights'}}){
                if (defined $$rh_meta_noa{$id}{'Highlights'}{$k}){
                    $highlights .= "$k:$$rh_meta_noa{$id}{'Highlights'}{$k};";
                }
                else{ $highlights .= "$k:-;"; }
            }

            print NODES "\t$highlights";
        }
        print NODES "\n";
    }
    close(NODES);

    # write metasubnets nnf file
    open(EDGES, ">$outputDir/meta_edges.nnf") || die "failed to open output file: $!";
    print EDGES "# Network\tProtein_A\tRelationship_type\tProtein_B\n";
    foreach my $subnetID (keys %{$rh_meta_nwa}){
        foreach my $edgeID (@{$$rh_meta_nwa{$subnetID}{'Edges'}}){
            my @temp = split("-", $edgeID);
            my $protA = $temp[0];
            my $protB = $temp[1];

            for my $type(keys %{$$rh_meta_eda{$edgeID}}){
                print EDGES $subnetID."\t".$protA."\t$type\t".$protB."\n";
            }
        }
    }
    close(EDGES);

    # writing subnets list file
    open(SUBNETS, ">$outputDir/subnets.txt") || die "failed to open output file: $!";
    print SUBNETS "# ID\tType\tNodes\tEdges\tScore\tPvalues\n";
    
    foreach my $id (keys %{$rh_subnets}){

        my @split = split("-", $id);
        my $type = "";
        if (Utils::in_array("reg", \@split)) { $type = "reg"; }
        elsif (Utils::in_array("int", \@split)) { $type = "int"; }

        my $nodes = join(";", keys %{$$rh_subnets{$id}{'Nodes'}});
        my $edges = join(";", @{$$rh_subnets{$id}{'Edges'}});

        my @ar=();
        foreach my $dataset(keys %{$$rh_subnets{$id}{'Scores'}}) {
            push(@ar, "$dataset:$$rh_subnets{$id}{'Scores'}{$dataset}");
        }

        my $scores = join(';', @ar);

        my @pval_k = keys(%{$$rh_subnets{$id}{'p-values'}});
        my @pval_v = values(%{$$rh_subnets{$id}{'p-values'}});

        my $pvalues = "$pval_k[0]:$pval_v[0];$pval_k[1]:$pval_v[1];$pval_k[2]:$pval_v[2]";

        print SUBNETS $id."\t".$type."\t".$nodes."\t".$edges."\t".$scores."\t".$pvalues."\n";
    }

    close(SUBNETS);
}

# reads metasubnets files and returns nwa, noa & eda hashes + subnets(int+reg)
sub readMetasubnets {
    my $f_metasubnets = shift;
    my $f_nodes = shift;
    my $f_nnf = shift;
    my $f_subnets = shift;
    my $f_eda   = shift;

    my %h_meta_nwa = ();
    my %h_meta_noa = ();
    my %h_meta_eda = ();
    my %h_subnets = ();

    open(META, $f_metasubnets) || die "Cannot open $f_metasubnets: $!";
    while (<META>) {
        next if($_ =~ m/^#/);
        $_ =~ s/[\r\n]+$//;

        my @a_row    = split(/\t/, $_);

        my $id = $a_row[0];
        my $inter = $a_row[2];
        my $connect = $a_row[3];
        my $type = $a_row[4];

        my $ints = $a_row[5];
        my $regs = $a_row[6];

        next if($id eq 'ID');

        my @a_interactors = split(';', $inter);
        my @a_connectors = split(';', $connect);
        my @a_ints = split(';', $ints);
        my @a_regs = split(';', $regs);

        $h_meta_nwa{$id}{'Interactors'} = \@a_interactors;
        $h_meta_nwa{$id}{'Connectors'} = \@a_connectors;
        $h_meta_nwa{$id}{'Type'} = $type;
        $h_meta_nwa{$id}{'Int_nodes'} = \@a_ints;
        $h_meta_nwa{$id}{'Reg_nodes'} = \@a_regs;
    }
    close(META);

    open(NODES, $f_nodes) || die "Cannot open $f_nodes: $!";
    while (<NODES>) {
        next if($_ =~ m/^#/);
        $_ =~ s/[\r\n]+$//;

        my @a_row    = split(/\t/, $_);

        my $geneID = $a_row[0];
        next if($geneID eq 'ID');

        my $symbol = $a_row[1];
        my $TF = $a_row[2];
        my $freq = $a_row[3];
        my $degree = $a_row[4];
        my $best = $a_row[5];
        my $subnets = $a_row[6];
        my $corr = $a_row[7];
        my $highlights = $a_row[8];

        $h_meta_noa{$geneID}{'Gene_symbol'} = $symbol;
        if ($TF eq "TF") { $h_meta_noa{$geneID}{'TF'} = 1; }
        $h_meta_noa{$geneID}{'Gene_frequency'} = $freq;
        $h_meta_noa{$geneID}{'Gene_degree'} = $degree;
        $h_meta_noa{$geneID}{'Best_subnetwork_average_score'} = $best;
        my @a_subnets = split(';', $subnets);
        $h_meta_noa{$geneID}{'Subnets'} = \@a_subnets;

        my @corr2 = split(';', $corr);
        foreach my $corr3(@corr2) {
            my @ar = split(':', $corr3);
            $h_meta_noa{$geneID}{'Correlations'}{$ar[0]} = $ar[1];
        }

        if (defined $highlights and $highlights ne '') {
            my @high = split(';', $highlights);
            for my $h (@high){
                my @h2 = split(':', $h);
                if ( $h2[1] ne "-" ) { $h_meta_noa{$geneID}{'Highlights'}{$h2[0]} = $h2[1]; }
                else { $h_meta_noa{$geneID}{'Highlights'}{$h2[0]} = undef; }
            }
        }

        foreach my $subID (@a_subnets){
            $h_meta_nwa{$subID}{'Nodes'}{$geneID} = 1;
        }
    }
    close(NODES);

    open(EDGES, $f_nnf) || die "Cannot open $f_nnf: $!";
    while (<EDGES>) {
        next if($_ =~ m/^#/);
        $_ =~ s/[\r\n]+$//;

        my @a_row    = split(/\t/, $_);

        my $metaID = $a_row[0];
        my $protA = $a_row[1];
        my $type = $a_row[2];
        my $protB = $a_row[3];

        my $edge = "$protA-$protB";

        $h_meta_eda{$edge}{$type} = undef;

        if (defined $h_meta_nwa{$metaID}{'Edges'}){
            my @edges = @{$h_meta_nwa{$metaID}{'Edges'}};
            if (! Utils::in_array($edge, \@edges)) {push(@edges, $edge);}
            $h_meta_nwa{$metaID}{'Edges'} = \@edges;
        }
        else{
            my @edges = ();
            push(@edges, $edge);
            $h_meta_nwa{$metaID}{'Edges'} = \@edges;
        }
    }
    close(EDGES);

    if(defined $f_eda){
        #print "Read edge attributes from $f_eda...\n";
        open(EDGEATTRIBUTES, $f_eda) || die "Cannot open $f_eda:\n$!";

        my $edgeNb = scalar(keys %h_meta_eda);

        my $isHeader = 1;
        my(@a_header, %h_atts);
        while (<EDGEATTRIBUTES>) {
            next if($_ =~ m/^#/);
            $_ =~ s/[\r\n]+$//;
            my @a_row = split(/\t/, $_);

            if($isHeader){
                @a_header = @a_row;
                $isHeader = 0;
                next;
            }
            for(my $i = 1; $i < scalar(@a_header); $i++){
                $h_atts{$a_header[$i]} = $a_row[$i];
            }

            my @a_int = split(/\ \(|\)\ /, $a_row[0]);
            my $node1 = $a_int[0];
            my $node2 = $a_int[2];
            my $dir   = $h_atts{'Direction'};
            my $edge  = "$node1-$node2";

            if($edgeNb){
                next if(!defined $h_meta_eda{$edge} ||
                        ($dir eq '--' and !exists $h_meta_eda{$edge}{'pp'}) ||
                        ((scalar(grep(/>/, $dir)) || scalar(grep(/</, $dir))) and
                         !exists $h_meta_eda{$edge}{'pd'})
                       );
            }

            for my $eda(keys %h_atts){
                $h_meta_eda{$edge}{$a_int[1]}{$eda} = $h_atts{$eda};
            }
        }
        close EDGEATTRIBUTES;
    }

    #print  "open $f_subnets\n";
    open(SUB, $f_subnets) || die "Cannot open $f_subnets: $!";
    while (<SUB>) {
        next if($_ =~ m/^#/);
        $_ =~ s/[\r\n]+$//;

        my @a_row    = split(/\t/, $_);

        my $id = $a_row[0];
        my $type = $a_row[1];
        my $nodes = $a_row[2];
        my $edges = $a_row[3];
        my $score = $a_row[4];
        my $pval = $a_row[5];

        next if($id eq 'ID');

        my @a_nodes = split(';', $nodes);
        my @a_edges = split(';', $edges);

        $h_subnets{$id}{'Type'} = $type;

        foreach my $geneID (@a_nodes){
            $h_subnets{$id}{'Nodes'}{$geneID} = 1;
        }

        $h_subnets{$id}{'Edges'} = \@a_edges;

        my @scores = split(';', $score);
        foreach my $corr(@scores) {
            my @ar = split(':', $corr);
            $h_subnets{$id}{'Scores'}{$ar[0]} = $ar[1];
        }

        my @pval = split(";", $pval);
        foreach my $p (@pval){
            my @kv = split(":", $p);
            my $k = $kv[0];
            my $v = $kv[1];
            $h_subnets{$id}{'p-values'}{$k} = $v;
        }

    }
    close(SUB);

    return (\%h_meta_nwa, \%h_meta_noa, \%h_meta_eda, \%h_subnets);
}

# Returns # of common nodes between 2 subnetworks (curr.: subnetA = int, subnetB = reg)
sub overlapCount {
    my ($subnetAid, $subnetBid, $rh_int_subnets, $rh_reg_subnets) = (shift, shift, shift, shift);

    my @connections = ();
    foreach my $geneID (keys %{$$rh_int_subnets{$subnetAid}{'Nodes'}}) {
        if (defined $$rh_reg_subnets{$subnetBid}{'Nodes'}{$geneID}) {
            push(@connections, $geneID);
        }
    }

    return (\@connections);
}

1;
