#!/usr/bin/perl
#
# ITI project
#
# Copyright Maxime Garcia, Ghislain Bidaut, Claire Rioualen, Quentin Da Costa
# (2009-2017)
#
# maxime.u.garcia@gmail.com
# ghislain.bidaut@inserm.fr
# claire.rioualen@inserm.fr
# quentin.da-costa@inserm.fr
#
# http://iti.sourceforge.net
#
# This software is a computer program whose purpose is to superimpose
# interactome over several gene expression datasets to find
# discriminative subnetworks predicting metastatic relapse in breast
# cancer.
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty and the software's author, the holder of the
# economic rights, and the successive licensors have only limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading, using, modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean that it is complicated to manipulate, and that also
# therefore means that it is reserved for developers and experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and, more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package Network;

use strict;
use warnings;
use Utils;

use Data::Dumper;

sub getSizeDistribution {
    my $rh_subnets = shift;
    my @a_sizes;
    for my $rh_snw (values %$rh_subnets) {
        push(@a_sizes, scalar(keys %{$$rh_snw{'Nodes'}}));
    }
    my ($av, $std) = Utils::avstd(@a_sizes);
    return ($av, $std);
}

=pod

=head2 readNetworkDir()

Reads and integrates all interactome data.

Returns:
\%h_totalNetwork(geneID => \@neighbor_genes);
\%h_totalNetworkEdas("$gene1 ($interaction_type) $gene2" => \%($networkID => \@names_of_databases_containing_interaction)).
Cytoscape notation is used for interactions: pp for protein-protein interaction, pd for protein-dna interaction.

=cut

sub readNetworkDir {
    my $d_network = shift;

    my @dirContent = Utils::readInputDir($d_network);

    print "Network directory content = @dirContent\n";

    my (%h_totalNetwork,     %h_totalNetworkEdas, %h_totalInteractome,
        %h_bidirectionalReg, %h_totalRegulome,    @a_allnodes,
        @a_TFs);
    my %h_count = ('int' => 0, 'reg' => 0, 'network' => 0);

    for my $f_current (@dirContent) {
        my @a_tmp = split(/\_/, $f_current);
        my $networkID = $a_tmp[0];
        my @a_nwType = split(/\./, $networkID);
        my $nwType = $a_nwType[0];

        print "\n*********** File = $f_current - Network ID = $networkID - Network type = $nwType\n";

        # Get all interactions (pp or pd); they are considered to be bidirectional.
        my ($rh_network, $rh_edgeAttributes) = parseNetwork("$d_network/$f_current");

        # Regulome
        if ($nwType eq "reg"){
            my $rh_regulome = parseRegulome("$d_network/$f_current");

            # %h_totalRegulome contains protein-dna interactions and there directions.
            for my $node(keys %$rh_regulome) {
                for my $neighborNode (@{$$rh_regulome{$node}}) {
                    unless(defined $h_totalRegulome{$node} and Utils::in_array($neighborNode, $h_totalRegulome{$node})){
                        push(@{$h_totalRegulome{$node}}, $neighborNode);
                        $h_count{"reg"}++;
                    }
                }
            }

            # %h_bidirectionalReg contains bidirectional protein-dna interactions.
            for my $node(keys %$rh_network) {
                for my $neighborNode (@{$$rh_network{$node}}) {
                    unless (defined $h_bidirectionalReg{$node} and Utils::in_array($neighborNode, $h_bidirectionalReg{$node})){
                        push(@{$h_bidirectionalReg{$node}}, $neighborNode);
                    }
                }
            }
        }
        # Interactome
        elsif ($nwType eq "int") {

            for my $node(keys %$rh_network) {
                for my $neighborNode (@{$$rh_network{$node}}) {
                    unless (defined $h_totalInteractome{$node} and Utils::in_array($neighborNode, $h_totalInteractome{$node})) {
                        push(@{$h_totalInteractome{$node}}, $neighborNode);
                        $h_count{"int"}++;
                    }
                }
            }
        }

        # Integration of current network to total network.
        for my $node(keys %$rh_network) {
            my $ra_neighborNodes = $$rh_network{$node};
            for my $neighborNode (@$ra_neighborNodes) {
                unless (defined $h_totalNetwork{$node} and Utils::in_array($neighborNode, $h_totalNetwork{$node})){
                    push(@{$h_totalNetwork{$node}}, $neighborNode);
                    $h_count{"network"}++;
                }
            }

            my $geneID1 = $node;
            for my $geneID2 (@$ra_neighborNodes) {
                my $gi1 = Utils::min($geneID1, $geneID2);
                my $gi2 = Utils::max($geneID1, $geneID2);

                my $ra_edas = $$rh_edgeAttributes{"$gi1-$gi2"};
                if ($nwType eq "reg"){
                    $h_totalNetworkEdas{"$gi1 (pd) $gi2"}{$networkID} = $ra_edas;
                }
                else {
                    $h_totalNetworkEdas{"$gi1 (pp) $gi2"}{$networkID} = $ra_edas;
                }
            }
        }
    }

    @a_allnodes = keys %h_totalNetwork;
    @a_TFs = keys %h_totalRegulome;

    print "\n" . scalar (keys %h_totalInteractome) . " node(s) found in total interactome\n";
    print scalar (@a_TFs) . " transcription factor(s) found in total regulome\n";
    print scalar (@a_allnodes) . " node(s) found in total network\n\n";

    my $nint = $h_count{"int"} / 2;
    print "$nint unique link(s) found in total interactome\n";

    my $nreg = $h_count{"reg"};
    print "$nreg directed link(s) found in total regulome\n";

    my $nglob = $h_count{"network"} / 2;
    print "$nglob unique link(s) found in total network\n\n";

    return \%h_totalNetwork,     \%h_totalNetworkEdas,
           \%h_totalInteractome, \%h_bidirectionalReg,
           \%h_totalRegulome,    \@a_allnodes,         \@a_TFs;
}

=pod

=head2 parseNetwork()

Parses network/interactome file.

=cut

sub parseNetwork {
    my $file = shift;

    print "Reading network file $file\n";
    if (!open(FILE, "<$file")) {die("cannot open $file : $!");}

    my %h_network = ();
    my %h_edgeAttributes = ();

    while (<FILE>) {
        next if ($_ =~ m/^#/);
        $_ =~ s/[\r\n]+$//;

        my @row     = split(/\t/, $_);
        my $geneID1 = $row[1];
        my $geneID2 = $row[3];

        next if ($geneID1 eq $geneID2);

        my @a_annot = ();
        if(defined $row[4]) {
           @a_annot = split(';', $row[4]);
        }
        if(defined $row[5] and $row[5] ne '-') {
           my @a_annot2 = split(';', $row[5]);
           @a_annot = (@a_annot, @a_annot2);
        }

        unless (defined $h_network{$geneID1} and Utils::in_array($geneID2, $h_network{$geneID1})){
            push(@{$h_network{$geneID1}}, $geneID2);
        }
        unless (defined $h_network{$geneID2} and Utils::in_array($geneID1, $h_network{$geneID2})){
            push(@{$h_network{$geneID2}}, $geneID1);
        }

        my $gi1 = Utils::min($geneID1, $geneID2);
        my $gi2 = Utils::max($geneID1, $geneID2);

        if (scalar @a_annot){
            foreach my $an (@a_annot) {
                unless (defined $h_edgeAttributes{"$gi1-$gi2"} and Utils::in_array($an, $h_edgeAttributes{"$gi1-$gi2"})) {
                    push(@{$h_edgeAttributes{"$gi1-$gi2"}}, $an);
                }
            }
        }
        else{
            unless(defined $h_edgeAttributes{"$gi1-$gi2"}){
                $h_edgeAttributes{"$gi1-$gi2"} = \@a_annot;
            }
        }
    }

    close(FILE);

    print "Number of individual node(s) = " . (scalar keys %h_network) . "\n";
    my $ninter = 0;
    while (my ($k, $v) = each(%h_network)) {$ninter += scalar @$v;}
    $ninter /= 2;
    print "$ninter unique link(s) found\n";
    return \%h_network, \%h_edgeAttributes;
}

=pod

=head2 parseRegulome()

Parses regulome file.

=cut

sub parseRegulome{
    my $file = shift;

    print "Reading regulome file $file\n";
    if (!open(FILE, "<$file")) {die("cannot open $file : $!");}

    my %h_regulome = ();
    while(defined(my $l = <FILE>)){
        next if ($l =~ m/^#/);
        $l =~ s/[\r\n]+$//;
        my @TFrow = split('\t', $l);

        next if ($TFrow[1] eq  $TFrow[3]);

        unless (defined $h_regulome{$TFrow[1]} and Utils::in_array($TFrow[3], $h_regulome{$TFrow[1]})) {
            push(@{$h_regulome{$TFrow[1]}}, $TFrow[3]);
        }
    }
    return \%h_regulome;
}

=pod

=head2 createRandomNetwork()

Generates random network.

=cut

sub createRandomNetworkNew {
    my $rh_network = shift;

    my $itaNb = 0;
    my %h_nNodes;
    my %h_randomNodes;
    while (my ($node, $ra_neiNodes) = each(%$rh_network)) {
        my $neiNb = scalar @$ra_neiNodes;
        $itaNb += $neiNb;

        my $maj = sprintf("%.0f", $neiNb/10);
        if($maj > 0){ $neiNb += $maj;}
        else        { $neiNb ++;}
        $h_nNodes{$node} = $neiNb;
        $h_randomNodes{$node} = 0;
    }
    $itaNb /= 2;

    my @a_nodes = keys %h_nNodes;
    my $nodeNb = scalar @a_nodes;

    print "Input network:
$nodeNb nodes
$itaNb unique interactions
Generating random interactome...\n";

    my($rdmItaNb, $compNb);
    my %h_randomNetwork;

    for($rdmItaNb = 0; $rdmItaNb < $itaNb; $rdmItaNb++) {
        my $i = int(rand($nodeNb));
        my $rdmNode1 = $a_nodes[$i];

        my($rdmNode2, $j, $times);
        do{
            $j = int(rand($nodeNb)); $rdmNode2 = $a_nodes[$j]; $times++;
            if($times == 5000){
                # warn "$nodeNb incomplete nodes:\n";
                # for my $n(@a_nodes){ warn "$h_randomNodes{$n}/$h_nNodes{$n} neighbors\n";}
                # warn $compNb . " complete nodes\n";
                last;
            }
        } while($rdmNode2 eq $rdmNode1 || Utils::in_array($rdmNode2, $h_randomNetwork{$rdmNode1}));

        push(@{$h_randomNetwork{$rdmNode1}}, $rdmNode2);
        push(@{$h_randomNetwork{$rdmNode2}}, $rdmNode1);
        $h_randomNodes{$rdmNode1}++;
        $h_randomNodes{$rdmNode2}++;

        if($h_randomNodes{$rdmNode1} == $h_nNodes{$rdmNode1}){
            splice(@a_nodes, $i, 1);
            $j-- if($j > $i);
            $nodeNb--;
            $compNb++;
        }
        if($h_randomNodes{$rdmNode2} == $h_nNodes{$rdmNode2}){
            splice(@a_nodes, $j, 1);
            $nodeNb--;
            $compNb++;
        }

        # if (($rdmItaNb % 10 == 0)) {
        #     print "$rdmItaNb/$itaNb done\n";
        # }
        if($nodeNb == 1){
            # warn "Node incomplete: $h_randomNodes{$a_nodes[0]}/$h_nNodes{$a_nodes[0]} neighbors\n"
            # . $compNb . " complete nodes\n";
            last;
        }
    }

    print "$rdmItaNb random interactions\n";

    return \%h_randomNetwork;
}


=pod

=head2 createRandomNetwork()

Generates random network.

=cut

sub createRandomNetwork {
    my $rh_interactome = shift;
    my %nNodes;
    while (my ($key, $value) = each(%$rh_interactome)) {
        $nNodes{$key} = scalar @$value;
    }
    print "Generating random interactome\n";
    print "Input interactome size: " . scalar(keys %nNodes) . "\n";
    my $ninter = 0;
    my %randomNNodes;
    while (my ($k, $v) = each(%nNodes)) {
        $ninter += $v;
        $randomNNodes{$k} = 0;
    }
    $ninter /= 2;

    my %randomInteractome;
    my $currentNinter = 0;

    print "$ninter unique interaction(s) found\n";
    my $hashSize = scalar keys %nNodes;

    my @nNodesAr = keys %nNodes;
    my $k;

    while ($currentNinter < $ninter) {
        do {$k = $nNodesAr[int(rand($hashSize))];}
          while (($randomNNodes{$k} - $nNodes{$k}) >= 1);

        my $randomKey;
        my $indic = 0;
        do {$randomKey = $nNodesAr[int(rand($hashSize))];}
          while ($randomKey eq $k
            || Utils::in_array($randomKey, $randomInteractome{$k})
            || ($randomNNodes{$randomKey} - $nNodes{$randomKey}) >= 4);

        if (defined $randomInteractome{$k}) {
            push(@{$randomInteractome{$k}}, $randomKey);
        }
        else {
            my @ar;
            push(@ar, $randomKey);
            $randomInteractome{$k} = \@ar;
        }
        if (defined $randomInteractome{$randomKey}) {
            push(@{$randomInteractome{$randomKey}}, $k);
        }
        else {
            my @ar;
            push(@ar, $k);
            $randomInteractome{$randomKey} = \@ar;
        }
        $randomNNodes{$k}++;
        $randomNNodes{$randomKey}++;
        $currentNinter++;
    }
    $ninter = 0;
    for my $key (keys %randomInteractome) {
        my @value = @{$randomInteractome{$key}};
        $ninter += scalar @value;
    }
    $ninter /= 2;
    print "Random interactome generated - ninteractions = $ninter\n";
    return \%randomInteractome;
}




=pod

=head2 readGOterms()

Parses associated GO subnetworks files

=cut
sub readGOterms{
    my $f_go_gene = shift;
    my $f_go_subnet = shift;
    my $rh_nwa = shift;
    my $rh_noa = shift;

    my %h_GOterms = ();

    open(GENE, $f_go_gene) || die "Cannot open $f_go_gene: $!";
    while (<GENE>) {
        next if($_ =~ m/^#/);
        $_ =~ s/[\r\n]+$//;
        my @a_row    = split(/\t/, $_);

        my $geneID = $a_row[0];
        my $GOID = $a_row[1];
        my $GOterm = $a_row[2];

        $h_GOterms{$GOID}{'term'} = $GOterm;
        push(@{$h_GOterms{$GOID}{'genes'}}, $geneID);

        push(@{$$rh_noa{$geneID}{'GO_terms'}}, $GOID);

    }
    close(GENE);

    open(SUBNET, $f_go_subnet) || die "Cannot open $f_go_subnet: $!";
    while (<SUBNET>) {
        next if($_ =~ m/^#/);
        $_ =~ s/[\r\n]+$//;
        my @a_row    = split(/\t/, $_);

        my $subnetID = $a_row[0];
        my $GOID = $a_row[1];
        my $GOterm = $a_row[2];
        my $subnetOcc = $a_row[3];
        my $totalOcc = $a_row[4];
        my $sub_gene_num = $a_row[5];
        my $anno_gene_num = $a_row[6];
        my $enrichRatio = $a_row[7];
        my $hyper = $a_row[8];
        my $corrpval = $a_row[9];
        my $signif = $a_row[10];

        $h_GOterms{$GOID}{'term'} = $GOterm;
        push(@{$h_GOterms{$GOID}{'subnets'}}, $subnetID);
        $h_GOterms{$GOID}{'network_occurrences'} = $totalOcc;

        $$rh_nwa{$subnetID}{'GO_terms'}{$GOID}{'hyper'} = $hyper;
        $$rh_nwa{$subnetID}{'GO_terms'}{$GOID}{'corrected_pval'} = $corrpval;
        $$rh_nwa{$subnetID}{'GO_terms'}{$GOID}{'BH_correction'} = $signif;
        $$rh_nwa{$subnetID}{'GO_terms'}{$GOID}{'Enrichment_ratio'} = $enrichRatio;
        $$rh_nwa{$subnetID}{'GO_terms'}{$GOID}{'occurrences'} = $subnetOcc;

    }
    close(SUBNET);

    return \%h_GOterms;
}


=pod

=head2 calcGlobalGOStatistics()

Calculate GO occurence in network

=cut
sub calcNetworkGlobalGOStatistics {

    my($rh_network, $rh_gene2GOs) = (shift, shift);

    my @a_allGenes  = keys %$rh_network;
    my $totalGeneNb = scalar(@a_allGenes);
    print "$totalGeneNb genes found in network file\n\n";

    my %h_GOccur;
    for my $gene(@a_allGenes){
        next if(!defined $$rh_gene2GOs{$gene});
        for my $GO(keys %{$$rh_gene2GOs{$gene}}){
            $h_GOccur{$GO}++;
        }
    }

    return ($totalGeneNb, \%h_GOccur);
}


=pod

=head2 calCdatasetsOverlapWithNetwork

Calculate GO occurence in network

=cut
sub calcDatasetsOverlapWithNetwork {
    my $rh_datasets = shift;
    my $rh_network = shift;
    
    my %h_datasetsOverlap;
    
    my @a_datasets = keys %$rh_datasets;
    for my $label (@a_datasets) {
	my $overlap = 0;
	
	for my $gene(keys %{$$rh_datasets{$label}{'collapseddata'}}) {
	    if(exists $$rh_network{$gene}) { $overlap++ };
	}
	$h_datasetsOverlap{$label} = $overlap;
    }

    for my $label (@a_datasets) {
	my $overlapRate = ($h_datasetsOverlap{$label}/scalar keys %$rh_network)*100;
	print "overlap dataset $label with network: $h_datasetsOverlap{$label}(". sprintf("%.2f", $overlapRate) ."%)\n";
	if($overlapRate<5) { die "Overlap too small(<5%) between dataset $label and interaction data\n";
	
	}
    }
    return \%h_datasetsOverlap;
}

1;
