#!/usr/bin/perl
#
# ITI project
#
# Copyright Maxime Garcia, Ghislain Bidaut, Claire Rioualen, Quentin Da Costa
# (2009-2017)
#
# maxime.u.garcia@gmail.com
# ghislain.bidaut@inserm.fr
# claire.rioualen@inserm.fr
# quentin.da-costa@inserm.fr
#
# http://iti.sourceforge.net
#
# This software is a computer program whose purpose is to superimpose
# interactome over several gene expression datasets to find
# discriminative subnetworks predicting metastatic relapse in breast
# cancer.
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty and the software's author, the holder of the
# economic rights, and the successive licensors have only limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading, using, modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean that it is complicated to manipulate, and that also
# therefore means that it is reserved for developers and experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and, more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package Subnetworks;

use strict;
use warnings;
use Utils;
use Data::Dumper;

=pod

=head2 writeSubnets()

Writes network file (.nnf) $f_outSnw and attribute files (node and network attributes) $f_outNoa and $f_outNwa
which are Cytoscape compatible, from hashes %$rh_subnets_nwa, %$rh_nodeAtts and %$rh_edgeAtts (See structures in the function).

File formats (supported by Cytoscape):
Network (nnf): Network    Protein_A    Relationship_type    Protein_B
Network attributes: Network    score-$dataset_label    p_value-$dist_head-$dataset_label
Node attributes: Gene_ID    Symbol    Transcription_factor    correlation-$dataset_label    Gene_frequency    Best_subnetwork_average_score    Gene_degree

Additional columns if CNV data are analyzed:
Network attributes: CNV_score-$dataset_label    CNV_p-value-$dist_head-$dataset_label
Node attributes: GISTIC_score-$dataset_label    CNV_final_score-$dataset_label    Putative_driver_gene
(4 GISTIC scores by dataset)

=cut

sub writeSubnets {
    my $rh_subnets_nwa    = shift;
    my $rh_nodeAtts       = shift;
    my $rh_edgeAtts       = shift;
    my $f_outSnw          = shift;
    my $f_outNwa          = shift;
    my $f_outNoa          = shift;
    my $randomNetworkUsed = shift;

    # General attributes:
    # Subnetworks / Network attributes hash:
    # \%rh_subnets_nwa($subnetwork_ID => \%('Edges' => \@("$gene_A-$gene_B"),
    #                                       'Nodes' => \%($gene_ID => 1),
    #                                       'Scores' => \%($dataset_label => $subnetwork_score),
    #                                       'p-values' => \%("$dist_head-$dataset_label" => $subnetwork_p_value)))
    # Node attributes hash:
    # \%rh_nodeAtts($gene_ID => \%('Correlations' => \%($dataset_label => $gene_score),
    #                              'Gene_symbol' => $gene_symbol,
    #                              'TF' => 1/undef,
    #                              'Gene_frequency' => $gene_frequency,
    #                              'Best_subnetwork_average_score' => $best_snw_average_score,
    #                              'Gene_degree' => $gene_degree))
    # Edge attributes hash:
    # \%rh_edgeAtts("$gene_A-$gene_B" => \%($interaction_type => \%('Direction' => $direction_symbol)))
    #
    # If CNV data are analyzed:
    # \%rh_subnets_nwa($subnetwork_ID => \%('CNV_scores' => \%($dataset_label => $subnetwork_score),
    #                                       'CNV_p-values' => \%("$dist_head-$dataset_label" => $subnetwork_p_value)))
    # \%rh_nodeAtts($geneID => \%('GISTIC_scores' => \%("CNV_score_label-$datasetLabel" => $CNV_score),
    #                             'Putative_driver_gene' => 1/undef))

    my @a_subnetIDs;
    if(defined $f_outSnw || defined $f_outNwa){
        if(!defined $rh_subnets_nwa || scalar(keys %$rh_subnets_nwa) == 0){
            warn "\nWarning: No subnetwork found!\n";
            return;
        }

        @a_subnetIDs = keys %$rh_subnets_nwa;
        #print "Writing " . scalar(@a_subnetIDs) . " subnetworks...\n";
    }


    my %h_presentGeneIDs;
    if(defined $f_outSnw){
        #print "> $f_outSnw\n";
        open(SUBNETS, ">$f_outSnw") || die "Cannot open $f_outSnw:\n$!";
        print SUBNETS "# Network\tProtein_A\tRelationship_type\tProtein_B\n";

        ## Write data ##

        my @a_networkSize;
        for my $netID (@a_subnetIDs) {
            my $rh_snw = $$rh_subnets_nwa{$netID};

            if(!defined $$rh_snw{'Edges'} || scalar(@{$$rh_snw{'Edges'}}) == 0){
                warn "\nWarning: No edge in $netID!\n\n";
                next;
            }

            for my $int(@{$$rh_snw{'Edges'}}){
                my @a_int = split(/-/, $int);
                my $g1 = $a_int[0];
                my $g2 = $a_int[1];

                $h_presentGeneIDs{$g1} = 1;
                $h_presentGeneIDs{$g2} = 1;

                if (!defined $randomNetworkUsed){
                    die "Missing interaction type(s)\n" if(!defined $$rh_edgeAtts{$int});
                    for my $intType (keys %{$$rh_edgeAtts{$int}}){
                        print SUBNETS "$netID\t$g1\t$intType\t$g2\n";
                    }
                }
                else{
                    print SUBNETS "$netID\t$g1\tRandom_interaction\t$g2\n";
                }
            }

            my $networkSize = scalar(keys %{$$rh_snw{'Nodes'}});
            push(@a_networkSize, $networkSize);
            #print "ID = $netID - Size = $networkSize\n";
        }

        close SUBNETS;

        my($avNetworkSize, $stdNetworkSize) = Utils::avstd(@a_networkSize);
        #print "Average subnetwork size = $avNetworkSize, stddev = $stdNetworkSize\n";
    }


    if(defined $f_outNwa){
        my @a_nwAttKeys = keys %{$$rh_subnets_nwa{$a_subnetIDs[0]}};
        my @a_scoreKeys = keys %{$$rh_subnets_nwa{$a_subnetIDs[0]}{'Scores'}};
        my @a_cnvScoreKeys;
        @a_cnvScoreKeys = keys %{$$rh_subnets_nwa{$a_subnetIDs[0]}{'CNV_scores'}}
            if(defined $$rh_subnets_nwa{$a_subnetIDs[0]}{'CNV_scores'});
        my @a_pValKeys;
        @a_pValKeys     = keys %{$$rh_subnets_nwa{$a_subnetIDs[0]}{'p-values'}}
            if(defined $$rh_subnets_nwa{$a_subnetIDs[0]}{'p-values'});
        my @a_CNVpValKeys;
        @a_CNVpValKeys  = keys %{$$rh_subnets_nwa{$a_subnetIDs[0]}{'CNV_p-values'}}
            if(defined $$rh_subnets_nwa{$a_subnetIDs[0]}{'CNV_p-values'});

        ## Write header ##

        #print "> $f_outNwa\n";
        open(NETWORKATTRIBUTES, ">$f_outNwa") || die "Cannot open $f_outNwa:\n$!";
        print NETWORKATTRIBUTES "Network";

        for my $head (@a_scoreKeys) {
            print NETWORKATTRIBUTES "\tscore-$head";
        }
        if(Utils::in_array('Average_score', \@a_nwAttKeys)){
            print NETWORKATTRIBUTES "\tAverage_score";
        }
        if(Utils::in_array('Seed_occurrence', \@a_nwAttKeys)){
            print NETWORKATTRIBUTES "\tSeed_occurrence";
        }
        if(Utils::in_array('Merged_snw_best_score', \@a_nwAttKeys)){
            print NETWORKATTRIBUTES "\tMerged_snw_best_score";
        }
        for my $head (@a_pValKeys) {
            print NETWORKATTRIBUTES "\tp_value-$head";
        }

        for my $head (@a_cnvScoreKeys) {
            print NETWORKATTRIBUTES "\tCNV_score-$head";
        }
        for my $head (@a_CNVpValKeys) {
            print NETWORKATTRIBUTES "\tCNV_p_value-$head";
        }
        if(Utils::in_array('Recurrent_nodes', \@a_nwAttKeys)){
            print NETWORKATTRIBUTES "\tRecurrent_nodes";
        }
        if(Utils::in_array('Merged_snw_seeds', \@a_nwAttKeys)){
            print NETWORKATTRIBUTES "\tMerged_snw_seeds";
        }

        print NETWORKATTRIBUTES "\n";


        ## Write data ##

        for my $netID (@a_subnetIDs) {
            print NETWORKATTRIBUTES $netID;
            for my $head (@a_scoreKeys) {
                if (defined $$rh_subnets_nwa{$netID}{'Scores'}{$head}) {
                    print NETWORKATTRIBUTES "\t$$rh_subnets_nwa{$netID}{'Scores'}{$head}";
                }
                else{
                    print NETWORKATTRIBUTES "\tundef";
                }
            }

            if(Utils::in_array('Average_score', \@a_nwAttKeys)){
                if(defined $$rh_subnets_nwa{$netID}{'Average_score'}){
                    print NETWORKATTRIBUTES "\t$$rh_subnets_nwa{$netID}{'Average_score'}";
                }
                else{
                    print NETWORKATTRIBUTES "\t-";
                }
            }

            if(Utils::in_array('Seed_occurrence', \@a_nwAttKeys)){
                print NETWORKATTRIBUTES "\t$$rh_subnets_nwa{$netID}{'Seed_occurrence'}";
            }

            if(Utils::in_array('Merged_snw_best_score', \@a_nwAttKeys)){
                if(defined $$rh_subnets_nwa{$netID}{'Merged_snw_best_score'}){
                    print NETWORKATTRIBUTES "\t$$rh_subnets_nwa{$netID}{'Merged_snw_best_score'}";
                }
                else{
                    print NETWORKATTRIBUTES "\t-";
                }
            }

            for my $head (@a_pValKeys){
                if (defined $$rh_subnets_nwa{$netID}{'p-values'}{$head}) {
                    print NETWORKATTRIBUTES "\t$$rh_subnets_nwa{$netID}{'p-values'}{$head}";
                }
                else{
                    print NETWORKATTRIBUTES "\tundef";
                }
            }

            for my $head (@a_cnvScoreKeys) {
                if (defined $$rh_subnets_nwa{$netID}{'CNV_scores'}{$head}) {
                    print NETWORKATTRIBUTES "\t$$rh_subnets_nwa{$netID}{'CNV_scores'}{$head}";
                }
                else{
                    print NETWORKATTRIBUTES "\tundef";
                }
            }

            for my $head (@a_CNVpValKeys){
                if (defined $$rh_subnets_nwa{$netID}{'CNV_p-values'}{$head}) {
                    print NETWORKATTRIBUTES "\t$$rh_subnets_nwa{$netID}{'CNV_p-values'}{$head}";
                }
                else{
                    print NETWORKATTRIBUTES "\tundef";
                }
            }

            if(Utils::in_array('Recurrent_nodes', \@a_nwAttKeys)){
                if(defined $$rh_subnets_nwa{$netID}{'Recurrent_nodes'} and
                   scalar @{$$rh_subnets_nwa{$netID}{'Recurrent_nodes'}}){
                    print NETWORKATTRIBUTES "\t" . join(';', @{$$rh_subnets_nwa{$netID}{'Recurrent_nodes'}});
                }
                else{
                    print NETWORKATTRIBUTES "\t-";
                }
            }

            if(Utils::in_array('Merged_snw_seeds', \@a_nwAttKeys)){
                if(defined $$rh_subnets_nwa{$netID}{'Merged_snw_seeds'} and
                   scalar @{$$rh_subnets_nwa{$netID}{'Merged_snw_seeds'}}){
                    print NETWORKATTRIBUTES "\t" . join(';', @{$$rh_subnets_nwa{$netID}{'Merged_snw_seeds'}});
                }
                else{
                    print NETWORKATTRIBUTES "\t-";
                }
            }

            print NETWORKATTRIBUTES "\n";
        }

        close NETWORKATTRIBUTES;
    }


    if(defined $f_outNoa){

        my @a_geneIDs;
        if(scalar(keys %h_presentGeneIDs)){
            @a_geneIDs = keys %h_presentGeneIDs;
        }
        else{
            @a_geneIDs = keys %$rh_nodeAtts;
        }

        my @a_nodeAttKeys  = keys %{$$rh_nodeAtts{$a_geneIDs[0]}};
        my @a_correlKeys   = keys %{$$rh_nodeAtts{$a_geneIDs[0]}{'Correlations'}};
        my @a_gisticScoreKeys;
        @a_gisticScoreKeys = keys %{$$rh_nodeAtts{$a_geneIDs[0]}{'GISTIC_scores'}}
            if(defined $$rh_nodeAtts{$a_geneIDs[0]}{'GISTIC_scores'});
        my @a_highlightKeys;
        @a_highlightKeys   = keys %{$$rh_nodeAtts{$a_geneIDs[0]}{'Highlights'}}
            if(defined $$rh_nodeAtts{$a_geneIDs[0]}{'Highlights'});

        ## Write header ##

        #print "> $f_outNoa\n";
        open(NODEATTRIBUTES, ">$f_outNoa") || die "Cannot open $f_outNoa:\n$!";
        print NODEATTRIBUTES "Gene_ID\tSymbol";

        if(Utils::in_array('TF', \@a_nodeAttKeys)){
            print NODEATTRIBUTES "\tTranscription_factor";
        }

        for my $head (@a_highlightKeys) {
            print NODEATTRIBUTES "\tHighlight-$head";
        }

        if(Utils::in_array('Present_in', \@a_nodeAttKeys)){
            print NODEATTRIBUTES "\tPresent_in";
        }

        for my $head (@a_correlKeys) {
            print NODEATTRIBUTES "\tcorrelation-$head";
        }

        if(Utils::in_array('Average_correlation', \@a_nodeAttKeys)){
            print NODEATTRIBUTES "\tAverage_correlation";
        }

        if(Utils::in_array('Gene_frequency', \@a_nodeAttKeys)){
            print NODEATTRIBUTES "\tGene_frequency";
        }
        if(Utils::in_array('Best_subnetwork_average_score', \@a_nodeAttKeys)){
            print NODEATTRIBUTES "\tBest_subnetwork_average_score";
        }
        if(Utils::in_array('Gene_degree', \@a_nodeAttKeys)){
            print NODEATTRIBUTES "\tGene_degree";
        }

        for my $head (@a_gisticScoreKeys) {
            print NODEATTRIBUTES "\t$head";
        }

        if(scalar(@a_gisticScoreKeys) > 1){
            print NODEATTRIBUTES "\tAverage_CNV_score\tPutative_driver_gene";
        }

        print NODEATTRIBUTES "\n";


        ## Write data ##

        for my $geneID (@a_geneIDs) {

            if (defined $$rh_nodeAtts{$geneID}{'Gene_symbol'}) {
                print NODEATTRIBUTES "$geneID\t$$rh_nodeAtts{$geneID}{'Gene_symbol'}";
            }
            else {
                print NODEATTRIBUTES "$geneID\t";
            }

            if(Utils::in_array('TF', \@a_nodeAttKeys)){
                if(defined $$rh_nodeAtts{$geneID}{'TF'}){
                    print NODEATTRIBUTES "\tTF";
                }
                else{
                    print NODEATTRIBUTES "\t-";
                }
            }

            for my $head (@a_highlightKeys) {
                my $annot = $$rh_nodeAtts{$geneID}{'Highlights'}{$head};
                if (defined $annot) {
                    print NODEATTRIBUTES "\t$annot";
                }
                else {
                    print NODEATTRIBUTES "\t-";
                }
            }

            if(Utils::in_array('Present_in', \@a_nodeAttKeys)){
                print NODEATTRIBUTES "\t";
                my $s;
                for my $study(keys %{$$rh_nodeAtts{$geneID}{'Present_in'}} ){
                    $s .= ";" if(defined $s);
                    $s .= "$study:" . join(',', keys(%{$$rh_nodeAtts{$geneID}{'Present_in'}{$study}}));
                }
                print NODEATTRIBUTES $s;
            }


            for my $head (@a_correlKeys) {
                if (defined $$rh_nodeAtts{$geneID}{'Correlations'}{$head}) {
                    print NODEATTRIBUTES "\t$$rh_nodeAtts{$geneID}{'Correlations'}{$head}";
                }
                else {
                    print NODEATTRIBUTES "\tundef";
                }
            }

            if(Utils::in_array('Average_correlation', \@a_nodeAttKeys)){
                if(defined $$rh_nodeAtts{$geneID}{'Average_correlation'}){
                    print NODEATTRIBUTES "\t$$rh_nodeAtts{$geneID}{'Average_correlation'}";
                }
                else{
                    print NODEATTRIBUTES "\t-";
                }
            }

            if(Utils::in_array('Gene_frequency', \@a_nodeAttKeys)){
                print NODEATTRIBUTES "\t$$rh_nodeAtts{$geneID}{'Gene_frequency'}";
            }
            if(Utils::in_array('Best_subnetwork_average_score', \@a_nodeAttKeys)){
                print NODEATTRIBUTES "\t$$rh_nodeAtts{$geneID}{'Best_subnetwork_average_score'}";
            }
            if(Utils::in_array('Gene_degree', \@a_nodeAttKeys)){
                print NODEATTRIBUTES "\t$$rh_nodeAtts{$geneID}{'Gene_degree'}";
            }

            for my $head (@a_gisticScoreKeys) {
                if (defined $$rh_nodeAtts{$geneID}{'GISTIC_scores'}{$head}) {
                    print NODEATTRIBUTES "\t$$rh_nodeAtts{$geneID}{'GISTIC_scores'}{$head}";
                }
                else {
                    print NODEATTRIBUTES "\tundef";
                }
            }

            if(scalar(@a_gisticScoreKeys) > 1){
                if(defined $$rh_nodeAtts{$geneID}{'Average_CNV_score'}){
                    print NODEATTRIBUTES "\t$$rh_nodeAtts{$geneID}{'Average_CNV_score'}";
                }
                else{
                    print NODEATTRIBUTES "\t-";
                }

                if(defined $$rh_nodeAtts{$geneID}{'Putative_driver_gene'}){
                    print NODEATTRIBUTES "\tPDG";
                }
                else{
                    print NODEATTRIBUTES "\t-";
                }
            }

            print NODEATTRIBUTES "\n";
        }

        close NODEATTRIBUTES;
    }
}

=pod

=head2 readSubnets()

Reads network file (.nnf) and attribute files (node, edge and network attributes) and stores informations in adapted hash.

Returns general attributes:

=over

=item Subnetworks / Network attributes hash:
%h_subnets_nwa($subnetwork_ID => \%('Edges' => \@("$gene_A-$gene_B"),
                                    'Nodes' => \%($gene_ID => 1))),
                                    'Scores' => \%($dataset_label => $subnetwork_score)),
                                    'p-values' => \%("$dist_head-$dataset_label" => $subnetwork_p_value))
=item Node attributes hash:
%h_nodeAtts($geneID => \%('Correlations' => \%($dataset_label => $gene_score),
                          'Gene_symbol' => $gene_symbol,
                          'TF' => 1/undef,
                          'Gene_frequency' => $gene_frequency,
                          'Best_subnetwork_average_score' => $best_snw_average_score,
                          'Gene_degree' => $gene_degree))
=item Edge attributes hash:
%h_edgeAtts("$gene_A-$gene_B" => \%($interaction_type => \%('Direction' => $direction_symbol,
                                                            'Sources' => $origins)))

=item If CNV data are analyzed:
%h_subnets_nwa($subnetwork_ID => \%('CNV_scores' => \%($dataset_label => $subnetwork_score),
                                    'CNV_p-values' => \%("$dist_head-$dataset_label" => $subnetwork_p_value)))
%h_nodeAtts($geneID => \%('GISTIC_scores' => \%("CNV_score_label-$dataset_label" => $CNV_score),
                          'Putative_driver_gene' => 1/undef))

=back

=cut

sub readSubnets {
    my $f_nnf   = shift;
    my $f_nwa   = shift;
    my $f_noa   = shift;
    my $f_eda   = shift;
    my $ra_snws = shift;    # Only these subnetworks will be loaded.

    # File formats (supported by Cytoscape):
    # Network (nnf): Network    Protein_A    Relationship_type    Protein_B
    # Network attributes: Network    score-$dataset_label    p_value-$dist_head-$dataset_label
    # Node attributes: Gene_ID    Symbol    Transcription_factor    correlation-$dataset    Average_correlation
    # Edge attributes: Edge    Direction    Sources
    #
    # Additional columns if CNV data are analyzed:
    # Network attributes: CNV_score-$dataset_label    CNV_p-value-$dist_head-$dataset_label
    # Node attributes: GISTIC_score-$dataset_label    CNV_final_score-$dataset_label    Putative_driver_gene
    # (4 GISTIC_score by dataset)

    my(%h_subnets_nwa, %h_edgeAtts,
       %h_nodeAtts);

    if(defined $f_nnf){
        #print "Read subnetworks from $f_nnf...\n";
        open(SUBNET, $f_nnf) || die "Cannot open $f_nnf:\n$!";

        while (<SUBNET>) {
            next if($_ =~ m/^#/);
            $_ =~ s/[\r\n]+$//;
            my @a_row    = split(/\t/, $_);
            my $subnetID = $a_row[0];
            next if(defined $ra_snws and !Utils::in_array($subnetID, $ra_snws));
            my $g1       = $a_row[1];
            my $intType  = $a_row[2];
            my $g2       = $a_row[3];

            my @a_subnetID = split(/-/, $subnetID);
            my $seed       = $a_subnetID[-1];
            $h_subnets_nwa{$subnetID}{'Seed'} = $seed;

            my $gi1 = Utils::min($g1, $g2);
            my $gi2 = Utils::max($g1, $g2);
            push(@{$h_subnets_nwa{$subnetID}{'Edges'}}, "$gi1-$gi2")
                unless(Utils::in_array("$gi1-$gi2", \@{$h_subnets_nwa{$subnetID}{'Edges'}}));

            $h_subnets_nwa{$subnetID}{'Nodes'}{$g1} = 1;
            $h_subnets_nwa{$subnetID}{'Nodes'}{$g2} = 1;

            $h_edgeAtts{"$gi1-$gi2"}{$intType} = undef;

            push(@{$h_nodeAtts{$g1}{'Subnets'}}, $subnetID) unless(defined $h_nodeAtts{$g1}{'Subnets'} and
                                                                   Utils::in_array($subnetID, $h_nodeAtts{$g1}{'Subnets'}));
            push(@{$h_nodeAtts{$g2}{'Subnets'}}, $subnetID) unless(defined $h_nodeAtts{$g2}{'Subnets'} and
                                                                   Utils::in_array($subnetID, $h_nodeAtts{$g2}{'Subnets'}));
        }
        close SUBNET;
    }

    if(defined $f_noa){
        #print "Read node attributes from $f_noa...\n";
        open(NODEATTRIBUTES, $f_noa)
          || die "Cannot open $f_noa:\n$!";

        my $isHeader = 1;
        my (@a_nodeAttHeads,     @a_corrIndexes,   @a_gisticScoreIndexes,
            $tfIndex,            $avCorrIndex,     $geneFreqIndex,
            $bestAvScoreIndex,   $degree,          $avCNVScoreIndex,
            $putDriverGeneIndex, @a_highlightIndexes, $studiesIndex);
        while (<NODEATTRIBUTES>) {
            next if($_ =~ m/^#/);
            $_ =~ s/[\r\n]+$//;
            my @a_row = split(/\t/, $_);

            if($isHeader){
                @a_nodeAttHeads = @a_row;

                for (my $i = 2; $i < scalar(@a_nodeAttHeads); $i++){
                    if($a_nodeAttHeads[$i] =~ m/^correlation/){
                        push(@a_corrIndexes, $i);
                        $a_nodeAttHeads[$i] =~ s/^correlation-//;
                    }
                    elsif($a_nodeAttHeads[$i] =~ m/Transcription_factor/){
                        $tfIndex = $i;
                    }
                    elsif($a_nodeAttHeads[$i] =~ m/Highlight/){
                        push(@a_highlightIndexes, $i);
                        $a_nodeAttHeads[$i] =~ s/^Highlight-//;
                    }
                    elsif($a_nodeAttHeads[$i] =~ m/Average_correlation/){
                        $avCorrIndex = $i;
                    }
                    elsif($a_nodeAttHeads[$i] =~ m/Gene_frequency/){
                        $geneFreqIndex = $i;
                    }
                    elsif($a_nodeAttHeads[$i] =~ m/Best_subnetwork_average_score$/){
                        $bestAvScoreIndex = $i;
                    }
                    elsif($a_nodeAttHeads[$i] =~ m/Gene_degree$/){
                        $degree = $i;
                    }
                    elsif($a_nodeAttHeads[$i] =~ m/GISTIC_score|CNV_final/){
                        push(@a_gisticScoreIndexes, $i);
                    }
                    elsif($a_nodeAttHeads[$i] =~ m/Average_CNV_score/){
                        $avCNVScoreIndex = $i;
                    }
                    elsif($a_nodeAttHeads[$i] =~ m/Putative_driver_gene/){
                        $putDriverGeneIndex = $i;
                    }
                    elsif($a_nodeAttHeads[$i] =~ m/Present_in/){
                        $studiesIndex = $i;
                    }
                }
                $isHeader = 0;
            }
            else{
                my $geneID = $a_row[0];
                my $sym = $a_row[1];
                $h_nodeAtts{$geneID}{'Gene_symbol'} = $sym;

                # for my $key(keys %h_indexes){
                #     for my $i (@{$h_indexes{$key}}){
                #         my $head = $a_nodeAttHeads[$i];
                #         my $noa  = $a_row[$i];

                #         if($head eq $key){
                #             $h_nodeAtts{$geneID}{$key}
                #         }
                #         else{
                #             $h_nodeAtts{$geneID}{$key}{$head} = $noa;
                #         }
                #     }
                # }

                if(defined $tfIndex){
                    if($a_row[$tfIndex] eq 'TF'){
                        $h_nodeAtts{$geneID}{'TF'} = 1;
                    }
                    else{
                        $h_nodeAtts{$geneID}{'TF'} = undef;
                    }
                }

                for my $i (@a_highlightIndexes){
                    my $head = $a_nodeAttHeads[$i];
                    my $noa  = $a_row[$i];

                    if($noa ne '-' and $noa ne '0'){
                        $h_nodeAtts{$geneID}{'Highlights'}{$head} = $noa;
                    }
                    else{
                        $h_nodeAtts{$geneID}{'Highlights'}{$head} = undef;
                    }
                }

                if(defined $studiesIndex){
                    for my $s(split(/;/, $a_row[$studiesIndex])){
                        my @a = split(/:/, $s);
                        my $study = $a[0];
                        for my $subnetID(split(/,/, $a[1])){
                            $h_nodeAtts{$geneID}{'Present_in'}{$study}{$subnetID} = 1;
                        }
                    }
                }

                for my $i (@a_corrIndexes){
                    my $head = $a_nodeAttHeads[$i];
                    my $noa  = $a_row[$i];

                    if($noa ne 'undef'){
                        $h_nodeAtts{$geneID}{'Correlations'}{$head} = $noa;
                    }
                    else{
                        $h_nodeAtts{$geneID}{'Correlations'}{$head} = undef;
                    }
                }

                if(defined $avCorrIndex){
                    if($a_row[$avCorrIndex] ne '-'){
                        $h_nodeAtts{$geneID}{'Average_correlation'} = $a_row[$avCorrIndex];
                    }
                    else{
                        $h_nodeAtts{$geneID}{'Average_correlation'} = undef;
                    }
                }

                $h_nodeAtts{$geneID}{'Gene_frequency'} = $a_row[$geneFreqIndex]
                    if(defined $geneFreqIndex);
                $h_nodeAtts{$geneID}{'Best_subnetwork_average_score'} = $a_row[$bestAvScoreIndex]
                    if(defined $bestAvScoreIndex);
                $h_nodeAtts{$geneID}{'Gene_degree'} = $a_row[$degree]
                    if(defined $degree);

                for my $i (@a_gisticScoreIndexes){
                    my $head = $a_nodeAttHeads[$i];
                    my $noa  = $a_row[$i];

                    if($noa ne 'undef'){
                        $h_nodeAtts{$geneID}{'GISTIC_scores'}{$head} = $noa;
                    }
                    else{
                        $h_nodeAtts{$geneID}{'GISTIC_scores'}{$head} = undef;
                    }
                }

                if(defined $avCNVScoreIndex){
                    if($a_row[$avCNVScoreIndex] ne '-'){
                        $h_nodeAtts{$geneID}{'Average_CNV_score'} = $a_row[$avCNVScoreIndex];
                    }
                    else{
                        $h_nodeAtts{$geneID}{'Average_CNV_score'} = undef;
                    }
                    if($a_row[$putDriverGeneIndex] eq 'PDG'){
                        $h_nodeAtts{$geneID}{'Putative_driver_gene'} = 1;
                    }
                    else{
                        $h_nodeAtts{$geneID}{'Putative_driver_gene'} = undef;
                    }
                }
            }
        }
        close NODEATTRIBUTES;
    }

    if(defined $f_nwa){
        #print "Read network attributes from $f_nwa...\n";
        open(NETWORKATTRIBUTES, $f_nwa)
          || die "Cannot open $f_nwa:\n$!";

        my $isHeader = 1;
        my (@a_nwAttHeads,      @a_scoreIndexes,
            @a_cnvScoreIndexes, $averageScoreIndex, $seedOccurIndex,
            @a_pValIndexes,     @a_CNVpValIndexes, $msbsIndex, $connIndex, $mssIndex);
        while (<NETWORKATTRIBUTES>) {
            next if($_ =~ m/^#/);
            $_ =~ s/[\r\n]+$//;
            my @a_row = split(/\t/, $_);

            if($isHeader){
                @a_nwAttHeads = @a_row;

                for (my $i = 1; $i < scalar(@a_nwAttHeads); $i++){
                    if($a_nwAttHeads[$i] =~ m/^score/){
                        push(@a_scoreIndexes, $i);
                        $a_nwAttHeads[$i] =~ s/^score-//;
                    }
                    elsif($a_nwAttHeads[$i] =~ m/^CNV_score/){
                        push(@a_cnvScoreIndexes, $i);
                        $a_nwAttHeads[$i] =~ s/^CNV_score-//;
                    }
                    elsif($a_nwAttHeads[$i] =~ m/Average_score/){
                        $averageScoreIndex = $i;
                    }
                    elsif($a_nwAttHeads[$i] =~ m/Seed_occurrence/){
                        $seedOccurIndex = $i;
                    }
                    elsif($a_nwAttHeads[$i] =~ m/^p_value/){
                        push(@a_pValIndexes, $i);
                        $a_nwAttHeads[$i] =~ s/^p_value-//;
                    }
                    elsif($a_nwAttHeads[$i] =~ m/^CNV_p_value/){
                        push(@a_CNVpValIndexes, $i);
                        $a_nwAttHeads[$i] =~ s/^CNV_p_value-//;
                    }
                    elsif($a_nwAttHeads[$i] =~ m/Merged_snw_best_score/){
                        $msbsIndex = $i;
                    }
                    elsif($a_nwAttHeads[$i] =~ m/Recurrent_nodes/){
                        $connIndex = $i;
                    }
                    elsif($a_nwAttHeads[$i] =~ m/Merged_snw_seeds/){
                        $mssIndex = $i;
                    }
                }
                $isHeader = 0;
            }
            else{
                my $subnetID = $a_row[0];
                next if(defined $ra_snws and !Utils::in_array($subnetID, $ra_snws));
                push(@{$h_subnets_nwa{$subnetID}{'Network_type'}}, 'int')
                    if(scalar(grep(/int/, $subnetID))) and !Utils::in_array('int', \@{$h_subnets_nwa{$subnetID}{'Network_type'}});
                push(@{$h_subnets_nwa{$subnetID}{'Network_type'}}, 'reg')
                    if(scalar(grep(/reg/, $subnetID))) and !Utils::in_array('reg', \@{$h_subnets_nwa{$subnetID}{'Network_type'}});

                for my $i (@a_scoreIndexes){
                    my $head = $a_nwAttHeads[$i];
                    my $nwa  = $a_row[$i];

                    if($nwa ne 'undef'){
                        $h_subnets_nwa{$subnetID}{'Scores'}{$head} = $nwa;
                    }
                    else{
                        $h_subnets_nwa{$subnetID}{'Scores'}{$head} = undef;
                    }
                }
                for my $i (@a_cnvScoreIndexes){
                    my $head = $a_nwAttHeads[$i];
                    my $nwa  = $a_row[$i];

                    if($nwa ne 'undef'){
                        $h_subnets_nwa{$subnetID}{'CNV_scores'}{$head} = $nwa;
                    }
                    else{
                        $h_subnets_nwa{$subnetID}{'CNV_scores'}{$head} = undef;
                    }
                }

                if(defined $averageScoreIndex){
                    if($a_row[$averageScoreIndex] ne "-"){
                        $h_subnets_nwa{$subnetID}{'Average_score'} = $a_row[$averageScoreIndex];
                    }
                    else{
                        $h_subnets_nwa{$subnetID}{'Average_score'} = undef;
                    }
                }
                if(defined $seedOccurIndex){
                    if($a_row[$seedOccurIndex] ne "-"){
                        $h_subnets_nwa{$subnetID}{'Seed_occurrence'} = $a_row[$seedOccurIndex];
                    }
                    else{
                        $h_subnets_nwa{$subnetID}{'Seed_occurrence'} = undef;
                    }
                }

                for my $i (@a_pValIndexes){
                    my $head = $a_nwAttHeads[$i];
                    my $nwa  = $a_row[$i];

                    if($nwa ne 'undef'){
                        $h_subnets_nwa{$subnetID}{'p-values'}{$head} = $nwa;
                    }
                    else{
                        $h_subnets_nwa{$subnetID}{'p-values'}{$head} = undef;
                    }
                }
                for my $i (@a_CNVpValIndexes){
                    my $head = $a_nwAttHeads[$i];
                    my $nwa  = $a_row[$i];

                    if($nwa ne 'undef'){
                        $h_subnets_nwa{$subnetID}{'CNV_p-values'}{$head} = $nwa;
                    }
                    else{
                        $h_subnets_nwa{$subnetID}{'CNV_p-values'}{$head} = undef;
                    }
                }

                if(defined $msbsIndex){
                    if($a_row[$msbsIndex] ne "-"){
                        $h_subnets_nwa{$subnetID}{'Merged_snw_best_score'} = $a_row[$msbsIndex];
                    }
                    else{
                        $h_subnets_nwa{$subnetID}{'Merged_snw_best_score'} = undef;
                    }
                }

                if(defined $connIndex){
                    if($a_row[$connIndex] ne "-"){
                        my @a_conn = split(/;/, $a_row[$connIndex]);
                        $h_subnets_nwa{$subnetID}{'Recurrent_nodes'} = \@a_conn;
                    }
                    else{
                        $h_subnets_nwa{$subnetID}{'Recurrent_nodes'} = undef;
                    }
                }

                if(defined $mssIndex){
                    if($a_row[$mssIndex] ne "-"){
                        my @a_seeds = split(/;/, $a_row[$mssIndex]);
                        $h_subnets_nwa{$subnetID}{'Merged_snw_seeds'} = \@a_seeds;
                    }
                    else{
                        $h_subnets_nwa{$subnetID}{'Merged_snw_seeds'} = undef;
                    }
                }
            }
        }
        close NETWORKATTRIBUTES;
    }

    if(defined $f_eda){
        #print "Read edge attributes from $f_eda...\n";
        open(EDGEATTRIBUTES, $f_eda)
          || die "Cannot open $f_eda:\n$!";

        my $edgeNb = scalar(keys %h_edgeAtts);

        my $isHeader = 1;
        my(@a_header, %h_atts);
        while (<EDGEATTRIBUTES>) {
            next if($_ =~ m/^#/);
            $_ =~ s/[\r\n]+$//;
            my @a_row = split(/\t/, $_);

            if($isHeader){
                @a_header = @a_row;
                $isHeader = 0;
                next;
            }
            for(my $i = 1; $i < scalar(@a_header); $i++){
                $h_atts{$a_header[$i]} = $a_row[$i];
            }

            my @a_int = split(/\ \(|\)\ /, $a_row[0]);
            my $node1 = $a_int[0];
            my $node2 = $a_int[2];
            my $dir   = $h_atts{'Direction'};
            my $edge  = "$node1-$node2";

            if($edgeNb){
                next if(!defined $h_edgeAtts{$edge} ||
                        ($dir eq '--' and !exists $h_edgeAtts{$edge}{'pp'}) ||
                        ((scalar(grep(/>/, $dir)) || scalar(grep(/</, $dir))) and
                         !exists $h_edgeAtts{$edge}{'pd'})
                       );
            }

            for my $eda(keys %h_atts){
                $h_edgeAtts{$edge}{$a_int[1]}{$eda} = $h_atts{$eda};
            }
        }
        close EDGEATTRIBUTES;
    }

    return \%h_subnets_nwa, \%h_nodeAtts,
           \%h_edgeAtts;
}

=pod

=head2 includeEdges()

Gets all interactions of subnetworks in the hash %$rh_snws, from the reference network %$rh_network.

=cut

sub includeEdges{
    my $rh_snws = shift;
    my $rh_network = shift;

    # Subnetworks hash:
    # \%rh_snws($subnetwork_ID => \%('Edges' => \@("$gene_A-$gene_B"),
    #                                'Nodes' => \%($gene_ID => 1)))
    # Reference network hash:
    # \%rh_network($gene_ID => \@neighbor_genes)

    #print "Find all links\n";
    my $count = 0;

    for my $subnetID(keys %$rh_snws){
        #if($count % 10 == 0) {
        #    print "$count done - ". scalar (keys  %$rh_snws) . "\n";
        #}
        #$count++;

        my $rh_snw = $$rh_snws{$subnetID};
        my %h_edges;
        for my $node(keys %{$$rh_snw{'Nodes'}}){
            for my $neighborNode (@{$$rh_network{$node}}) {
                next unless(defined $$rh_snw{'Nodes'}{$neighborNode});

                my $g1 = Utils::min($node, $neighborNode);
                my $g2 = Utils::max($node, $neighborNode);
                $h_edges{"$g1-$g2"} = 1;
            }
        }

        @{$$rh_snw{'Edges'}} = keys %h_edges;
    }
    #print "\n";
}

=pod

=head2 networkScore()

Computes and returns the subnetwork score.

=cut

sub networkScore {
    my $rh_datasets       = shift;
    my $rh_snwProfile     = shift;
    my $rh_snwProfileSize = shift;
    my $rh_snwGeneScores  = shift;
    my $scoreType         = shift;
    my $label             = shift;
    my $maxSampleNb       = shift;
    my $useAbsVal         = shift;
    my $useAnti           = shift;

    # \%rh_datasets($dataset_label => \%('cond' => \%($sample_ID => 0/1/undef),
    #                                    'expe' => \@sampleList,
    #                                    'MultiValues' => undef/1))
    # \%rh_snwProfile($dataset_label => \%($sample_ID => $expression))
    # \%rh_snwProfileSize($dataset_label => \%($sample_ID => $gene_nb))
    # \%rh_snwGeneScores($dataset_label => \%($gene_ID => $gene_score))

    if ($scoreType eq 'averageExpression') {
        my $score = 0;
        my $ra_samples = $$rh_datasets{$label}{'expe'};

        my (@a_dataVect);

        for (my $i = 0; $i < scalar(@$ra_samples); $i++) {
            my $sample = $$ra_samples[$i];

            if (defined $$rh_snwProfile{$label}{$sample}) {
                $a_dataVect[$i] =
                    $$rh_snwProfile{$label}{$sample} /
                    $$rh_snwProfileSize{$label}{$sample};
            }
            else {
                $a_dataVect[$i] = undef;
            }
        }

        # Case "Mono Condition"
        if(scalar keys %{$$rh_datasets{$label}{'cond'}}) {

            my $rh_conditions = $$rh_datasets{$label}{'cond'};
            my @a_condVect = ();
            if(defined $$rh_datasets{$label}{'MultiValues'}) {
                for my $sample (@$ra_samples) {
                    push(@a_condVect, $$rh_conditions{$sample});
                }

                $score  = Utils::multiconditionalDifferentialExpression(\@a_dataVect, \@a_condVect);
                $score *= sqrt(scalar @$ra_samples) / sqrt($maxSampleNb);
            }
            else{
                for my $sample (@$ra_samples) {push(@a_condVect, $$rh_conditions{$sample});}
                $score  = Utils::pearsonCorrelation(\@a_dataVect, \@a_condVect);
                $score *= sqrt(scalar @$ra_samples) / sqrt($maxSampleNb);
            }

            if (defined $useAbsVal) {
                $score = abs($score);
            }
            elsif (defined $useAnti) {
                $score = -$score;
            }

            return $score;
        }
        # Case "MultiCondition"
        else {
            my $rh_conditions = $$rh_datasets{$label}{'mcond'};
            my $rh_conditionsWeight = $$rh_datasets{$label}{'mcondweight'};
            foreach my $condType(keys %$rh_conditionsWeight) {
                my @a_condVect = ();
                for my $sample (@$ra_samples) {push(@a_condVect, $$rh_conditions{$sample}{$condType});}

                $score += $$rh_conditionsWeight{$condType} * abs(Utils::pearsonCorrelation(\@a_dataVect, \@a_condVect));
            }

            if (defined $useAbsVal) {
                $score = abs($score);
            }
            elsif (defined $useAnti) {
                $score = -$score;
            }
            else {
                $score = $score;
            }

            return $score;
        }
    }
    elsif ($scoreType eq 'averageCorrelation') {
        my @a_vect = ();
        foreach my $curNode (keys %{$$rh_snwGeneScores{$label}}) {
            push(@a_vect, $$rh_snwGeneScores{$label}{$curNode});
        }
        my($score, $std) = Utils::avstd(@a_vect);

        return $score;
    }
}

=pod

=head2 overlap()

Computes and returns the overlap rate of 2 subnetworks.

=cut

sub overlap {
    my ($rh_subnetworkA, $rh_subnetworkB) = (shift, shift);

    # \%rh_subnetwork($gene_ID => 1)

    my $overlap = 0;
    for my $key (keys %$rh_subnetworkB) {
        if (exists $$rh_subnetworkA{$key}) {$overlap++}
    }
    if (scalar(keys %$rh_subnetworkB) > 0) {
        return ($overlap / scalar(keys %$rh_subnetworkB));
    }
    else {return 0;}
}


1;
