#!/usr/bin/perl
#
# ITI project
#
# Copyright Maxime Garcia, Ghislain Bidaut, Claire Rioualen, Quentin Da Costa
# (2009-2017)
#
# maxime.u.garcia@gmail.com
# ghislain.bidaut@inserm.fr
# claire.rioualen@inserm.fr
# quentin.da-costa@inserm.fr
#
# http://iti.sourceforge.net
#
# This software is a computer program whose purpose is to superimpose
# interactome over several gene expression datasets to find
# discriminative subnetworks predicting metastatic relapse in breast
# cancer.
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty and the software's author, the holder of the
# economic rights, and the successive licensors have only limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading, using, modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean that it is complicated to manipulate, and that also
# therefore means that it is reserved for developers and experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and, more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package Utils;

use strict;
use warnings;
use IO::Uncompress::Gunzip qw(gunzip $GunzipError);
use Data::Dumper;

# return max of a list of numbers
sub max {
    my @a_number = @_;
    my $maximum  = $a_number[0];
    foreach my $number (@a_number) {$maximum = $number if $number > $maximum;}
    return $maximum;
}


# return min of a list of numbers
sub min {
    my @a_number = @_;
    my $minimum  = $a_number[0];
    foreach my $number (@a_number) {$minimum = $number if $number < $minimum;}
    return $minimum;
}

# return average and standard dev of a vector
sub avstd {
    my @vect = @_;
    
    if (scalar @vect == 1) {
        return ($vect[0], 0);
    }
    if (scalar @vect == 0) {
        return (0, 0);
    }

    my $av  = 0;
    my $std = 0;

    foreach (@vect) {
        $av += $_;
    }
    $av /= scalar @vect;
    foreach (@vect) {
        $std += ($_ - $av) * ($_ - $av);
    }
    $std = sqrt($std / ((scalar @vect) - 1));
    return ($av, $std);
}


# Compute Median Value of a vector
sub median {
    my @vect = @_;

    if (scalar @vect == 1) {return $vect[0];}
    if (scalar @vect == 0) {return 0;}

    my @sorted = sort @vect;
    my $med;

    if ((scalar @sorted) % 2 == 0) {
        $med =
          ($sorted[scalar(@sorted) / 2] + $sorted[scalar(@sorted) / 2 - 1]) / 2;
    }
    else {
        $med = $sorted[scalar(@sorted) / 2];
    }

    return $med;
}


# Computes personcorrelation betwen two vectors.  Empty values are allowed. If two empty vector are submitted, correlation is set to 0.
sub pearsonCorrelation {
    my @vect1 = @{$_[0]};
    my @vect2 = @{$_[1]};

    my @index = ();
    for (my $i = 0; $i < scalar @vect1; $i++) {
        if (defined $vect1[$i] && defined $vect2[$i]) {
            push(@index, $i);
        }
    }

    my $n = scalar @index;
    if ($n == 0) {return 0;}

    my $sumxy = 0;
    for my $i (@index) {
        $sumxy += $vect1[$i] * $vect2[$i];
    }

    my $sumx = 0;
    my $sumy = 0;

    my $sumx2 = 0;
    my $sumy2 = 0;

    for my $i (@index) {

        $sumx  += $vect1[$i];
        $sumy  += $vect2[$i];
        $sumx2 += $vect1[$i] * $vect1[$i];
        $sumy2 += $vect2[$i] * $vect2[$i];
    }

    my $p =  ($sumx2 - (($sumx * $sumx) / $n)) * ($sumy2 - (($sumy * $sumy) / $n));
    my $den;

    if   ($p < 0) {$den = 0;}
    else          {$den = sqrt($p);}

    if ($den != 0) {
        my $c = ($sumxy - ($sumx * $sumy) / $n) / $den;
        return $c;
    }
    else {
        return 0;
    }
}


# Computes multiconditional differential expression measurement, as defined in Tusher et al. PNAS (2001)
# 1st arg: the expression profile
# 2nd arg: the group
# Returns: the SAM score

sub multiconditionalDifferentialExpression {
    my @profile    = @{$_[0]};
    my @conditions = @{$_[1]};

    my @index = ();
    for (my $i = 0; $i < scalar @profile; $i++) {
        if (defined $profile[$i] && defined $conditions[$i]) {
            push(@index, $i);
        }
    }

    my $nconds = scalar @index;

    my $d;
    my $r;
    my $s;
    my $av = 0;

    my %h_c;

    # Compute the Nk
    foreach my $i(@index) {
        $h_c{$conditions[$i]}++;
    }

    # Compute the average of the profile
    foreach (@index) { $av += $profile[$_]; }
    $av /= $nconds;

    # Initialize
    my $sum_nk         = 0; # Sum(Nk)
    my $pi_nk          = 1; # Pi(Nk)
    my $sum_diff       = 0; # Sum(Nk(Xk-X)2)
    my $sum_invert_nk  = 0; # Sum (1/Nk)
    my $sum_nk_minus1  = 0; # Sum (Nk-1)
    my $sum_diff_av_xj = 0; # Sum (xj-Xk)2
    my %av_x;               # hash containing av(Xk) indexed by k
    my %diff_av_xj;         # hash containing diff(av(j)-av(Xk) indexed by k

    # Loop across conditions
    foreach my $cd (sort keys %h_c) {

        # print "current cond = $cd  ncd = $h_c{$cd}\n";

        # Some easily computed coefficients
        $sum_nk += $h_c{$cd};
        $pi_nk  *= $h_c{$cd};

        $sum_invert_nk += 1 / $h_c{$cd};
        $sum_nk_minus1 += $h_c{$cd} - 1;

        # Calc (Xk-X)
        $av_x{$cd} = 0;
        for my $i(@index) {
            if($conditions[$i] eq $cd) { $av_x{$cd} += $profile[$i]; }
        }
        $av_x{$cd} /= $h_c{$cd};

        $sum_diff += $h_c{$cd} * ($av_x{$cd} - $av)**2;

        # Calc Sum (Xj-av(Xk))
        $diff_av_xj{$cd} = 0;
        for my $i(@index) {
            if($conditions[$i] eq $cd) {
                $diff_av_xj{$cd} += ($profile[$i]-$av_x{$cd})**2;
            }
        }

        $sum_diff_av_xj += $diff_av_xj{$cd};
    }

    # Final computation
    $r = sqrt(($sum_nk/$pi_nk) * $sum_diff);
    $s = sqrt(($sum_invert_nk/$sum_nk_minus1) * $sum_diff_av_xj);

    my $s0 = 3.3; # This should be adapted to the data

    $d = $r/($s+$s0); # The score

    # Adding the score sign
    if(scalar(keys %av_x) == 2){
        my @a_cd = sort keys %av_x;
        $d = -$d if($av_x{$a_cd[1]} < $av_x{$a_cd[0]});
    }

    return $d;
}


# computes factorial
sub fact {
    my $n = shift;

    if   ($n == 0) {return 1;}
    else           {return ($n * fact($n - 1))}
}

# Returns a list of file names contained in a given directory
sub readInputDir {
    my $inputDirectory = shift;
    #print "Input directory = $inputDirectory\n";

    opendir(DIRECTORY, "$inputDirectory") || die "Cannot open $inputDirectory : $!";

    my @a_file;
    while (my $file = readdir(DIRECTORY)) {
        if (!($file eq '.') && !($file eq '..')) {
            push(@a_file, $file);
        }

    }
    close DIRECTORY;
    return @a_file;
}

# test if value is in array
sub in_array {
    my ($element, $ra_array) = (shift, shift);
    return (scalar(grep({"$_" eq "$element"} @$ra_array)) != 0) ? 1 : 0;
}

# takes an array and return an array without duplicates
sub unique {
    my %seen = ();
    my @r = ();
    foreach my $a (@_) {
        unless ($seen{$a}) {
            push @r, $a;
            $seen{$a} = 1;
        }
    }
    return \@r;
}

# reads an input file (here a threshold file) and returns a hash
sub readInputLight {
    my $file = shift;
    open(FILE, "$file") || die "Cannot open $file : $!";

    my %h_hash;
    my @a_tmp;
    while (<FILE>) {
        next if ($_ =~ m/^#/);
        $_ =~ s/[\r\n]+$//;
        @a_tmp = split(/\t/, $_);
        $h_hash{$a_tmp[0]} = $a_tmp[1];
    }
    close(FILE);
    return %h_hash;
}

# returns a double hash from reading a file
sub readInputDataset {
    my $fileName = shift;

    print "Reading compressed file\n";

    unless(-e $fileName) { die "$fileName does not exist\n"; }

    my $z = new IO::Uncompress::Gunzip $fileName
      || die "IO::Uncompress::Gunzip failed: $GunzipError\n";

    my @data;
    while (<$z>) {
        next if ($_ =~ m/^#/);
        $_ =~ s/[\r\n]+$//;
        push(@data, $_);
    }
    close($z);

    my %data;
    my @expeList;
    my @geneList;

    my @tmp = split(/\t/, $data[0]);

    for (my $i = 1; $i < scalar @tmp; $i++) {
        push(@expeList, $tmp[$i]);
    }

    print "Number of experiments: " . scalar @expeList . "\n";
    
    for (my $i = 1; $i < scalar @data; $i++) {

        my @row = split(/\t/, $data[$i]);

        my $geneID = $row[0];

        push(@geneList, $geneID);

        for (my $j = 1; $j < scalar @row; $j++) {
            if ($row[$j] eq '' || $row[$j] !~ m/^NaN$/) {
                $data{$geneID}{$expeList[$j - 1]} = $row[$j];
            }
            else {
                $data{$geneID}{$expeList[$j - 1]} = undef;
            }
        }
    }
    
    print "Number of gene expression profiles populated: "
      . (scalar keys %data) . "\n";
    return \%data, \@expeList, \@geneList;
}


=pod

=head2 readHTSData()

Returns a double hash from reading a HTS z-score file.
$sample is a unique fiels with the name of the header
Returns:
\%h_data($geneID => \%($sample => $expression));
\@a_sampleList.

=cut

sub readHTSData {
    my $fileName   = shift;
    my $rh_conditions = shift;

    print "Reading z-score file\n";

    my @a_data;

    open(FILE, "$fileName") || die "Cannot open $fileName:$!";
    
    my %h_data;
    my @a_expeList;
    my @expeIndex;
    my $i=0;
    
    # Read header and test number of conditons
    
    my $header = <FILE>;
    $header =~ s/[\r\n]+$//;
    
    my @tmp = split(/\t/, $header);

    if (scalar @tmp != 2) {
	die "File $fileName contains an extra column in header. It must be tab delimited with exactly 2 colums.\n";
    }

    push(@a_expeList,  $tmp[1]);
    
    die "No experiments found\n" unless(scalar @a_expeList);
    
    while(<FILE>) {
	$i++;
	next if $_ =~ m/^#/;
	
	$_ =~ s/[\r\n]+$//;
	my @row = split(/\t/, $_);

	if (scalar @row != 2) {
	    die "File $fileName contains an extra column row $i. It must be tab delimited with exactly 2 colums.\n";
	}
	
        my $geneID = $row[0];
        my $zscore = $row[1];
	
	if($zscore !~ m/^\s?(NaN|NA)?\s?$/) {
	    
	    if($zscore !~ /^[0-9\.eE\-\+]+$/) {
		die "File $fileName contains malformed float value: $zscore at row $i.\n";
	    }
	    
	    $h_data{$geneID}{$a_expeList[0]} = $zscore;
	}
	else {
	    $h_data{$geneID}{$a_expeList[0]} = undef;
        }
    }

    close(FILE);
    
    print "Number of z-score found = " . scalar(keys %h_data) . "\n";
    die "No usable Data found in input file $fileName.\n" if( (scalar(keys %h_data)) == 0);
    return \%h_data, \@a_expeList;
}



# normal inner product including normalisation
sub innerProduct {
    my @vect1 = @{$_[0]};
    my @vect2 = @{$_[1]};

    my ($av1, $std1) = avstd(@vect1);
    my ($av2, $std2) = avstd(@vect2);

    my $res = 0;
    for (my $j = 0; $j < scalar @vect1; $j++) {

        print "$j, $vect1[$j] "
          . ($vect1[$j] / $av1)
          . "   $vect2[$j] "
          . $vect2[$j] / $av2 . "\n";

        $res += ($vect1[$j] / $av1) * ($vect2[$j] / $av2);
    }
    return $res;
}


=pod

=head2 hypergeometricTest()

This has been copied from the catscore package.
For the implementation in PERL, see:
Kevin J. Cheung, Vasudeo Badarinarayana, Douglas Selinger, Daniel Janse, and George M. Church
(2003) "A microarray-based antibiotic screen identifies a regulatory role for supercoiling in the osmotic stress response of Escherichia coli," Genome Res. 2003-2005 Feb;13(2):206-15.
Files of CATSCORE are at http://arep.med.harvard.edu/supercoiling/supplement.htm.

Following tavazoie et al:
k = number of gene in the cluster involved in the function of interest;
f = total annotated gene number involved in the function of interest;
n = number of gene in the cluster;
g = total annotated gene number.

=cut

sub hypergeometricTest {

    my $k = $_[0];
    my $f = $_[1];
    my $n = $_[2];
    my $g = $_[3];

    my $cump;
    my $logterm;
    my $i;

    # the maximum worst case distribution is if
    # all of cluster (n) is your functional category
    # OR if all of functional category is in your cluster.

    my $max_allowed = $n;
    if ($n > $f) {$max_allowed = $f;}

    $i = $k;
    $logterm =
      log_n_choose_x($f,      $i) +
      log_n_choose_x($g - $f, $n - $i) -
      log_n_choose_x($g,      $n);

    $cump = exp($logterm);

    for ($i = $k + 1; $i <= $max_allowed; $i++) {
        $logterm =
          $logterm +
          log($f - $i + 1) -
          log($i) -
          log($g - $f - ($n - $i + 1) + 1) +
          log(($n - $i + 1));

        $cump = $cump + exp($logterm);
    }

    return $cump;
}

=pod

=head2 log_n_choose_x()

Calculates (n choose x).

=cut

sub log_n_choose_x {
    my $a = $_[0];
    my $b = $_[1];
    my ($c0, $c1);
    my $i;

    if ($b == 0) {0;}    ## only 1 way to choose 0 things

    $c0 = 0;
    for ($i = ($a - $b + 1); $i <= $a; $i++) {
        $c0 = $c0 + log($i);
    }

    $c1 = 0;
    for ($i = 1; $i <= $b; $i++) {
        $c1 = $c1 + log($i);
    }

    return ($c0 - $c1);
}

=pod

=head2 readGene2Go()

Parses gene2go file, selects and counts GOs of the taxon and populates correspondance tables.

=head3 Prerequisite
Download gene2go file from NCBI FTP site.

=head3 Arguments
$gene2GO_fileName, $taxonID.

=head3 Returns
=over

=item \%h_gene2GOs
Reference of hash table indexed by gene ID and containing the list of corresponding GOs; each go is described by a string of the form: "$GOID - $go_term";

=back

=cut

sub readGene2Go {
    my $gene2GOFile   = shift;
    my $soughtTaxonID = shift;

    print "\nReading gene_to_GO file and counting GOs...\n";

    unless(-e $gene2GOFile) { die "$gene2GOFile does not exist\n"; }
    
    my $z = new IO::Uncompress::Gunzip $gene2GOFile
      || die "IO::Uncompress::Gunzip failed:\n$GunzipError\n";
    my (%h_gene2GOs, %h_GOs);
    my $i = 0;
    while (<$z>) {
        next if ($_ =~ m/^#/);
        $_ =~ s/[\r\n]+$//;
        my @a_row = split(/\t/, $_);

        $i++;
        my $taxonID = $a_row[0];
        if ($taxonID eq $soughtTaxonID) {
            die "Gene ID not defined (line $i) - file corrupted" if (!defined $a_row[1]);
            my $geneID = $a_row[1];

            die "GO ID for $geneID not defined (line $i) - file corrupted" if (!defined $a_row[2]);
            my $goID = $a_row[2];

            my $go_term = "-";
            if (defined $a_row[5] and $a_row[5] ne "") {
                $go_term = $a_row[5];
            }
            else {
                warn "Warning: description for $goID not defined (line $i)\n";
            }

            my $GO = "$goID - $go_term";
            $h_gene2GOs{$geneID}{$GO} = 1;
            $h_GOs{$GO}               = 1;
        }
    }

    close($z);
    print scalar(keys %h_gene2GOs) . " genes and " .
          scalar(keys %h_GOs) . " GOs found in gene2go file\n";

    return \%h_gene2GOs;
}


# read gene_info file
sub readGeneInfo {
    my $fileName = shift;
    my $taxID    = shift;

    my %GIDSymbolHash;
    my %SymGeneIDHash;

    print "Reading compressed gene_info file\n";

    unless(-e $fileName) { die "$fileName does not exists\n"; }

    my $z = new IO::Uncompress::Gunzip $fileName || die "IO::Uncompress::Gunzip failed: $GunzipError\n";
    my $i = 0;
    while (<$z>) {
        $i++;
        next if ($_ =~ m/^#/);
        $_ =~ s/[\r\n]+$//;
        my @a_row = split(/\t/, $_);

        if (!defined $a_row[1]) {
            die "Gene id not defined (line $i): $!";
        }
        next if(defined $taxID and $a_row[0] ne $taxID);

        my $gene_id = $a_row[1];
        my $symbol  = $a_row[2];
        $GIDSymbolHash{$gene_id} = $symbol;
        $SymGeneIDHash{$symbol}  = $gene_id;
    }
    close($z);

    return \%GIDSymbolHash, \%SymGeneIDHash;
}


# read gene2ensembl file
sub readGene2EnsemblGene {
    my $fileName = shift;
    my $taxID    = shift;

    my %h_entrez2ensembl;
    my %h_ensembl2entrez;

    print "Reading compressed gene2ensembl file\n";

    unless(-e $fileName) { die "$fileName does not exists\n"; }

    my $FILE = new IO::Uncompress::Gunzip $fileName || die "IO::Uncompress::Gunzip failed: $GunzipError\n";
    my $i = 0;
    while (<$FILE>) {
        $i++;
        next if ($_ =~ m/^#/);
        $_ =~ s/[\r\n]+$//;
        my @a_row = split(/\t/, $_);

        if (!defined $a_row[1] || !defined $a_row[2]) {
            die "Gene ID missing (line $i)\n";
        }
        next if(defined $taxID and $a_row[0] ne $taxID);

        my $entrezID   = $a_row[1];
        my $ensemblID  = $a_row[2];
        $h_entrez2ensembl{$entrezID}   = $ensemblID;
        $h_ensembl2entrez{$ensemblID}  = $entrezID;
    }
    close($FILE);

    return \%h_entrez2ensembl, \%h_ensembl2entrez;
}


# read gene2ensembl file
sub readGene2EnsemblProt {
    my $fileName = shift;
    my $taxID    = shift;

    my %h_entrez2ensembl;
    my %h_ensembl2entrez;

    print "Reading compressed gene2ensembl file\n";

    unless(-e $fileName) { die "$fileName does not exists\n"; }

    my $FILE = new IO::Uncompress::Gunzip $fileName || die "IO::Uncompress::Gunzip failed: $GunzipError\n";
    my $i = 0;
    while (<$FILE>) {
        $i++;
        next if ($_ =~ m/^#/);
        $_ =~ s/[\r\n]+$//;
        my @a_row = split(/\t/, $_);

        if (!defined $a_row[1] || !defined $a_row[2]) {
            die "Gene ID missing (line $i)\n";
        }
        next if(defined $taxID and $a_row[0] ne $taxID);

        my $entrezID   = $a_row[1];
        my $ensemblID  = $a_row[6];
        $h_entrez2ensembl{$entrezID}   = $ensemblID;
        $h_ensembl2entrez{$ensemblID}  = $entrezID;
    }
    close($FILE);

    return \%h_entrez2ensembl, \%h_ensembl2entrez;
}


# read PPI HPRD file / HPRD ID to Gene Symbol
sub readHPRDtoGsym {
    my $fileName = shift;

    my %h_HPRDtoSym;

    print "Reading binary_protein_protein_interactions file\n";

    unless(-e $fileName) { die "$fileName does not exists\n"; }

    my $z = new IO::Uncompress::Gunzip $fileName
      || die "IO::Uncompress::Gunzip failed: $GunzipError\n";


    my @data = <$z>;
    while (<$z>) {
        $_ =~ s/[\r\n]+$//;
        push(@data, $_);
    }
    close($z);

    for (my $i = 1; $i < scalar @data; $i++) {

        my @row = split(/\t/, $data[$i]);

        my $Sym1  = $row[0];
        my $HPRD1 = $row[1];

        my $Sym2  = $row[3];
        my $HPRD2 = $row[4];

        $h_HPRDtoSym{$HPRD1} = $Sym1;
        $h_HPRDtoSym{$HPRD2} = $Sym2;
    }

    return \%h_HPRDtoSym;
}

# read gene_info file / Ensembl ID to gene id
sub readEnsemblInfo {
    my $fileName = shift;

    my %h_EnsemblToGeneID;

    unless(-e $fileName) { die "$fileName does not exists\n"; }

    print "Reading compressed gene_info file\n";
    my $z = new IO::Uncompress::Gunzip $fileName
      || die "IO::Uncompress::Gunzip failed: $GunzipError\n";


    my @data = <$z>;
    while (<$z>) {
        $_ =~ s/[\r\n]+$//;
        push(@data, $_);
    }
    close($z);

    for (my $i = 1; $i < scalar @data; $i++) {

        my @row = split(/\t/, $data[$i]);

        my $taxon_id = $row[0];
        if ($taxon_id eq "9606" or $taxon_id eq "10090" or $taxon_id eq "10116")
        {    #keep only mouse, human and rat

            if (!defined $row[1]) {
                die "gene id not defined (line $i) : $!";
            }
            my $gene_id     = $row[1];
            my $gene_symbol = $row[2];
            my $other_id    = $row[5];    # dbXrefs field
            my $ensembl_id  = '';

            my @a_other_id = split('\|', $other_id);    # dbXrefs field parsing
            my @grep_ensembl = grep(/^Ensembl/, @a_other_id);

            if (scalar @grep_ensembl == 0) {            ## no ensembl ID found
                ;
            }
            elsif (scalar @grep_ensembl == 1) {         ## 1 ensembl ID found
                my @split = split(':', $grep_ensembl[0]);
                $ensembl_id = $split[1]
                  ;    # if ensembl ID already defined keep shortest gene symbol
                if (  !defined($h_EnsemblToGeneID{$ensembl_id})
                    || length($gene_symbol) <
                    length($h_EnsemblToGeneID{$ensembl_id})) {
                    $h_EnsemblToGeneID{$ensembl_id} = $gene_id;
                }
            }
            elsif (scalar @grep_ensembl >= 2) {    ## several ensembl ID found
                foreach (@grep_ensembl) {
                    my @split = split(':', $_);
                    $ensembl_id = $split[1]
                      ; # if ensembl ID already defined keep shortest gene symbol
                    if (  !defined($h_EnsemblToGeneID{$ensembl_id})
                        || length($gene_symbol) <
                        length($h_EnsemblToGeneID{$ensembl_id})) {
                        $h_EnsemblToGeneID{$ensembl_id} = $gene_id;
                    }
                }
            }
        }
    }
    return \%h_EnsemblToGeneID;
}


# Get file name from a fully qualified name
sub getName {
    my $file   = shift;
    my @a_temp = split(/\//, $file);
    my $name   = $a_temp[-1];
    @a_temp = split(/\./, $name);
    $name = $a_temp[0];
    return $name;
}


# generate random number from a normal distribution
sub approximateNormalDistribution {
    my ($mean, $std, $acc) = (shift, shift, shift);
    my $sum = 0;
    for (my $i = 0; $i < $acc; $i++) {
        $sum += rand();
    }
    $sum -= $acc / 2.0;
    return (($sum / sqrt($acc / 2.0)) * $std + $mean);
}


# Read tab delimited inout file and populates a hash table
sub readHashFile {
    my $f_input = shift;
    my %h_output;

    open(FILE, "<$f_input") || die "Cannot open $f_input: $!";
    
    while(defined(my $l = <FILE>)){
        next if ($l =~ m/^#/);
        $l =~ s/[\r\n]+$//;
        my @row = split('\t', $l);
        push(@{$h_output{$row[0]}}, $row[1]);
    }
    close(FILE);

    return(\%h_output);
}


# Print HTS Net Logo
sub printHTSNETLogo {
    print '

 #     # #######  #####        #     #              
 #     #    #    #     #       ##    # ###### ##### 
 #     #    #    #             # #   # #        #   
 #######    #     #####  ##### #  #  # #####    #   
 #     #    #          #       #   # # #        #   Detection of
 #     #    #    #     #       #    ## #        #   Deregulated
 #     #    #     #####        #     # ######   #   Gene Networks

Claire Rioualen (Developer)
Quentin Da Costa (Developer)
Ghislain Bidaut (Senior developer)
Maxime Garcia (Former developer)
Licence : CeCILL (http://www.cecill.info)
Web: http://htsnet.marseille.inserm.fr/


';
}

1;
