#!/usr/bin/perl
#
# HTS-Net project
#
# Copyright Maxime Garcia, Ghislain Bidaut, Claire Rioualen, Quentin Da Costa
# (2009-2017)
#
# maxime.u.garcia@gmail.com
# ghislain.bidaut@inserm.fr
# claire.rioualen@inserm.fr
# quentin.da-costa@inserm.fr
#
# http://iti.sourceforge.net
#
# This software is a computer program whose purpose is to superimpose
# interactome over several gene expression datasets to find
# discriminative subnetworks predicting metastatic relapse in breast
# cancer.
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty and the software's author, the holder of the
# economic rights, and the successive licensors have only limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading, using, modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean that it is complicated to manipulate, and that also
# therefore means that it is reserved for developers and experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and, more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod

=head1 Name
detectSubnetworks.pl

=head1 Synopsis

=head2 Multiple output directories
detectSubnetworks.pl
        -id | --inputDir     <Input directory>
        -od | --outputDir    <General output directory>

=head2 Output files in current directory
detectSubnetworks.pl
        -so | --singleOutput

        -ef | --expFile  <Expression file>

=head2 Required
        -nwf  | --networkFile     <Interaction network file>
        -edaf | --edgeAttFile     <Edge attribute file>
        -gif  | --geneInfoFile    <gene_info file>
        -gsf  | --gene2RefSeqFile <gene2refseq file>
    
        -th | -pcth  <Pearson correlation threshold>
        -r  | --rate <Score improvement threshold>

=head2 Optional
        -cd | --conditionDir <Condition directory> (With multiple output directories only)

        -slf | --seedListFile <List of subnetwork seeds (gene IDs)>

        -nwt | --networkType <Network type>
        -st  | --scoreType   <Subnetwork score type>
        -d   | --depth       <Maximal depth of subnetwork>
        -c   | --consensus   <Consensus number of datasets>
        -nrand               <Number of random subnetworks to generate>
        -start               <Start index in interactome>
        -stop                <Stop index in interactome>
        -index               <Script index>
        -s   | --suffix      <Suffix string>

        -abs
OR      -anti
        -sam

        -h | --help | -u | --usage
        -debug

=head1 Examples
perl detectSubnetworks.pl -id ~/workspace/data/test_dataset/ -od ~/workspace/data/iti_test/test/ -nwf ~/workspace/data/network_data-test/network.txt -edaf ~/workspace/data/network_data-test/edge_attributes.txt -gif /home/data/annotations/stable/ncbi/Homo_sapiens.gene_info.gz  -r 0.02 -th 0.15 -c 1 -nrand 100 -abs -sc averageCorrelation
perl detectSubnetworks.pl -singleOutput  -ef ~/workspace/data/test_dataset/IPC.txt.gz -p GPL6244 -nwf ~/workspace/data/network_data-test/network.txt -edaf ~/workspace/data/network_data-test/edge_attributes.txt -gif /home/data/annotations/stable/ncbi/Homo_sapiens.gene_info.gz -stop 1000 -rate 0.02 -th 0.2 -d 15 -c 1 -nrand 300 -abs -sc averageCorrelation

=head1 Description
Detects discriminative regions in interaction network by superimposition of Interactome/Regulome and HTS


=head1 Subroutines
=over

=item detectDiscriminativeSubnetworks()
=item detectSubnets()
=item printSubnets()
=item exploreSeed()
=item testIfImprovementInHTSScore()
=item addNodeToSubnetwork()
=item addSubnetInHash()

=back

=head1 Authors
Ghislain Bidaut     ghislain.bidaut@inserm.fr
Quentin Da Costa    quentin.da-costa@inserm.fr
Maxime Garcia       maxime.u.garcia@gmail.com
Claire Rioualen     claire.rioualen@inserm.fr

=cut


## Modules ##

use strict;
use warnings;

use Getopt::Long;
use Sys::Hostname;

use Utils;
use HTS;
use Subnetworks;
use Network;


## Arguments management ##

my ($help,             $debug,             $d_output,
    $d_inputData,
    $networkType,      $f_network,
    $f_totalEda,       $f_geneInfo,        $f_gene2RefSeq,
    $fileIndex,
    $pearsonCorrTh,    $improvRate,
    $stop,
    $useAbsVal,        $useAnti,           $outInCurrDir,
    $f_exp,             
    $f_seedList,       $useSAMScore);

my $snwDepth    = 10;
my $nrand       = 200;
my $start       = 0;
my $datasetCons = 1;
my $scoreType   = 'averageCorrelation';

GetOptions(
    'help|usage|h|u' => \$help,
    'debug'          => \$debug,

    'outputDir|od=s' => \$d_output,

    'inputDataDir|id=s'     => \$d_inputData,
    
    'networkFile|nwf=s'     => \$f_network,
    'edgeAttFile|edaf=s'    => \$f_totalEda,
    'seedListFile|slf=s'    => \$f_seedList,
    'geneInfoFile|gif=s'    => \$f_geneInfo,
    
    'networkType|nwt=s'     => \$networkType,
    'scoreType|st=s'        => \$scoreType,

    'index=i' => \$fileIndex,

    'pcth|th=f'     => \$pearsonCorrTh,
    'rate|r=f'      => \$improvRate,
    'depth=i'       => \$snwDepth,
    'consensus|c=i' => \$datasetCons,
    
    'start=i' => \$start,
    'stop=i'  => \$stop,
    'nrand=i' => \$nrand,

    'abs'  => \$useAbsVal,
    'anti' => \$useAnti,
    'sam'  => \$useSAMScore,

    'singleOutput|so' => \$outInCurrDir,
    'iFile|if=s'    => \$f_exp,
    
);

if(defined $d_inputData){
    if($d_inputData   !~ m/\/$/)    {$d_inputData   .= "/";}
}
else{
    $d_inputData = "" if(defined $f_exp);
}

my $fiTag;
if (defined $fileIndex) {
    srand($fileIndex);
    $fiTag = "$fileIndex-";
}
else {
    $fileIndex = "";
    $fiTag     = "";
}

if(defined $debug) {$debug = 1;}
else               {$debug = 0;}

my $nwtTag = "";
$nwtTag = "$networkType-" if(defined $networkType);

# Time measurement system - START  #TMS
my $host = hostname;
my $date = localtime time;
my $time = time;

# Mobyle - specific code
if (defined $outInCurrDir    and
    defined $f_exp           and
    defined $f_network       and
    defined $f_totalEda      and
    defined $f_geneInfo      and
    defined $improvRate      and
    defined $pearsonCorrTh){

    $d_output = "";

    Utils::printHTSNETLogo();
    print "Mobyle Code - Process started the $date on $host
$0:

    Input Data file                    = $f_exp
   
    Network file                       = $f_network
    Edge attribute file                = $f_totalEda
    gene_info file                     = $f_geneInfo
    
    Subetwork score type               = $scoreType
    Score improvement rate             = $improvRate
    Pearson correlation threshold      = $pearsonCorrTh
    Consensus number of datasets       = $datasetCons
    Subetwork depth                    = $snwDepth
    Number of random subnetworks       = $nrand
    Start index in interactome         = $start
";

    if(defined $stop) {
        print "    Stop index in interactome = $stop\n";
    }

    if($debug) {
        print "    DEBUG option activated\n";
    }
}

# Generic code (i.e. non-Mobyle)
elsif (defined $d_inputData     and
       defined $d_output        and
       defined $f_network       and
       defined $f_totalEda      and
       defined $f_geneInfo      and
       defined $improvRate      and
       defined $pearsonCorrTh){

    if($d_output          !~ m/\/$/)    {$d_output          .= "/";}
    unless (-d $d_output)               {mkdir "$d_output";}
    

    Utils::printHTSNETLogo();
    print "Generic Code - Process started the $date
$0:
";

if(defined $f_exp) {
    print "Input File:                 $f_exp\n";
}
else {
    print  "Input Directory:           $d_inputData\n";
}

print "

    Output directory                   = $d_output

    Network file                       = $f_network
    Edge attribute file                = $f_totalEda
    gene_info file                     = $f_geneInfo

    Subetwork score type               = $scoreType
    Score improvement rate             = $improvRate
    Pearson correlation threshold      = $pearsonCorrTh
    Consensus number of datasets       = $datasetCons
    Subetwork depth                    = $snwDepth
    Number of random subnetworks       = $nrand
    Start index in interactome         = $start
";

    if(defined $stop) {
        print "    Stop index in interactome          = $stop\n";
    }

    if($debug) {
        print "    DEBUG option activated\n";
    }
}

else{
    Utils::printHTSNETLogo();
    print "
NAME
detectSubnetworks.pl

DESCRIPTION
Detects discriminative subnetworks by superimposition of Interactome/Regulome and HTS data.

PARAMETERS

Multiple output directories
-id | --inputDir <Transcriptome directory>
-od | --outputDir <General output directory>

Output files in current directory
-so | --singleOutput
-if | --inputFile <Input Data file>

Required
-nwf | --networkFile <Interaction network file>\n\tInteractome and/or regulome
-edaf | --edgeAttFile <Edge attribute file>
-gif | --geneInfoFile <gene_info file>\n\tFull path of local copy of gene_info file gzip-compressed (.gz) downloaded from NCBI
-th | -pcth <Pearson correlation threshold>
-r | --rate <Score improvement threshold>

Optional

-slf | --seedListFile <Subnetwork seed file>\n\tA list of gene IDs to use as seed of subnetworks (in the first column)
-nwt | --networkType <Network type>\n\t\"int\" or \"reg\" which will appear in the subnetwork ID
-st | --scoreType <Subnetwork score type>\n\t\"averageCorrelation\" (DEFAULT) or \"averageScore\"
-d | --depth <Depth of subnetworks>\n\t(DEFAULT = 10)
-c | --consensus <Consensus number of datasets>\n\t(DEFAULT = 1)
-nrand <Number of random subnetworks to generate>\n\t(DEFAULT = 200)
-start <Start index in network>\n\t(DEFAULT = 0)
-stop <Stop index in network>
-index <Script index>
-s | --suffix <Suffix string>\n\tWrite the given string in created files so that you can identify from which run they come from
-abs\n\tUse absolute value for subnetwork score (DEFAULT = FALSE)
-anti\n\tUse anticorrelation value for subnetwork score (DEFAULT = FALSE)
-sam\n\tUse SAM score instead of Pearson correlation (DEFAULT = FALSE)

-h | --help | -u | --usage\n\tDisplay this user guide
-debug\n\tPrint additional informations

EXAMPLES
$0 -td ~/workspace/data/test_dataset/ -od ~/workspace/data/iti_test/test/ -nwf ~/workspace/data/network_data-test/network.txt -edaf ~/workspace/data/network_data-test/edge_attributes.txt -gif /home/data/annotations/stable/ncbi/Homo_sapiens.gene_info.gz -r 0.02 -th 0.15 -c 1 -nrand 100 -abs -sc averageCorrelation
$0 -singleOutput  -ef ~/workspace/data/test_dataset/IPC.txt.gz -p GPL6244 -nwf ~/workspace/data/network_data-test/network.txt -edaf ~/workspace/data/network_data-test/edge_attributes.txt -gif /home/data/annotations/stable/ncbi/Homo_sapiens.gene_info.gz  -stop 1000 -rate 0.02 -th 0.2 -d 15 -c 1 -nrand 300 -abs -sc averageCorrelation

Please report problems at <ghislain.bidaut\@inserm.fr>

";
    exit(0);
}


if (defined $useAnti) {
    print "\n    Detecting profiles ANTI-correlated with experimental condition\n";
    $useAbsVal = undef;
}
else {
    print "\n    Detecting profiles normally-correlated with experimental condition\n";
}

if (defined $useAbsVal) {
    print "    Absolute Value for scores/correlations is USED\n\n\n";
}
else {
    print "    Absolute Value for scores/correlations is NOT USED\n\n\n";
}


## Main program ##

print "\n********* Read edge attributes\n";
# %h_totalEdgeAtts("$gi1-$gi2" => \%($interaction_type => \%('Direction' => $direction_symbol)))
my (%h_totalEdgeAtts, @a_TFs);
open(FILE, "<$f_totalEda") or die "Can't open $f_totalEda:\n$!";
my $isHeader = 1;
my $motInd;
while(<FILE>){
    next if ($_ =~ m/^#/);
    $_ =~ s/[\r\n]+$//;
    my @a_row = split('\t', $_);

    if($isHeader){
        my @a_motInd = grep{$a_row[$_] eq 'Direction'}(0..$#a_row);
        $motInd = $a_motInd[0];
        $isHeader = 0;
        next;
    }

    my @a_int = split(/\ \(|\)\ /, $a_row[0]);
    my $node1 = $a_int[0];
    my $intType = $a_int[1];
    my $node2 = $a_int[2];

    my $mot = $a_row[$motInd];

    if(defined $networkType){
        next if(($networkType eq 'int' and $intType ne 'pp') ||
                ($networkType eq 'reg' and $intType ne 'pd'));
    }

    $h_totalEdgeAtts{"$node1-$node2"}{$intType}{'Direction'} = $mot;

    push(@a_TFs, $node1) if(scalar(grep(/>/, $mot)));
    push(@a_TFs, $node2) if(scalar(grep(/</, $mot)));
}
close FILE;

print "********* Read Network File\n";

# \%(geneID => \@neighbor_genes)
my $rh_network       = Utils::readHashFile($f_network);
my $rh_randomNetwork = Network::createRandomNetwork($rh_network);


print "\n********* Read Annotations\n";

my ($rh_geneIDtoSymbol, $rh_geneSymbolToID) = Utils::readGeneInfo($f_geneInfo);

print "\n********* Read HTS Data\n";

# \%($dataset_name => \%('cond'          => \%($sample_ID => $binary_value),
#                        'rancond'       => \%($sample_ID => $binary_value),
#                        'collapseddata' => \%($geneID => \%($sample => $score)),
#                        'expe'          => \@sampleList))
my $rh_datasets = HTS::readInputDataDir($d_inputData, $f_exp);

Network::calcDatasetsOverlapWithNetwork($rh_datasets, $rh_network);

my @a_datasets = keys %$rh_datasets;

#print "\n********* Correlation Calculation on HTS Data\n";

my $noCond;
for my $label(@a_datasets) {

    if(scalar(keys(%{$$rh_datasets{$label}{'cond'}})) == 1){    # NoConds
        if($scoreType ne 'averageCorrelation'){
            warn "\nSingle Conditions mode detected in dataset $label: score type set to averageCorrelation\n";
            $scoreType = 'averageCorrelation';
        }
        $noCond = 1;
        $useSAMScore = undef;
    }
    elsif(defined $$rh_datasets{$label}{'MultiValues'}){
        if($scoreType ne 'averageCorrelation'){
            warn "\nMultiValues Conditions mode detected in dataset $label: score type set to averageCorrelation\n";
            $scoreType = 'averageCorrelation';
        }
        warn "\nMultiValues Conditions mode detected in dataset $label: score type switched to SAM score\n";
        $useSAMScore = 1;
    }
    elsif(defined $useSAMScore){
        $$rh_datasets{$label}{'MultiValues'} = 1;
    }
}


# Node attribute hash:
# \%($geneID => \%('Correlations' => \%("correlation-$datasetLabel" => $gene_correlation))).
my($rh_nodeAtts, $maxSampleNb) = HTS::projectsInputData($rh_datasets, $useAbsVal, $useAnti);


#print "\n********* Processes node attribute hash\n";
for my $geneID(keys %$rh_nodeAtts){

    if (defined $$rh_geneIDtoSymbol{$geneID}) {
        # \%($geneID => \%('Gene_symbol' => $gene_symbol)
        $$rh_nodeAtts{$geneID}{'Gene_symbol'} = $$rh_geneIDtoSymbol{$geneID};
    }
    else{
        $$rh_nodeAtts{$geneID}{'Gene_symbol'} = undef ;
    }

    if(scalar(@a_TFs)){
        if(Utils::in_array($geneID, \@a_TFs)){
            # \%($geneID => \%('TF' => 1)
            $$rh_nodeAtts{$geneID}{'TF'} = 1;
        }
        else{
            $$rh_nodeAtts{$geneID}{'TF'} = undef;
        }
    }

    for my $label(@a_datasets) {
        if(!defined  $$rh_nodeAtts{$geneID}{'Correlations'}{$label}) {
            $$rh_nodeAtts{$geneID}{'Correlations'}{$label} = undef;
        }
    }

    if(scalar(@a_datasets) > 1) {
        my @a_tmp;
        for my $label(@a_datasets) {
            if(defined  $$rh_nodeAtts{$geneID}{'Correlations'}{$label}) {
                push(@a_tmp, $$rh_nodeAtts{$geneID}{'Correlations'}{$label});
            }
        }

        my($avCorr, $std) = Utils::avstd(@a_tmp);
        # \%($geneID => \%('Average_correlation' => $average_correlation)
        $$rh_nodeAtts{$geneID}{'Average_correlation'} = $avCorr;
    }
}

#print "\n********* Processes nodes degree\n";
for my $edge (keys %h_totalEdgeAtts){

    my @nodes = split("-", $edge);
    my $gene1 = $nodes[0];
    my $gene2 = $nodes[1];

    $$rh_nodeAtts{$gene1}{'Gene_degree'}++ if(defined $$rh_nodeAtts{$gene1});
    $$rh_nodeAtts{$gene2}{'Gene_degree'}++ if(defined $$rh_nodeAtts{$gene2});
}

detectDiscriminativeSubnetworks(
    $rh_network,     $rh_randomNetwork, $start,         $stop,
    $rh_nodeAtts,    \%h_totalEdgeAtts, $f_seedList,    $rh_datasets, $maxSampleNb,
    $scoreType,        $pearsonCorrTh,
    $datasetCons,    $nrand,         $snwDepth,
    $improvRate,     $d_output,      $fiTag,
    $nwtTag,         $useAbsVal,     $useAnti,
    $outInCurrDir,   $debug);

# Time measurement system - ENDING  #TMS
my $oldTime = $time;
$date       = localtime(time);
$time       = time;
my $period  = $time - $oldTime;

# Printing exit message
#print "\n$date: process complete\n";
#print "It took $period seconds to run\n";
#print "$0 done.\n\n";


## Subroutines ##

=pod

=head2 detectDiscriminativeSubnetworks()

Executes functions which detect and print discriminative subnetworks for each condition (real, random).

=cut

sub detectDiscriminativeSubnetworks {

    my $rh_network       = shift;
    my $rh_randomNetwork = shift;
    my $start            = shift;
    my $stop             = shift;
    my $rh_nodeAtts      = shift;
    my $rh_totalEdgeAtts = shift;
    my $f_seedList       = shift;
    my $rh_datasets      = shift;
    my $maxSampleNb      = shift;
    my $scoreType        = shift;
    my $pearsonCorrTh    = shift;
    my $datasetCons      = shift;
    my $nrand            = shift;
    my $snwDepth         = shift;
    my $improvRate       = shift;
    my $d_output         = shift;
    my $fiTag            = shift;
    my $nwtTag           = shift;
    my $useAbsVal        = shift;
    my $useAnti          = shift;
    my $outInCurrDir     = shift;
    my $debug            = shift;

    my @a_nodes;
    if(defined $f_seedList){
        my $rh_nodes = Utils::readHashFile($f_seedList);
        @a_nodes     = sort(keys %$rh_nodes);

        if (!defined $stop || $stop > scalar(@a_nodes)) {
            $stop = scalar(@a_nodes);
        }
    }
    else{
        @a_nodes = sort(keys %$rh_network);

        if (!defined $stop || $stop > scalar(@a_nodes)) {
            $stop = scalar(@a_nodes);
        }
    }

    my $randomNetworkUsed;

    print "\nPART I\n";
    print "Looking for discriminative subnetworks\n";

    my $realSnwMinSize  = 5;
    my $snwTag = $nwtTag . "snw";

    # detectSubnets(\@all_nodes, $start, $stop, \%all_interactions(geneID => \@neighbor_genes),
    # \%node_attributes($geneID => \%('Correlations' => \%("correlation-$datasetLabel" => $gene_correlation),
    #                                 'Average_correlation' => $average_correlation,
    #                                 'Gene_symbol' => $gene_symbol,
    #                                 'TF' => $1_ou_undef,
    # \%datasets($dataset_name => \%('cond'          => \%($sample_ID => $binary_value),
    #                                'rancond'       => \%($sample_ID => $binary_value),
    #                                'collapseddata' => \%($geneID => \%($sample => $score)),
    #                                'expe'          => \@sampleList)),
    # $pearson_correlation_threshold, $consensus, $random_type, $minSize, $average_size, $standard_deviation_size,
    # $subnetwork_depth, $improvement_rate, $useAbsVal, $useAnti, $tag, $debug)
    my ($rh_subnets_nwa, $ra_rate) =
        detectSubnets(\@a_nodes,         $start,            $stop,           $rh_network,
                      $rh_datasets,      $maxSampleNb,      $scoreType,      $pearsonCorrTh,
                      $datasetCons,      0,                 $realSnwMinSize, $snwDepth,
		      undef,             $improvRate,       $useAbsVal,      $useAnti,
                      $snwTag,           1,                 $debug);

    # printSubnets($general_output_directory, $output_directory_of_subnetworks, $tag, $file_index,
    # \%subnetworks_nwas($subnetwork_ID => \%('Edges' => \@("$gene1-$gene2"),
    #                                         'Nodes' => \%($geneID => 1)),
    #                                      \%('Scores' => \%($datasetLabel => $subnetwork_score),
    # \%node_attributes,
    # \%edge_attributes("$gi1-$gi2" => \%($interaction_type => \%('Direction' => $direction_symbol))),
    # \@rate, $outInCurrDir, $debug)
    printSubnets($d_output,          "subnets/",        "$fiTag$snwTag", $rh_subnets_nwa,
                 $rh_nodeAtts,       $rh_totalEdgeAtts, $ra_rate,
                 $randomNetworkUsed, $outInCurrDir,     1,               $debug);

    my $randomth        = 0.0;
    my $randomConsensus = 1;
    my @a_av_std        = Network::getSizeDistribution($rh_subnets_nwa);

    if(scalar(keys %$rh_subnets_nwa) != 0){
        print "PART II\n";
        print "Generating $nrand random subnets (Type A p-values: randomizing network aggregation decision)\n";

        $snwTag = $nwtTag . "rand1";
	
        ($rh_subnets_nwa, $ra_rate) =
            detectSubnets(\@a_nodes,         0,                $nrand,           $rh_network,
                          $rh_datasets,      $maxSampleNb,     $scoreType,       $randomth,
			  $randomConsensus,  1,                2,                $snwDepth,
			  \@a_av_std,        undef,            $useAbsVal,       $useAnti,
			  $snwTag,           0,                $debug);
	
        printSubnets($d_output,          "random1-subnets/", "$fiTag$snwTag", $rh_subnets_nwa,
                     $rh_nodeAtts,       $rh_totalEdgeAtts,  $ra_rate,
                     $randomNetworkUsed, $outInCurrDir,      0,               $debug);
    }
    else{
        print "PART II skipped\n";
    }

    print "\nPART III\n";
    print "Generating random subnets (Type B p-values: shuffled clinical condition)\n";

    $snwTag = $nwtTag . "rand2";

    ($rh_subnets_nwa, $ra_rate) =
        detectSubnets(\@a_nodes,         0,                $nrand,           $rh_network,
                      $rh_datasets,      $maxSampleNb,     $scoreType,       $randomth,
		      $randomConsensus,  2,                2,                $snwDepth,
		      undef,             $improvRate,      $useAbsVal,       $useAnti,
		      $snwTag,           0,                $debug);

    printSubnets($d_output,          "random2-subnets/", "$fiTag$snwTag", $rh_subnets_nwa,
                 $rh_nodeAtts,       $rh_totalEdgeAtts,  $ra_rate,
                 $randomNetworkUsed, $outInCurrDir,      0,               $debug);

    $randomNetworkUsed = 1;
    print "\nPART IV\n";
    print "Looking for discriminative subnetworks in random interactome (Type C p-values)\n";

    $snwTag = $nwtTag . "rand3";

    ($rh_subnets_nwa, $ra_rate) =
        detectSubnets(\@a_nodes,         0,                $nrand,           $rh_randomNetwork,
                      $rh_datasets,      $maxSampleNb,     $scoreType,       $randomth,
                      $randomConsensus,  0,                2,                $snwDepth,
		      undef,             $improvRate,      $useAbsVal,       $useAnti,
		      $snwTag,           0,                $debug);

    printSubnets($d_output,          "random3-subnets/", "$fiTag$snwTag", $rh_subnets_nwa,
                 $rh_nodeAtts,       $rh_totalEdgeAtts,  $ra_rate,
                 $randomNetworkUsed, $outInCurrDir,      0,               $debug);
    
}


=pod

=head2 detectSubnets()

Detects subnetworks and calculates their score, respectively using exploreSeed() and networkScore().

Returns:
\%h_subnets_nwa($subnetwork_ID => \%('Edges' => \@("$gene1-$gene2"),
                                     'Nodes' => \%($geneID => 1))),
                                  \%('Scores' => \%($datasetLabel => $subnetwork_score),

=cut

sub detectSubnets {

    my $ra_nodes         = shift;
    my $start            = shift;
    my $stop             = shift;
    my $rh_network       = shift;
    my $rh_datasets      = shift;
    my $maxSampleNb      = shift;
    my $scoreType        = shift;
    my $pearsonCorrTh    = shift;
    my $datasetCons      = shift;
    my $randomType       = shift;
    my $minSize          = shift;
    my $snwDepth         = shift;
    my $ra_av_std        = shift;
    my $improvRate       = shift;
    my $useAbsVal        = shift;
    my $useAnti          = shift;
    my $snwTag           = shift;
    my $realSnws         = shift;
    my $debug            = shift;

    my $nodecount = 0;
    my $total     = $stop - $start;
    my (@a_rate, %h_subnets_nwa, %h_networkAtts);

    print "Type $randomType networks - Consensus is set to $datasetCons (Pearson) for [$snwTag] subnetworks\n";
    
#    print "Pearson correlation threshold is set to $pearsonCorrTh  - score Type $scoreType\n";

    for (my $nodeIndex = $start; $nodeIndex < $stop; $nodeIndex++) {
        my $seed;

        if ($randomType > 0) {
            $seed = $$ra_nodes[rand(scalar @$ra_nodes)];
        }
        else {
            $seed = $$ra_nodes[$nodeIndex];
        }
        next if(!defined $seed);

        $nodecount++;
#        if (($nodecount % 10 == 0)) {
#            print "$nodecount/$total done\t(Current seed = $seed)\n";
#        }

        my $checkSeed = 0;
        foreach my $label (keys %$rh_datasets) {
            if (defined $$rh_datasets{$label}{'Gene_scores'}{$seed}{"correlation-$label"}) {
                $checkSeed++;
            }
        }

        if($checkSeed == 0){
	    if($debug) { print "DEBUG: The considered seed $seed is ABSENT from the selected datasets\n"; }
            next;
        }

        my $rh_subnet = exploreSeed($seed,           $rh_network,    $rh_datasets,      $maxSampleNb,
                                    \@a_rate,        $scoreType,     $pearsonCorrTh,    $datasetCons,
                                    $improvRate,     $randomType,    $snwDepth,         $ra_av_std,
				    $useAbsVal,      $useAnti,       $snwTag,           $debug);

        next if (scalar(keys %{$$rh_subnet{'Nodes'}}) < $minSize);

        for my $label (keys %$rh_datasets) {
            my $score = Subnetworks::networkScore($rh_datasets,
                                     \%{$$rh_subnet{'subNetworkProfiles'}},
                                     \%{$$rh_subnet{'subNetworkProfilesSize'}},
                                     \%{$$rh_subnet{'subNetworkScoreList'}},
                                     $scoreType, $label, $maxSampleNb, $useAbsVal, $useAnti);

            $h_networkAtts{"$snwTag-$seed"}{'Scores'}{$label} = $score;
        }
	
        my $snwID = "$snwTag-$seed";
        # Add subnetwork in hash.
        $h_subnets_nwa{$snwID} = $rh_subnet;
        for my $key(keys %{$h_networkAtts{$snwID}}) {
            $h_subnets_nwa{$snwID}{$key} = $h_networkAtts{$snwID}{$key};
        }
    }

    print "Number of subnetworks found = " . scalar(keys %h_subnets_nwa) . " for tag $snwTag.\n\n";

    if($realSnws || $debug){
        Subnetworks::includeEdges(\%h_subnets_nwa, $rh_network);
    }

    return \%h_subnets_nwa, \@a_rate;
}

=pod

=head2 printSubnets()

Executes Subnetworks::writeSubnets().

=cut

sub printSubnets {
    my $d_output          = shift;
    my $d_subOut          = shift;
    my $bigTag            = shift;
    my $rh_subnets_nwa    = shift;
    my $rh_nodeAtts       = shift;
    my $rh_totalEdgeAtts  = shift;
    my $ra_rate           = shift;
    my $randomNetworkUsed = shift;
    my $outInCurrDir      = shift;
    my $realSnws          = shift;
    my $debug             = shift;

    if(defined $outInCurrDir){
        $d_subOut = "";
    }
    else{
        mkdir "$d_output/$d_subOut";

        if($debug){
            my $d_subOutRate;
            $d_subOutRate = "$d_output/$d_subOut/rate/";
            mkdir $d_subOutRate;
            open(SUBNETSRATE,">$d_subOutRate$bigTag-rate.txt") || die "Cannot open $bigTag-rate.txt:\n$!";
            foreach(@$ra_rate) { print SUBNETSRATE "$_\n"; }
            close SUBNETSRATE;
        }
    }

    my($f_nnf, $f_noa);
    if($realSnws || $debug){
        $f_nnf = "$d_output$d_subOut$bigTag.nnf";
        $f_noa = "$d_output$d_subOut$bigTag-gene-scores-noa.txt";
    }

    Subnetworks::writeSubnets(
        $rh_subnets_nwa, $rh_nodeAtts, $rh_totalEdgeAtts,
        $f_nnf,
        "$d_output$d_subOut$bigTag-scores-nwa.txt",
        $f_noa,
        $randomNetworkUsed);
}

=pod

=head2 exploreSeed()

Explores the network from a seed gene to get a subnetwork.

=cut

sub exploreSeed {
    my $seed             = shift;
    my $rh_network       = shift;
    my $rh_datasets      = shift;
    my $maxSampleNb      = shift;
    my $ra_rate          = shift;
    my $scoreType        = shift;
    my $pearsonCorrTh    = shift;
    my $datasetCons      = shift;
    my $improvRate       = shift;
    my $randomType       = shift;
    my $snwDepth         = shift;
    my $ra_av_std        = shift;
    my $useAbsVal        = shift;
    my $useAnti          = shift;
    my $snwTag           = shift;
    my $debug            = shift;

    my %h_subnetwork;      # Subnetwork Object (contains all subnetwork infos)
    my %h_snwNodes;        # Contains all subnetwok nodes
    my %h_nodesToExplore;  # Contains all future node to explore

    my %h_snwProfile;      # Data: \%($dataset => \%($sample => $gene_score_sum))
    my %h_snwProfileSize;  # Nb of genes in snw: \%($dataset => \%($sample => $size))

    my %h_snwGeneScores;   # Contains subnetwork score of all datasets

    my($snwSize, $wantedSize);
    if ($randomType == 1) {
        $wantedSize = Utils::approximateNormalDistribution($$ra_av_std[0],
                                                  $$ra_av_std[1],
                                                  20);
        $wantedSize = sprintf('%.0f', $wantedSize);
        $snwSize = 1;
    }

    # Initialisation:
    # The subnetwork contains only the seed, so the seed is the only node to explore
    addNodeToSubnetwork($seed,            \%h_snwNodes,
                        \%h_snwProfile,   \%h_snwProfileSize,
                        \%h_snwGeneScores,
                        $rh_datasets,
                        $scoreType,       $randomType);
    $h_nodesToExplore{$seed} = $seed;

    if($debug) {
        print "\nDEBUG: *** Explore SEED = $seed ***
MAIN LOOP of maximum network depth (i.e. distance to the subnetwork seed)\n";
    }

    my %h_rejectedNodes;
    # Main loop of maximum network depth (i.e. distance to the subnetwork seed)
    for (my $depth = 0; $depth < $snwDepth; $depth++) {
        if($debug) { print "DEBUG: DEPTH = $depth\n"; }

        my $nFlag;
       
        my %h_neighborNodes;  # Contains full list of neighbor nodes to test
        my %h_acceptedNodes;  # Contains full list of accepted nodes

        if($debug) { print "DEBUG: $seed NEIGHBORHOOD exploration of " . scalar(keys %h_nodesToExplore) . " valid nodes\n"; }

        foreach my $node (keys %h_nodesToExplore) {
            foreach my $neighborNode (@{$$rh_network{$node}}) {
                # Test if neighbor has already been rejected - If yes it is never considered again;
                # Test if neighbor and links was already included. In this case we skip it.
                next if (defined $h_rejectedNodes{$neighborNode} ||
                         defined $h_snwNodes{$neighborNode}      ||
                         defined $h_neighborNodes{$neighborNode});

                # Case 1: It is a complete new node: we add node and link.
                $h_neighborNodes{$neighborNode} = $neighborNode;
            }
        }

        if($debug) { print "DEBUG: Number of potential NEIGHBORS = " . scalar(keys %h_neighborNodes) . "\n"; }

        # If no more neighbors (highly improbable), stop exploration.
        last if (scalar(keys %h_neighborNodes) == 0);

        if($debug) { print "DEBUG: VALIDATION step of neighborhood nodes\n"; }

        # Then, we proceed with list of neightbors to be validated.
        # For each neighbor, computes and tests HTS scores.
        # If a score passes the threshold, node is added to subnetwork.
        foreach my $neighborNode (keys %h_neighborNodes) {
            if($debug) { print "DEBUG: Current neighbor = $neighborNode\n"; }
            if($randomType == 1){
                last if($snwSize >= $wantedSize);
                $h_acceptedNodes{$neighborNode} = $neighborNode;
                $snwSize++;
                next;
            }

            # Test improvement on HTS scores
            $nFlag = testIfImprovementInHTSScore($neighborNode,
                                                      \%h_snwProfile, \%h_snwProfileSize, \%h_snwGeneScores,
                                                      $rh_datasets,   $maxSampleNb,
                                                      $ra_rate,
                                                      $pearsonCorrTh, $improvRate, $scoreType,
                                                      $randomType,    $useAbsVal,
                                                      $useAnti,       $debug);

	    
	    
	    if ($nFlag >= $datasetCons) {
		$h_acceptedNodes{$neighborNode} = $neighborNode;
	    }
	    else { $h_rejectedNodes{$neighborNode} = $neighborNode; }
	    
        }

        if($debug) { print "DEBUG: There are " . scalar(keys %h_acceptedNodes) . " ACCEPTED nodes\n"; }
        # Populate final subnetworks profiles and scores

        last if (scalar(keys %h_acceptedNodes) == 0);

        # Reinitialization
        %h_nodesToExplore = ();
        # h_nodesToExplore contains only accepted nodes
        foreach my $acceptedNode (keys %h_acceptedNodes) {
            $h_nodesToExplore{$acceptedNode} = $acceptedNode;
            # For accepted nodes, link subnetworkprofile and subnetworkprofilesize are transferred to subnetork
            # Update scores/average profiles
            addNodeToSubnetwork($acceptedNode,     \%h_snwNodes,
                                \%h_snwProfile,    \%h_snwProfileSize,
                                \%h_snwGeneScores,
                                $rh_datasets,
                                $scoreType,        $randomType);
        }

        last if($randomType == 1 and $snwSize >= $wantedSize);

        #warn "\nWarning: Maximum subnetwork depth reached (snw type: $snwTag; seed: $seed)!\n\n" if($depth == $snwDepth - 1);
    } #for $depth

    $h_subnetwork{'Nodes'}                  = \%h_snwNodes;
    $h_subnetwork{'subNetworkProfiles'}     = \%h_snwProfile;
    $h_subnetwork{'subNetworkProfilesSize'} = \%h_snwProfileSize;
    $h_subnetwork{'subNetworkScoreList'}    = \%h_snwGeneScores;

    if($debug) {
        print "DEBUG: SIZE of $seed subnetwork: " . scalar(keys(%h_snwNodes)) . "\n";
        print "WANTED size: $wantedSize\n" if($randomType == 1);
    }

    return \%h_subnetwork;
}

=pod

=head2 testIfImprovementInHTSScore()

Tests if there is an improvement of score on HTS data.

=cut

sub testIfImprovementInHTSScore {
    my $node              = shift;
    my $rh_snwProfile     = shift;
    my $rh_snwProfileSize = shift;
    my $rh_snwGeneScores  = shift;
    my $rh_datasets       = shift;
    my $maxSampleNb       = shift;
    my $ra_rate           = shift;
    my $pearsonCorrTh     = shift;
    my $improvRate        = shift;
    my $scoreType         = shift;
    my $randomType        = shift;
    my $useAbsVal         = shift;
    my $useAnti           = shift;
    my $debug             = shift;

    my $nFlag = 0;

    # Score computed by averaging gene score in subnetwork then calculating correlation.
    # for each dataset label, test if there is an improvement, if yes flag node
    foreach my $label(keys %$rh_datasets) {
        last if($scoreType ne 'averageScore');
        if($debug) { print "DEBUG: DATASET = $label\n"; }
        next if(!defined $$rh_datasets{$label}{'Gene_scores'}{$node}{"correlation-$label"});
	
        my $rh_data          = $$rh_datasets{$label}{'collapseddata'};
        my $ra_samples       = $$rh_datasets{$label}{'expe'};
        my $rh_conditions    = $$rh_datasets{$label}{'cond'};
        my $rh_ranconditions = $$rh_datasets{$label}{'rancond'};

        my (@a_dataVect, @a_vectSize);
        foreach(@$ra_samples) {
            push(@a_dataVect, 0);
            push(@a_vectSize, 0);
        }

        my $subnetworkScore;
        if(defined $$rh_snwProfileSize{$label}) { # Subnetwork not empty

            $subnetworkScore = Subnetworks::networkScore($rh_datasets,       $rh_snwProfile,
                                            $rh_snwProfileSize, $rh_snwGeneScores, $scoreType,
                                            $label,             $maxSampleNb,
                                            $useAbsVal,         $useAnti);

            for(my $i=0; $i < scalar(@$ra_samples); $i++) {
                my $sample = $$ra_samples[$i];
                if(defined $$rh_snwProfile{$label}{$sample}) {
                    $a_dataVect[$i] = $$rh_snwProfile{$label}{$sample};
                    $a_vectSize[$i] = $$rh_snwProfileSize{$label}{$sample};
                }
                if(defined $$rh_data{$node}{$sample}) {
                    $a_dataVect[$i] += $$rh_data{$node}{$sample};
                    $a_vectSize[$i] += 1;
                }
                # Average profile of genes for each sample
                if($a_vectSize[$i] != 0) { $a_dataVect[$i] /= $a_vectSize[$i]; }
            }
        }
        else { # Subnetwork empty
            for(my $i=0; $i < scalar(@$ra_samples); $i++) {
                my $sample = $$ra_samples[$i];
                if(defined $$rh_data{$node}{$sample}) { $a_dataVect[$i] = $$rh_data{$node}{$sample}; }
                else                                  { $a_dataVect[$i] = undef; }
            }
        }

        # compare prospective value with threshold
        my $prospectiveValue;
        if(defined $$rh_datasets{$label}{'MultiValues'}) {
            # random = 2 -> conditions are randomized
            my @a_condVect = ();
            if($randomType == 2) { for my $sample(@$ra_samples) { push(@a_condVect, $$rh_ranconditions{$sample}); } }
            else                 { for my $sample(@$ra_samples) { push(@a_condVect, $$rh_conditions{$sample}); } }

            if($debug) { print "DEBUG: vect 1 = @a_dataVect\nDEBUG: vect 2 = @a_condVect\n"; }

            $prospectiveValue  = Utils::multiconditionalDifferentialExpression(\@a_dataVect, \@a_condVect);
            $prospectiveValue *= sqrt(scalar @$ra_samples) / sqrt($maxSampleNb);
        }
        elsif(defined  $$rh_datasets{$label}{'mcond'}) {

            my $rh_conditions;
            my $rh_conditionsWeight;
            if($randomType == 2) {
                $rh_conditions = $$rh_datasets{'mcondrandom'};
                $rh_conditionsWeight = $$rh_datasets{'mcondweightrandom'};
            }
            else {
                $rh_conditions = $$rh_datasets{$label}{'mcond'};
                $rh_conditionsWeight = $$rh_datasets{$label}{'mcondweight'};
            }
            my $score = 0;
            foreach my $condType(keys %$rh_conditionsWeight) {
                my @a_condVect = ();
                for my $sample (@$ra_samples) {push(@a_condVect, $$rh_conditions{$sample}{$condType});}

                $score += $$rh_conditionsWeight{$condType} * Utils::pearsonCorrelation(\@a_dataVect, \@a_condVect);
            }
            $prospectiveValue = $score;
        }
        else{

            my @a_condVect = ();

            if($randomType == 2) { for my $sample(@$ra_samples) { push(@a_condVect, $$rh_ranconditions{$sample}); } }
            else                 { for my $sample(@$ra_samples) { push(@a_condVect, $$rh_conditions{$sample}); } }

            if($debug) { print "DEBUG: vect 1 = @a_dataVect\nDEBUG: vect 2 = @a_condVect\n"; }

            $prospectiveValue  = Utils::pearsonCorrelation(\@a_dataVect, \@a_condVect);
            $prospectiveValue *= sqrt(scalar @$ra_samples) / sqrt($maxSampleNb);
        }

        if (defined $useAbsVal) {
            $prospectiveValue = abs($prospectiveValue);
        }
        elsif (defined $useAnti) {
            $prospectiveValue = -$prospectiveValue;
        }

        if ( !defined $subnetworkScore ) { $subnetworkScore = -$improvRate; }
        if($debug) { print "DEBUG: node = $node pv = $prospectiveValue ss = $subnetworkScore ir = $improvRate\n"; }

        if ($prospectiveValue >= $pearsonCorrTh &&
            $prospectiveValue >= ($subnetworkScore + $improvRate)) {
            $nFlag++;
            if($debug) { print "DEBUG: node $node accepted for dataset $label\n"; }
            if($debug and $subnetworkScore > -$improvRate) {
                push(@$ra_rate, $prospectiveValue-$subnetworkScore);
            }
        }
    }# for each dataset labels


    # Score computed by averaging correlations among genes.
    # for each dataset label, test if there is an improvement, if yes flag node
    foreach my $label(keys %$rh_datasets) {
        last if($scoreType ne 'averageCorrelation');
        next if(!defined $$rh_datasets{$label}{'Gene_scores'}{$node}{"correlation-$label"});

        my @a_vect = ();
        my $subnetworkScore;
        if(defined $$rh_snwGeneScores{$label}) { # Subnetwork not empty
            $subnetworkScore = Subnetworks::networkScore($rh_datasets,       $rh_snwProfile,
                                            $rh_snwProfileSize, $rh_snwGeneScores, $scoreType,
                                            $label,             $maxSampleNb,
                                            $useAbsVal,         $useAnti);

            foreach my $curNode (keys %{$$rh_snwGeneScores{$label}}) {
                push(@a_vect, $$rh_snwGeneScores{$label}{$curNode});
            }
        }

        # Random = 2 - conditions are randomized
        if($randomType == 2) {
            push(@a_vect, $$rh_datasets{$label}{'Gene_scores'}{$node}{"randProjectedInputData-$label"});
        }
        # All other cases - conditions are not randomized
        else {
            push(@a_vect, $$rh_datasets{$label}{'Gene_scores'}{$node}{"correlation-$label"});
        }

        # compare prospective value with threshold
        my($prospectiveValue, $std2) = Utils::avstd(@a_vect);

        if ( !defined $subnetworkScore ) { $subnetworkScore = -$improvRate; }

        if ($prospectiveValue >= $pearsonCorrTh &&
            $prospectiveValue >= ($subnetworkScore + $improvRate)) {
            $nFlag++;
        }
    }

    return $nFlag;
}


=pod

=head2 addNodeToSubnetwork()

Adds node to subnetwork.

=cut

sub addNodeToSubnetwork {
    my $node              = shift;
    my $rh_snwNodes       = shift;
    my $rh_snwProfile     = shift;
    my $rh_snwProfileSize = shift;
    my $rh_snwGeneScores  = shift;
    my $rh_datasets       = shift;
    my $scoreType         = shift;
    my $randomType        = shift;

    $$rh_snwNodes{$node} = $node;

    # Update average profile
    foreach my $label(keys %$rh_datasets) { # transcriptome label
        next if(!defined $$rh_datasets{$label}{'Gene_scores'}{$node}{"correlation-$label"});

        if($scoreType eq 'averageScore'){
            my $rh_data    = $$rh_datasets{$label}{'collapseddata'};
            my $ra_samples = $$rh_datasets{$label}{'expe'};

            if(defined $$rh_snwProfileSize{$label}) {
                for(my $i = 0; $i < scalar @$ra_samples; $i++) {
                    my $sample = $$ra_samples[$i];
                    if(defined $$rh_data{$node}{$sample}) {
                        $$rh_snwProfile{$label}{$sample} += $$rh_data{$node}{$sample};
                        $$rh_snwProfileSize{$label}{$sample}++;
                    }
                }
            }
            else {
                for(my $i = 0; $i < scalar @$ra_samples; $i++) {
                    my $sample = $$ra_samples[$i];
                    if(defined $$rh_data{$node}{$sample}) {
                        $$rh_snwProfile{$label}{$sample}     = $$rh_data{$node}{$sample};
                        $$rh_snwProfileSize{$label}{$sample} = 1;
                    }
                    else {
                        $$rh_snwProfile{$label}{$sample}     = undef;
                        $$rh_snwProfileSize{$label}{$sample} = 0;
                    }
                }
            }
        }
        else{
            # Update subnetworkScoreList
            # Case Random = 2 - conditions are randomized
            if($randomType == 2) {
                $$rh_snwGeneScores{$label}{$node} =
                    $$rh_datasets{$label}{'Gene_scores'}{$node}{"randProjectedInputData-$label"};
            }
            # All other cases - conditions are not randomized
            else {
                $$rh_snwGeneScores{$label}{$node} =
                    $$rh_datasets{$label}{'Gene_scores'}{$node}{"correlation-$label"};
            }
        }
    }    
}
