#!/usr/bin/perl
#
# ITI project
#
# Copyright Garcia Maxime, Bidaut Ghislain, Stahl Olivier, Rioualen Claire, Da Costa Quentin
# (2009-2017)
#
# maxime.garcia@inserm.fr
# ghislain.bidaut@inserm.fr
# claire.rioualen@inserm.fr
# quentin.dacosta@inserm.fr
# http://iti.sourceforge.net
#
# This software is a computer program whose purpose is to superimpose
# interactome  over  several  gene  expression  datasets  to  find
# discriminative subnetworks predicting metastatic relapse in breast
# cancer.
# 
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security.
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME
generateConditionFiles.pl

=head1  SYNOPSIS

=head1 DESCRIPTION
generateConditionFiles.pl generates proper conditions files (.cond) from a list of .pheno files for bootstrapping several datasets.

=head1 SUBROUTINES
none

=head1 Author
Maxime Garcia	maxime.garcia@inserm.fr

=cut

use strict;
use warnings;
use POSIX qw(ceil floor);
use Utils;
use Cwd;
use Data::Dumper;
use File::Basename;
# GLOBAL PARAMETERS
my $VERSION = 1.0;

########################
#     MAIN SECTION     #
########################

print "\n";

if ( $#ARGV < 1 ) {
	print "Usage: $0
		 -d  <Input directory phenotype files>
		 -e(--e)  <Input event> (e.g. selection on e.event and t.time annotation
                 -c <input condition> (e.g. selection on \"Annot\")
                 -ct <input condition true> (e.g. selection on Annot=Basal versus rest)
                 -o  <output directory>
		[-r  <number of random clinical conditions files>]
		[-l  <number of layer for stratification>]
		[-s  <selection=value>]
		[-t  <Censor Time (in month)>]
		[-te <Censor Time on event (in month)>]
		[-i  <items for stratification>]
		[-f  (Flag = ON)]
		[-n  (Stratification = OFF)]\n";
	exit;
}

my ( $inputDirectory,
     $event,
     $inputCondition,
     $inputConditionTrue,
     $outputDirectory,
     $layer,
     @a_temp,
     @a_headerSelect,
     @a_valueSelect,
     $stratification,
     $censorTime,
     $censorTimeEvent,
     @a_stratificationItem,
     $flag);

my $n_random = 0;

for ( my $i = 0; $i < (scalar @ARGV); $i++ ) {
	if($ARGV[$i] eq "-d")  { $inputDirectory  = "$ARGV[$i+1]"; }
	if($ARGV[$i] eq "-o")  { $outputDirectory = "$ARGV[$i+1]"; }
	if(($ARGV[$i] eq "-e") || ($ARGV[$i] eq "--e"))  { $event           = "$ARGV[$i+1]"; }
	if($ARGV[$i] eq "-c")  { $inputCondition  = "$ARGV[$i+1]";}
	if($ARGV[$i] eq "-ct") { $inputConditionTrue  = "$ARGV[$i+1]";}
	if($ARGV[$i] eq "-l")  { $layer           = "$ARGV[$i+1]"; $stratification = 0;}
	if($ARGV[$i] eq "-t")  { $censorTime      = "$ARGV[$i+1]"; }
	if($ARGV[$i] eq "-r")  { $n_random        = "$ARGV[$i+1]"; srand($n_random); $stratification  = 1; $layer = 2;}
	if($ARGV[$i] eq "-te") { $censorTimeEvent = "$ARGV[$i+1]"; }
	if($ARGV[$i] eq "-f")  { $flag            = 1; }
	if($ARGV[$i] eq "-n")  { $stratification  = 1; $layer = 2; }
	if($ARGV[$i] eq "-s")  {
		for ( my $j= $i+1; $j < scalar @ARGV; $j++) {
			if ($ARGV[$j] =~ m/^-/) { last; }
			push(@a_temp, $ARGV[$j]);
		}
	}
	if($ARGV[$i] eq "-i") {
		for ( my $j= $i+1; $j < scalar @ARGV; $j++) {
			if ($ARGV[$j] =~ m/^-/) { last; }
			push(@a_stratificationItem, $ARGV[$j]);
		}
	}
}

!defined $inputDirectory  and die "No Input Directory Defined";
!defined $outputDirectory and die "No Output Directory Defined";

mkdir $outputDirectory;

my @a_directoryContent = sort(Utils::readInputDir($inputDirectory));
my @a_inputFile;

foreach my $file (@a_directoryContent) { $file =~ m/.pheno$/ and push(@a_inputFile, "$inputDirectory/$file"); }

if(scalar @a_inputFile < 1) { die "No pheno file(s) in $inputDirectory"; }

(!(defined $event or defined $inputCondition)) and die "No Event nor InputCondition specified.";

if(defined $stratification && !defined $layer) { die "No Layer specified."; }

if($layer !~ m/^[0-9]+$/) { die "Number $layer is not in the proper format"; }

if($layer <= 1) { die "Number $layer is not large enough.\nIt should be at least 2."; }
if($n_random !~ m/^[0-9]+$/) { die "Number $n_random is not in the right format."; }

if ( $n_random ne 0) { print "$n_random conditions files are generated randomly for each dataset.\n"; $stratification = 1; }

if(defined $stratification && $stratification eq 1 ) { $layer = 1; }

(defined $censorTime && $censorTime !~ m/^[0-9]*\.*[0-9]$/)           and die "Censor time $censorTime is not in the right format : $!";
(defined $censorTimeEvent && $censorTimeEvent !~ m/^[0-9]*\.*[0-9]$/) and die "Censor time on event $censorTimeEvent is not in the right format : $!";

if(defined $event) {
    if(defined $censorTimeEvent) { print "# Following and cut off\nEvent Selected          = E.$event/T.$event\nCensor Time Event (m)   = $censorTimeEvent\n"; }
    else                          { print "Event Selected          = E.$event\n"; }
    if (defined $censorTime)      { print "Censor Time (in month)  = $censorTime\n"; }
    
    # Setting up stratification
    unshift(@a_stratificationItem, "E.$event");
    foreach my $item (@a_stratificationItem) { print "Stratification on item  = $item\n"; }
}
if(defined $inputCondition) {
    print "Input Condition = $inputCondition, selection if $inputCondition is equal to $inputConditionTrue\n";

    # Setting up stratification
    unshift(@a_stratificationItem, "$inputCondition");
    foreach my $item (@a_stratificationItem) { print "Stratification on item  = $item\n"; }
}

defined $flag and unshift(@a_temp,'FLAG=0');

foreach my $element (@a_temp) {
    my @a_temp2 = split (/=/, $element);
    push (@a_headerSelect, $a_temp2[0]);
    push (@a_valueSelect,  $a_temp2[1]);
}

if ( scalar @a_headerSelect > 0 ) {
    print "# Samples Selection\n";
    for ( my $i = 0; $i < scalar @a_headerSelect; $i++) {
	print "Phenotype Selected      = $a_headerSelect[$i]\n";
	print "Value Selected          = $a_valueSelect[$i]\n";
    }
}

print "# Partitioning\nNumber of layers        = $layer\n";

# Open and read pheno file
my $rh_pheno = readPhenoFile(\@a_inputFile);

if (!defined $censorTime && scalar @a_headerSelect < 1) {
	for my $dataset (sort keys %$rh_pheno) { print "Number of Samples       = " . (scalar keys %{$$rh_pheno{$dataset}}) . "\t($dataset)\n"; }
}

# FIRST FILTERING
# Check if phenotype, event and time exist
# Remove unwanted sample on given phenotype (example : ER = 0)
# %$rh_pheno countains every sample
# %h_removedSample countains every sample removed
my $s_log;
my %h_removedSample;
if ( scalar @a_headerSelect > 0 ) {
	for my $dataset (sort keys %$rh_pheno) {
		for my $sample (sort keys %{$$rh_pheno{$dataset}}) {
			for ( my $i = 0; $i < scalar @a_headerSelect; $i++) {
				if ( !defined $$rh_pheno{$dataset}{$sample}{$a_headerSelect[$i]} ) {
					die "Phenotype $a_headerSelect[$i] does not exist.\nPlease check your argument(s) and/or your .pheno file(s)";
				}
				
				# Event mode
				if(defined $event) {
				    if ( !defined $$rh_pheno{$dataset}{$sample}{"E.$event"} ) {
					die "Event E.$event does not exist.\nPlease check your argument(s) and/or your .pheno file(s).";
				    }
				    if ( !defined $$rh_pheno{$dataset}{$sample}{"T.$event"} ) {
					die "Time T.event does not exist.\nPlease check your argument(s) and/or your .pheno file(s).";
				    }
				}
				

				# InputCondition mode
				
				if(defined $inputCondition) {
				    if ( !defined $$rh_pheno{$dataset}{$sample}{$inputCondition} ) {
					die "Input Condition $inputCondition does not exist.\nPlease check your argument(s) and/or your .pheno file(s)";
				    }
				}
				
				# Keep a track of a removed sample, and why it was removed
				if ($$rh_pheno{$dataset}{$sample}{$a_headerSelect[$i]} ne "$a_valueSelect[$i]") {
					$h_removedSample{$dataset}{$sample} = 1;
					$s_log .= "$dataset\t$sample\tRemoved\t$a_headerSelect[$i] eq $a_valueSelect[$i]\n";
				}
			}
		}
	}
	# Print information on Filtering
	print "# Filtering on selected phenotype\n";
	for my $dataset (sort keys %$rh_pheno) { print "Samples Kept            = " . ((scalar keys %{$$rh_pheno{$dataset}})-(scalar keys %{$h_removedSample{$dataset}})) . "/" . (scalar keys %{$$rh_pheno{$dataset}}) . "\t($dataset)\n";	}
}

if ( scalar @a_stratificationItem > 0 ) {
    for my $dataset (sort keys %$rh_pheno) {
	for my $sample (sort keys %{$$rh_pheno{$dataset}}) {
	    if (defined $h_removedSample{$dataset}{$sample}) { next; }
	    foreach my $item (@a_stratificationItem) {
		if ( !defined $$rh_pheno{$dataset}{$sample}{$item} ) {
		    die "Phenotype $item does not exist.\nPlease check your argument(s) and/or your .pheno file(s) : $!";
		}
		if (("$$rh_pheno{$dataset}{$sample}{$item}" eq "NA") or ("$$rh_pheno{$dataset}{$sample}{$item}" eq "undef")) {
		    $h_removedSample{$dataset}{$sample} = 1;
		    $s_log .= "$dataset\t$sample\t$item = undef\n";
		}
	    }
	}
    }
}

# SECOND FILTERING
# %h_condition countains only samples kept after first and second filtering

my %h_condition;

# Case censortime AND censortime event defined
# If Value is 0 and Time < CensorTime => Sample is Censored
# If Value is 1 and Time > censorTimeEvent => Sample Changed, new Value is 0

if (defined $event and defined $censorTime && defined $censorTimeEvent) {

    for my $dataset (sort keys %$rh_pheno) {
	for my $sample (sort keys %{$$rh_pheno{$dataset}}) {
	    if (defined $h_removedSample{$dataset}{$sample}) { next; }
	    my ($value, $time);
	    
	    for my $header (sort keys %{$$rh_pheno{$dataset}{$sample}}) {
		
		# Keep value
		if ($header =~m/^E.$event$/) {
		    $value = $$rh_pheno{$dataset}{$sample}{$header};
		    # Remove Unwanted Samples
		    if (("$value" eq "NA") or ("$value" eq "undef") or !(defined $value)) {
			$s_log .= "$dataset\t$sample\tCensored\tE.$event = NA\n";
			last;
		    }
		}
		#  Keep time
		if ($header =~m/^T.$event$/) {
		    $time = $$rh_pheno{$dataset}{$sample}{$header};
		    # Remove Unwanted Samples
		    if (("$time" eq "NA") or ("$time" eq "undef") or !(defined $time) ) {
			$s_log .= "$dataset\t$sample\tCensored\tT.$event = NA\n";
			last;
		    }
		    if ($value == 0) {
			# If Value is 0 and Time < CensorTime => Sample is Censored
			if ($time < $censorTime) {
			    $s_log .= "$dataset\t$sample\tCensored\t$time < $censorTime\n";
			}
			# Else Sample is Kept
			# Info is kept on sample for stratification
			else {
			    ($$rh_pheno{$dataset}{$sample}{"E.$event"}, $h_condition{$dataset}{$sample}) = (0, 0);
			    $s_log .= "$dataset\t$sample\tKept\t$time > $censorTime\n";
			}
		    }
		    if ($value == 1) {
			# If Value is 1 and Time > censorTimeEvent => Sample Changed, new Value is 0
			# Info is kept on sample for stratification
			if ($time > $censorTimeEvent) {
			    ($$rh_pheno{$dataset}{$sample}{"E.$event"}, $h_condition{$dataset}{$sample}) = (0, 0);
			    $s_log .= "$dataset\t$sample\tChanged\t$time > $censorTimeEvent\n";
			}
			# Else Sample Kept, Nothing is changed
			# Info is kept on sample for stratification
			else {
			    ($$rh_pheno{$dataset}{$sample}{"E.$event"}, $h_condition{$dataset}{$sample}) = (1, 1);
			    $s_log .= "$dataset\t$sample\tUnchanged\t$time < $censorTimeEvent\n";
			}
		    }
		    
		    ($value, $time) = (undef, undef);
		    
		}
		
	    }
	}
    }
    
    # Print information on Filtering
    print "# Filtering on Censoring\\Following Time\n";
    for my $dataset (sort keys %h_condition) {
	print "Samples Kept            = " . (scalar keys %{$h_condition{$dataset}}) . "/" . (scalar keys %{$$rh_pheno{$dataset}}) . "\t($dataset)\n";
    }
}

# Censortime defined but censortime evvent undefined
# If Value is 0 and Time < CensorTime => Sample is Censored
# If Value is 1 Nothing is changed
elsif (defined $event and defined $censorTime && !defined $censorTimeEvent) {
    for my $dataset (sort keys %$rh_pheno) {
	for my $sample (sort keys %{$$rh_pheno{$dataset}}) {
	    if (defined $h_removedSample{$dataset}{$sample}) { next; }
	    my ($value, $time);
	    for my $header (sort keys %{$$rh_pheno{$dataset}{$sample}}) {
		# Keep value
		if ($header =~m/^E.$event$/) {
		    $value = $$rh_pheno{$dataset}{$sample}{$header};
		    # Remove Unwanted Samples
		    if (("$value" eq "NA") or ("$value" eq "undef") or !(defined $value)) {
			$s_log .= "$dataset\t$sample\tCensored\tE.$event = NA\n";
			last;
		    }
		}
		# Keep time
		if ($header =~m/^T.$event$/) {
		    $time = $$rh_pheno{$dataset}{$sample}{$header};
		    # Remove Unwanted Samples
		    if (("$time" eq "NA") or ("$time" eq "undef") or !(defined $time) ) {
			$s_log .= "$dataset\t$sample\tCensored\tT.$event = NA\n";
			last;
		    }
		    if ($value == 0) {
			# If Value is 0 and Time < CensorTime => Sample is Censored
			if ($time < $censorTime) {
			    $s_log .= "$dataset\t$sample\tCensored\t$time < $censorTime\n";
			}
			# Else Sample is Kept
			# Info is kept on sample for stratification
			else {
			    ($$rh_pheno{$dataset}{$sample}{"E.$event"}, $h_condition{$dataset}{$sample}) = (0, 0);
			    $s_log .= "$dataset\t$sample\tKept\t$time > $censorTime\n";
			}
		    }
		    if ($value == 1) {
			# If Value is 1 Nothing is changed
			($$rh_pheno{$dataset}{$sample}{"E.$event"}, $h_condition{$dataset}{$sample}) = (1, 1);
			$s_log .= "$dataset\t$sample\tUnchanged\tEvent 1 has no need to be changed\n";
		    }
		    ($value, $time) = (undef, undef);
		}
	    }
	}
    }
    # Print information on Filtering
    print "# Filtering on Censoring Time\n";
    for my $dataset (sort keys %h_condition) {
	print "Samples Kept            = " . (scalar keys %{$h_condition{$dataset}}) . "/" . (scalar keys %{$$rh_pheno{$dataset}}) . "\t($dataset)\n";
    }
}
# Case censorTimeEvent defined and censortime undefined
# If Value is 0 and Time < CensorTimeEvent => Sample is Censored
# If Value is 1 and Time > censorTimeEvent => Sample Changed, new Value is 0

elsif (defined $event and defined $censorTimeEvent && !defined $censorTime) {
    for my $dataset (sort keys %$rh_pheno) {
	for my $sample (sort keys %{$$rh_pheno{$dataset}}) {
	    if (defined $h_removedSample{$dataset}{$sample}) { next; }
	    my ($value, $time);
	    for my $header (sort keys %{$$rh_pheno{$dataset}{$sample}}) {
		# Keep value
		if ($header =~m/^E.$event$/) {
		    $value = $$rh_pheno{$dataset}{$sample}{$header};
		    # Remove Unwanted Samples
		    if (("$value" eq "NA") or ("$value" eq "undef") or !(defined $value)) {
			$s_log .= "$dataset\t$sample\tCensored\tE.$event = NA\n";
			last;
		    }
		}
		# Keep time
		if ($header eq "T.$event") {
		    
		    $time = $$rh_pheno{$dataset}{$sample}{$header};
		    # Remove Unwanted Samples
		    if (("$time" eq "NA") or ("$time" eq "undef") or !(defined $time) ) {
			$s_log .= "$dataset\t$sample\tCensored\tT.$event = NA\n";
			last;
		    }
		    if ($value == 0) {
			# If Value is 0 and Time < CensorTimeEvent => Sample is Censored
			if ($time < $censorTimeEvent) {
			    $s_log .= "$dataset\t$sample\tCensored\t$time < $censorTimeEvent\n";
			}
			# Else Sample is Kept
			# Info is kept on sample for stratification
			else {
			    ($$rh_pheno{$dataset}{$sample}{"E.$event"}, $h_condition{$dataset}{$sample}) = (0, 0);
			    $s_log .= "$dataset\t$sample\tKept\t$time > $censorTimeEvent\n";
			}
		    }
		    if ($value == 1) {
			# If Value is 1 and Time > censorTimeEvent => Sample Changed, new Value is 0
			# Info is kept on sample for stratification
			if ($time > $censorTimeEvent) {
			    ($$rh_pheno{$dataset}{$sample}{"E.$event"}, $h_condition{$dataset}{$sample}) = (0, 0);
			    $s_log .= "$dataset\t$sample\tChanged\t$time > $censorTimeEvent\n";
			}
			# Else Sample Kept, Nothing is changed
			# Info is kept on sample for stratification
			else {
			    ($$rh_pheno{$dataset}{$sample}{"E.$event"}, $h_condition{$dataset}{$sample}) = (1, 1);
			    $s_log .= "$dataset\t$sample\tUnchanged\t$time < $censorTimeEvent\n";
			}
		    }
		    ($value, $time) = (undef, undef);
		}
	    }   
	}
	# Print information on Filtering
	print "# Filtering on Following Time\n";
	for my $dataset (sort keys %h_condition) {
	    print "Samples Kept            = " . (scalar keys %{$h_condition{$dataset}}) . "/" . (scalar keys %{$$rh_pheno{$dataset}}) . "\t($dataset)\n";
	}
    }
}

# Case No censor time defined: event is reported directly
elsif(defined $event) { 
    for my $dataset (sort keys %$rh_pheno) {
	for my $sample (sort keys %{$$rh_pheno{$dataset}}) {
	    if (defined $h_removedSample{$dataset}{$sample}) { next; }
	    my $value;
	    for my $header (sort keys %{$$rh_pheno{$dataset}{$sample}}) {
		# Keep value
		if ($header =~m/^E.$event$/) {
		    $value = $$rh_pheno{$dataset}{$sample}{$header};
		    if (("$value" eq "NA") or ("$value" eq "undef") or !(defined $value)) {
			$s_log .= "$dataset\t$sample\tCensored\tE.$event = NA\n";
			last;
		    }
		    elsif ($value == 0) {
			# If Value is 0
			# Info is kept on sample for stratification
			($$rh_pheno{$dataset}{$sample}{"E.$event"}, $h_condition{$dataset}{$sample}) = (0, 0);
			$s_log .= "$dataset\t$sample\tKept\n";
		    }
		    if ($value == 1) {
			# If Value is 1
			# Info is kept on sample for stratification
			($$rh_pheno{$dataset}{$sample}{"E.$event"}, $h_condition{$dataset}{$sample}) = (1, 1);
			$s_log .= "$dataset\t$sample\tKept\n";
		    }
		}
	    }
	}
    }
}

# Condition Mode
else {
    for my $dataset (sort keys %$rh_pheno) {
	for my $sample (sort keys %{$$rh_pheno{$dataset}}) {
	    if (defined $h_removedSample{$dataset}{$sample}) { next; }
	    my $value;
	    for my $header (sort keys %{$$rh_pheno{$dataset}{$sample}}) {
		if ($header eq $inputCondition) {
		    $value = $$rh_pheno{$dataset}{$sample}{$header};
		    
		    # Remove Unwanted Samples if data malformed 
		    if (($value eq "NA") or ($value eq "undef") or !(defined $value)) {
			$s_log .= "$dataset\t$sample\tCensored\t$inputCondition = NA\n";
			last;
		    }
		    
		    # Test if condition is equal to conditiontrue
		    if ($value eq $inputConditionTrue) {
			
			($$rh_pheno{$dataset}{$sample}{"$inputCondition"}, $h_condition{$dataset}{$sample}) = (1, 1);
			$s_log .= "$dataset\t$sample\tKept\t$inputCondition = $inputConditionTrue\n";
		    }
		    else {
			($$rh_pheno{$dataset}{$sample}{"$inputCondition"}, $h_condition{$dataset}{$sample}) = (0, 0);
			$s_log .= "$dataset\t$sample\tKept\t$inputCondition neq $inputConditionTrue\n";
			
		    }
		}
	    }
	}
    }
}

# Perform stratification on stratification item
my %h_dataset;
my %h_countSample;
my %h_countItem;
my %h_count3;
my @a_number;

foreach my $item (@a_stratificationItem) {
    for my $dataset (sort keys %h_condition) {
	for my $sample (sort keys %{$h_condition{$dataset}}) {
	    if (defined $h_removedSample{$dataset}{$sample}) { next; }
	    my $value = $$rh_pheno{$dataset}{$sample}{$item};
	    if ( (!defined $value) or ("$value" eq "NA") or ("$value" eq "undef") or ("$value" eq "") ) {
		$s_log .= "$dataset\t$sample\tRemoved\tE.$item = NA\n";
		$h_removedSample{$dataset}{$sample}=1;
	    }
	    else {
		push (@{$h_countSample{$dataset}{$sample}{'item'}},  $item);
		push (@{$h_countSample{$dataset}{$sample}{'value'}}, $value);
	    }
	}
    }
}

for my $dataset (sort keys %h_countSample) {
    for my $sample (sort keys %{$h_countSample{$dataset}}) {
	my $s_temp;
	for (my $i=0; $i < scalar @{$h_countSample{$dataset}{$sample}{'item'}}; $i++) {
	    $s_temp.= "$h_countSample{$dataset}{$sample}{'item'}[$i]$h_countSample{$dataset}{$sample}{'value'}[$i]";
	}
	if ( !defined $h_countItem{$dataset}{$s_temp} ) { $h_countItem{$dataset}{$s_temp} = 1; }
	else                                            { $h_countItem{$dataset}{$s_temp} ++; }
	$h_count3{$dataset}{$sample} = $s_temp;
    }
}

for my $dataset (sort keys %h_countItem) {
    for my $item (sort keys %{$h_countItem{$dataset}}) {
	push (@a_number, $h_countItem{$dataset}{$item});
    }
    
    my $min = Utils::min(@a_number);
    if ( $min <= 1 ) { die "Not enough Samples in $dataset for Stratification. Numbers are @a_number"; }
    
    if ( $min < $layer ) {
	my $temp = $min;
	my $fold = 1;
	while ($min < $layer) {
	    $min += $temp;
	    $fold ++;
	}
	print "There is not enough samples for $layer layer\n";
	print "Apply multiplying ($fold-fold) on $dataset\n";
	$h_dataset{$dataset}{'fold'} = $fold;
    }
    @a_number = ();
}

# Gather information on size of dataset
# $h_dataset{$dataset}{'size'} countains the size of the dataset
# $h_dataset{$dataset}{'used'} countains the number of sample used in the dataset for testing
# $h_dataset{$dataset}{'fold'} countains the number of fold for the dataset
my %h_count;
for my $dataset (sort keys %h_condition) {
    for my $item (sort keys %{$h_countItem{$dataset}}) {
	if (!defined $h_dataset{$dataset}{'fold'}) { $h_dataset{$dataset}{'fold'} = 1; }
	$h_count{$dataset}{$item}{'size'} = $h_dataset{$dataset}{'fold'}*($h_countItem{$dataset}{$item});
	$h_count{$dataset}{$item}{'left'} = $h_count{$dataset}{$item}{'size'};
	$h_count{$dataset}{$item}{'used'} = 0;
    }
}

my $headerLabel;
if(defined $event) {
    $headerLabel = "E.$event";
}
else {
    $headerLabel = "$inputCondition";
}

if ($n_random eq 0) {
    my %h_countLog;
    # If dataset's size isn't large enough, it is doubled, tripled...
    # So each samples is used as many times as any other
    for (my $i=0; $i < $layer; $i++) {
	# Make directory for learning-data (data), and for testing-data (test)
	mkdir "$outputDirectory/run-".sprintf("%03d", $i+1)."-train";
	mkdir "$outputDirectory/run-".sprintf("%03d", $i+1)."-test";
	for my $dataset (sort keys %h_count3) {
	    
	    my $s_trainFile = "# ID\t$headerLabel\n";
	    my $s_testFile  = "# ID\t$headerLabel\n";
	    
	    my @a_item = ();
	    for my $sample (sort keys %{$h_count3{$dataset}}) { Utils::in_array($h_count3{$dataset}{$sample},\@a_item) ? next : push(@a_item, $h_count3{$dataset}{$sample}); }
	    # 		print "DATASET : (".sprintf("%03d", $i+1).") $dataset\tLOOP : $i\n";
	    my %h_test   = ();
	    my %h_sample = ();
	    foreach my $item (sort @a_item) {
		# 			print "\tITEM : $item\t";
		# 			print "$dataset\t$item\n";
		if ($h_count{$dataset}{$item}{'left'}%($layer-$i) == 0) {
		    $h_count{$dataset}{$item}{'used'} += ($h_count{$dataset}{$item}{'left'})/($layer-$i);
		}
		else {
		    $h_count{$dataset}{$item}{'used'} += (sprintf ("%1d",$h_count{$dataset}{$item}{'left'}/($layer-$i))+1);
		}
		# 			print "SIZE :".$h_count{$dataset}{$item}{'size'}."\tLEFT :".$h_count{$dataset}{$item}{'left'};
		# 			print "\t(*".$h_dataset{$dataset}{'fold'}.")USED : $h_count{$dataset}{$item}{'used'}\n";
		my $count = 0;
		for (my $fold = 0; $fold < $h_dataset{$dataset}{'fold'}; $fold++) {
		    foreach my $sample (sort keys %{$h_count3{$dataset}}) {
			if ($h_count3{$dataset}{$sample} eq "$item") {
			    if ( $count >= ($h_count{$dataset}{$item}{'size'}-$h_count{$dataset}{$item}{'left'}) and $count < $h_count{$dataset}{$item}{'used'}) {
				$h_test{$sample}=1;
			    }
			    $count++;
			}
		    }
		}
		if ($h_count{$dataset}{$item}{'left'}%($layer-$i) == 0) {
		    $h_count{$dataset}{$item}{'left'} -= ($h_count{$dataset}{$item}{'left'})/($layer-$i);
		}
		else {
		    $h_count{$dataset}{$item}{'left'} -= (sprintf ("%1d",$h_count{$dataset}{$item}{'left'}/($layer-$i))+1);
		}
	    }
	    for my $sample (sort keys %{$h_condition{$dataset}}) {
		if ( defined $h_test{$sample} ) {
		    if ( !defined $h_sample{$sample} ) { $s_testFile .= "$sample\t$h_condition{$dataset}{$sample}\n"; }
		    if ( !defined $h_sample{$sample} ) { if ( !defined $h_countLog{'test'}{sprintf("%03d", $i+1)}{$dataset}{$h_count3{$dataset}{$sample}} ) {$h_countLog{'test'}{sprintf("%03d", $i+1)}{$dataset}{$h_count3{$dataset}{$sample}} = 1 ; } else {$h_countLog{'test'}{sprintf("%03d", $i+1)}{$dataset}{$h_count3{$dataset}{$sample}}++ ; } }
		    
		    
		    
		    $h_sample{$sample} = 1;
		}
		else {
		    if ( !defined $h_sample{$sample})  { $s_trainFile .= "$sample\t$h_condition{$dataset}{$sample}\n"; }
		    if ( !defined $h_sample{$sample} ) { if ( !defined $h_countLog{'train'}{sprintf("%03d", $i+1)}{$dataset}{$h_count3{$dataset}{$sample}} ) {$h_countLog{'train'}{sprintf("%03d", $i+1)}{$dataset}{$h_count3{$dataset}{$sample}} = 1 ; } else {$h_countLog{'train'}{sprintf("%03d", $i+1)}{$dataset}{$h_count3{$dataset}{$sample}}++ ; } }
		    
		    $h_sample{$sample} = 1;
		}
	    }
	    # Printing train-file
	    my $file = "$outputDirectory/run-".sprintf("%03d", $i+1)."-train/$dataset.cond";
	    open (CONDITION_TRAIN_FILE, ">$file") or die "Cannot open $file : $!";
	    print CONDITION_TRAIN_FILE $s_trainFile;
	    close CONDITION_TRAIN_FILE;
	    
	    # Printing test-file
	    $file = "$outputDirectory/run-".sprintf("%03d", $i+1)."-test/$dataset.cond";
	    open (CONDITION_TEST_FILE, ">$file") or die "Cannot open $file : $!";
	    print CONDITION_TEST_FILE $s_testFile;
	    close CONDITION_TEST_FILE;
	}
    }
    for my $type (sort keys %h_countLog ) {
	for my $layer (sort keys %{$h_countLog{$type}} ) {
	    for my $dataset (sort keys %{$h_countLog{$type}{$layer}} ) {
		for my $condition (sort keys %{$h_countLog{$type}{$layer}{$dataset}}) {
		    if ( $stratification == 1 ) {
			$s_log .= "$dataset\tcondition = $condition\tnumber of samples = $h_countLog{$type}{$layer}{$dataset}{$condition} \n";
		    }
		    else {
			my $total = $h_countLog{'train'}{$layer}{$dataset}{$condition}+$h_countLog{'test'}{$layer}{$dataset}{$condition};
			$s_log .= "layer = $layer\t$type\t$dataset\tcondition = $condition\tnumber of samples = $h_countLog{$type}{$layer}{$dataset}{$condition} | $total\n";
		    }
		}
	    }
	}
    }
}
else {
    my %h_countLog;
    # Make directory for random, and for true conditions
    for my $dataset (sort keys %h_count3) {
	
	mkdir "$outputDirectory/$dataset-random";
	mkdir "$outputDirectory/$dataset-true";

	for (my $i=0; $i < $n_random; $i++) {
	    my $s_randomFile = "# ID\t$headerLabel\n";
	    
	    my $randomCondition = randomizeConditions($h_condition{$dataset});
	    
	    foreach my $sample (sort keys %$randomCondition) {
		$s_randomFile .= "$sample\t$randomCondition->{$sample}\n";
	    }
	    # Printing random-file
	    my $file = "$outputDirectory/$dataset-random/$dataset-$i.cond";
	    open (CONDITION_RANDOM_FILE, ">$file") or die "Cannot open $file : $!";
	    print CONDITION_RANDOM_FILE $s_randomFile;
	    close CONDITION_RANDOM_FILE;
	}
	my $s_trueFile = "# ID\t$headerLabel\n";
	foreach my $sample (sort keys %{$h_condition{$dataset}}) { $s_trueFile .= "$sample\t$h_condition{$dataset}{$sample}\n"; }
	# Printing true-file
	my $file = "$outputDirectory/$dataset-true/$dataset.cond";
	open (CONDITION_TRUE_FILE, ">$file") or die "Cannot open $file : $!";
	print CONDITION_TRUE_FILE $s_trueFile;
	close CONDITION_TRUE_FILE;
    }
    for my $type (sort keys %h_countLog ) {
	for my $layer (sort keys %{$h_countLog{$type}} ) {
	    for my $dataset (sort keys %{$h_countLog{$type}{$layer}} ) {
		for my $condition (sort keys %{$h_countLog{$type}{$layer}{$dataset}}) {
		    $s_log .= "$dataset\tcondition = $condition\tnumber of samples = $h_countLog{$type}{$layer}{$dataset}{$condition} \n";
		}
	    }
	}
    }
}


# Printing log-file
my $baseName = basename $inputDirectory;
my $logFile = "$outputDirectory/$baseName.log";
open (LOG, ">$logFile") or die "Cannot open $logFile : $!";
print LOG $s_log;
close LOG;

sub readPhenoFile {
    my $ra_inputFile = shift;
    my %h_pheno;
    foreach my $file (@$ra_inputFile){
	my @a_temp   = split (/\//, $file);
	my $fileTemp = $a_temp[scalar @a_temp -1];
	@a_temp      = split (/\./, $fileTemp);
	my $dataset  = $a_temp[0];
        
    open(PHENO_FILE, "$file") || die "Cannot open $file : $!";
	my @a_row = <PHENO_FILE>; 
	close PHENO_FILE;
    
    foreach (@a_row) {
         $_ =~ s/[\r\n]+$//;
    }
	my @a_header = split (/\t/, $a_row[0]);
	for ( my $j = 1; $j < (scalar @a_header); $j++) {
	    for ( my $k = 1; $k < (scalar @a_row); $k++) {
		my @a_item = split (/\t/, $a_row[$k]);
		$h_pheno{$dataset}{$a_item[0]}{$a_header[$j]}=$a_item[$j];
	    }
	}
    }
    return \%h_pheno;
}

sub randomizeConditions {
    my $rh_condition = shift;
    
    my @a_randomSample;
    my %h_randomCondition;
    
    foreach my $sample(sort keys %$rh_condition) {
	push(@a_randomSample, $sample);
	$h_randomCondition{$sample} = $$rh_condition{$sample};
    }
    
    my $n = 100 * (scalar keys %$rh_condition);
    
    while ($n>1) {
	my $k1 = rand() * (scalar keys %$rh_condition);
	my $k2 = rand() * (scalar keys %$rh_condition);
	
	$n--;
	
	my $tmp = $a_randomSample[$k1];
	$a_randomSample[$k1] = $a_randomSample[$k2];
	$a_randomSample[$k2] = $tmp;
	
	$tmp = $h_randomCondition{$a_randomSample[$k1]};
	$h_randomCondition{$a_randomSample[$k1]} = $h_randomCondition{$a_randomSample[$k2]};
	$h_randomCondition{$a_randomSample[$k2]} = $tmp;
    }
    return \%h_randomCondition;
}

