#!/usr/bin/perl
#
# HTS-Net project
#
# Copyright Maxime Garcia, Ghislain Bidaut, Claire Rioualen, Quentin Da Costa
# (2009-2017)
#
# maxime.u.garcia@gmail.com
# ghislain.bidaut@inserm.fr
# claire.rioualen@inserm.fr
# quentin.da-costa@inserm.fr
#
# http://iti.sourceforge.net
#
# This software is a computer program whose purpose is to superimpose
# interactome over several gene expression datasets to find
# discriminative subnetworks predicting metastatic relapse in breast
# cancer.
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty and the software's author, the holder of the
# economic rights, and the successive licensors have only limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading, using, modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean that it is complicated to manipulate, and that also
# therefore means that it is reserved for developers and experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and, more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod

=head1 Name
generateHTMLReport.pl

= head1 Description
This script generates final HTML report for an HTS Net run

=head1 Synopsis

=head2 Input directories

    --interactomeRunDir|-intDir| -id  <Directory containing interactome-derived subnetworks>
    --regulomeRunDir | -regDir | -rd  <Directory containing interactome-derived subnetwork> (At least one of these two must be defined).
    --metaSubnetsDir| -md             <Directory containing integrated subnetworks (metasubnetworks)>
    --edgeAttributeFile | -eda        <Edge Attribute file>

=head2 Required

    -htsnet | --htsnetPath            <HTS-Net Path>
    -graph | --graphVizPath           <graphViz Executable Path>

    -s | --studies                    <List of studies>
    -rph | --resultsPageHeading       <HTML report Page Heading>
    -project | --projectDescription   <Project description>
    -od | --outputDir                 <output Directory>

    -dd | --inputDir                 <Input data directory>
    
    -gif | --geneInfoFile            <gene_info file>
   
    
=head2 Optional
   --help | --usage | -h | -u
    -go | --GOEnrichment             <Presence of GO enrichment data>


=head1 Authors
Claire Rioualen     claire.rioualen@inserm.fr
Ghislain Bidaut     ghislain.bidaut@inserm.fr
Quentin Da Costa    quentin.da-costa@inserm.fr
Maxime Garcia       maxime.u.garcia@gmail.com

=head1 Usage

~/bin/htsnet-project/pipeline/generateHTMLReport.pl -studies aldh1-run-large-inhi aldh1-run-large-stim aldh1-run-standard-inhi aldh1-run-standard-stim -project Aldh1-Score-CSC -od html-output-2 -intDir runs-int/out* -regDir runs-reg/out* -eda ./network-data/edge_attributes-eda.txt -id /home/data/htsnet/travail/project-CSC/analysis-screen-aldh1/Aldh1Score/ -cd /home/data/htsnet/travail/project-CSC/analysis-screen-aldh1/Aldh1Score/ -htsnet htsnet-project/ -graph /home/progs/visualisation/graphviz-2.30.1/bin/ -gif /home/data/annotations/stable/ncbi/Homo_sapiens.gene_info.gz -go

=cut

# Modules

use strict;
use warnings;

use Getopt::Long;
use Cwd;

use Subnetworks;
use Metasubnet;
use HTS;
use Utils;
use Network;
use Web;


## Arguments management

my (
    $help,
    @a_d_studies,
    $project,
    @a_f_desc,
    $d_output,

    $d_int,
    $d_reg,
    $f_eda,
    $d_metaSubnets,

    $d_inputData,
    
    $d_htsnet,
    $d_rscript,
    $graphVizPath,

    $resHeading,

    $geneInfo,

    $noImage,
    $noHeatmap,
    $sortByScore,
    $GO,

    $outInCurrDir,
    $zip,
    @a_int,
    @a_reg,
    @a_meta,
    @a_f_intGO,
    @a_f_regGO,
    @a_f_metaGO,
    $f_exp
    );


GetOptions(
    'help|usage|h|u'                => \$help,
    'studies|s=s{,}'                => \@a_d_studies,
    'resultsPageHeading|rph=s'      => \$resHeading,
    'projectName|project=s'         => \$project,
    'descriptionFiles|df=s{,}'      => \@a_f_desc,
    'outputDir|od=s'                => \$d_output,

    'interactomeRunDir|intDir|id=s{,}' => \@a_int,
    'regulomeRunDir|regDir|rd=s{,}'    => \@a_reg,
    'metaSubnetsDir|metaDir|md=s{,}'   => \@a_meta,
    'edgeAttributeFile|eda=s'       => \$f_eda,

    'inputDir|dd=s'                 => \$d_inputData,
        
    'htsnetPath|htsnet=s'           => \$d_htsnet,
    'rscriptDir|rsd=s'              => \$d_rscript,
    'graphVizPath|graph=s'          => \$graphVizPath,

    'geneInfoFile|gif=s'            => \$geneInfo,
    
    'GOEnrichment|go'               => \$GO,
    'noImage|noPict'                => \$noImage,
    'noHeatmap'                     => \$noHeatmap,
    'sortByScore|sbs'               => \$sortByScore,

    'singleOutput|so'               => \$outInCurrDir,
    'zip'                           => \$zip,
    'iFile|if=s'                    => \$f_exp,
    'intGOFiles=s{,}'               => \@a_f_intGO,
    'regGOFiles=s{,}'               => \@a_f_regGO,
    'metaGOFiles=s{,}'              => \@a_f_metaGO
);

if(defined $d_inputData){
    
    if($d_inputData !~ m/\/$/) {$d_inputData .= "/";}
}
else{
    $d_inputData = "" if(defined $f_exp);
}

if(defined $d_rscript) {
    if($d_rscript !~ m/\/$/){$d_rscript .= "/";}
}
else {
    $d_rscript = "";
}

if(defined $outInCurrDir       and
   defined $project            and
   defined $d_output           and
   (scalar @a_int              or
    scalar @a_reg)             and
   (scalar @a_f_intGO          or
    scalar @a_f_regGO)         and
   defined $f_eda              and
   defined $d_inputData        and
   defined $d_htsnet           and
   defined $graphVizPath       and
   defined $geneInfo){

    @a_d_studies = ("");
    $d_int = "" if(scalar @a_int);
    $d_reg = "" if(scalar @a_reg);
    $d_metaSubnets = "" if(scalar @a_meta);
    $GO = 1;
}

elsif(scalar  @a_d_studies      and
      defined $project          and
      defined $d_output         and
      (scalar @a_int            or
       scalar @a_reg)           and
      defined $f_eda            and
      defined $d_inputData      and
      defined $d_htsnet         and
      defined $graphVizPath     and
      defined $geneInfo){

    
    for(my $i = 0; $i < scalar(@a_d_studies); $i++){
        if($a_d_studies[$i] !~ m/\/$/) {$a_d_studies[$i] .= "/";}
    }
    
    $d_int = $a_int[0];
    $d_reg = $a_reg[0];
    $d_metaSubnets = $a_meta[0];
    
    if(defined $d_int and $d_int !~ m/\/$/) {$d_int .= "/";}
    if(defined $d_reg and $d_reg !~ m/\/$/) {$d_reg .= "/";}
    if(defined $d_metaSubnets and $d_metaSubnets !~ m/\/$/) {$d_metaSubnets .= "/";}

}

else{

print "
NAME
$0

DESCRIPTION
This script generates final HTML report for an HTS-Net run

PARAMETERS

Input directories

    --interactomeRunDir|-intDir| -id  <Directory containing interactome-derived subnetworks>
    --regulomeRunDir | -regDir | -rd  <Directory containing interactome-derived subnetwork> (At least one of these two must be defined).
    --metaSubnetsDir | -md            <Directory containing integrated subnetworks (metasubnetworks)>
    --edgeAttributeFile | -eda        <Edge Attribute file>

Required

    -htsnet | --htsnetPath                  <HTS-Net Path>
-rsd | --rscriptDir <Rscript directory>
    -graph | --graphVizPath           <graphViz Executable Path>

    -s | --studies                    <List of studies>
    -project | --projectDescription   <Project description>
    -od | --outputDir                 <output Directory>

    -dd | --dataDir                  <Input data directory>
    
    -gif | --geneInfoFile            <gene_info file>

Optional
   --help | --usage | -h | -u
    -go | --GOEnrichment             <Presence of GO enrichment data>
    -rph | --resultsPageHeading      <HTML report Page Heading (default = \"Results\">

AUTHORS
Claire Rioualen     claire.rioualen\@inserm.fr
Ghislain Bidaut     ghislain.bidaut\@inserm.fr
Quentin Da Costa    quentin.da-costa\@inserm.fr
Maxime Garcia       maxime.u.garcia\@gmail.com

EXAMPLE

~/bin/htsnet-project/pipeline/generateHTMLReport.pl -studies aldh1-run-large-inhi aldh1-run-large-stim aldh1-run-standard-inhi aldh1-run-standard-stim -project Aldh1-Score-CSC -od html-output-2 -intDir runs-int/out* -regDir runs-reg/out* -eda ./network-data/edge_attributes-eda.txt -dd /home/data/htsnet/travail/project-CSC/analysis-screen-aldh1/Aldh1Score/ -cd /home/data/htsnet/travail/project-CSC/analysis-screen-aldh1/Aldh1Score/  -htsnet htsnet-project/ -graph /home/progs/visualisation/graphviz-2.30.1/bin/ -gif /home/data/annotations/stable/ncbi/Homo_sapiens.gene_info.gz  -go

Please report any problem at <ghislain.bidaut\@inserm.fr>

";
    exit(0);
}


$resHeading = "Results" unless(defined $resHeading);


print "
Studies List:              @a_d_studies
Results Page Heading:      $resHeading
Project Description:       $project

Output Directory:          $d_output
";
print "Interactome Run Dir:       $d_int\n" if(defined $d_int);
print "Regulome Run Dir:          $d_reg\n" if(defined $d_reg);
print "MetaSubnets Dir:           $d_metaSubnets\n" if(defined $d_metaSubnets);
print "edgeAttributeFile:         $f_eda
";

if(defined $f_exp) {
    print "Input File:                 $f_exp\n";
}
else {
    print  "Input Directory:           $d_inputData\n";
}

print "
HTS-Net Path:              $d_htsnet
GraphVizPath:              $graphVizPath

GeneInfoFile:              $geneInfo

";
print "GOEnrichment:              Yes\n\n\n" if(defined $GO);


my $oldTime = time;

my ($int, $reg, $itg, $class);
# Determination of analysis type (interactome, regulome or both)
if(defined $d_int && defined $d_reg){
    $int = $reg = $itg = 1;
    $class = "itg";
    die "No metasubnet directory defined. Please see usage information." unless(defined $d_metaSubnets);
}
elsif(defined $d_int){
    $int = 1;
    $class = "int";
}
elsif(defined $d_reg){
    $reg = 1;
    $class = "reg";
}

#print "Analysis type: $class\n\n";

mkdir($d_output);
$d_output =~ s/\/+$//;
#print "Importing KickStart...";
system "rsync -a --exclude=\".*\" $d_htsnet/KickStart/ $d_output/KickStart/";
#print "\n";

# Read Annotations

my ($rh_geneIDtoSymbol, $rh_symbolToGeneID) = Utils::readGeneInfo($geneInfo);


# Read Input Data

my $rh_datasets = HTS::readInputDataDir($d_inputData, $f_exp);

if(defined $noHeatmap){
    for my $ds(keys %$rh_datasets){
        $$rh_datasets{$ds}{'noHeatmap'} = 1;
    }
}


# Process each study
my(@a_studies, $desc);
my $i = 0;
foreach my $d_studyInput (@a_d_studies){

    my $study = $d_studyInput;
    if(defined $outInCurrDir){
        $study = $project;
    }
    else{
        $study =~ s/\/+$//;
        my @a_study = split('\/', $study);
        $study      = $a_study[-1];
    }

    push(@a_studies, $study);
    #print "\n******************* Current study: $study\n";

    my $d_studyOutput = "$d_output/$study";
    mkdir $d_studyOutput;

    # Read subnets and write in corresponding output dir
    my ($rh_int_subnets, $rh_int_eda, $rh_int_subnets_nwa,
        $rh_int_noa, $rh_reg_subnets, $rh_reg_eda,
        $rh_reg_subnets_nwa, $rh_reg_noa, $rh_meta_nwa,
        $rh_meta_noa, $rh_meta_eda, $rh_subnets);

    if (defined $int){
        my $netDir = "$d_output/$study/int";
        mkdir $netDir;
        my $downloadDir = "$d_output/$study/int/downloads";
        mkdir $downloadDir;

        #print "Read interactome subnet data in $d_studyInput$d_int\n";

        my @a_files;
        if(defined $outInCurrDir){
            @a_files = @a_int;
        }
        else{
            @a_files = Utils::readInputDir("$d_studyInput$d_int");
        }

        my @a_snw = grep(/.nnf$/, @a_files);
        my @a_noa = grep(/noa.txt$/, @a_files);
        my @a_nwa = grep(/nwa.txt$/, @a_files);

        ($rh_int_subnets_nwa, $rh_int_noa, $rh_int_eda) = Subnetworks::readSubnets("$d_studyInput$d_int$a_snw[0]", "$d_studyInput$d_int$a_nwa[0]", "$d_studyInput$d_int$a_noa[0]", $f_eda);

        Subnetworks::writeSubnets($rh_int_subnets_nwa, $rh_int_noa, $rh_int_eda,
                             "$downloadDir/int-subnetworks.nnf",
                             "$downloadDir/int-subnetworks_attributes-nwa.txt",
                             "$downloadDir/int-genes_attributes-noa.txt");

        unless(defined $itg){
            system "perl -I $d_htsnet/pipeline $d_htsnet/parsing/extractSubnetwork.pl -nnf $downloadDir/int-subnetworks.nnf -od $downloadDir/subnetworks/";

            my $cwd = getcwd;
            chdir $downloadDir;
            system "zip -rq int-subnetworks.zip subnetworks/
                    cp int-subnetworks_attributes-nwa.txt int-genes_attributes-noa.txt subnetworks/
                    zip -rq int-results.zip subnetworks/
                    rm -r subnetworks/ int-subnetworks.nnf";
            chdir $cwd;
        }
    }
    if (defined $reg){
        my $netDir = "$d_output/$study/reg";
        mkdir $netDir;
        my $downloadDir = "$d_output/$study/reg/downloads";
        mkdir $downloadDir;

        #print "Read regulome subnet data in $d_studyInput$d_reg\n";

        my @a_files;
        if(defined $outInCurrDir){
            @a_files = @a_reg;
        }
        else{
            @a_files = Utils::readInputDir("$d_studyInput$d_reg");
        }

        my @a_snw = grep(/.nnf$/, @a_files);
        my @a_noa = grep(/noa.txt$/, @a_files);
        my @a_nwa = grep(/nwa.txt$/, @a_files);

        ($rh_reg_subnets_nwa, $rh_reg_noa, $rh_reg_eda) = Subnetworks::readSubnets("$d_studyInput$d_reg$a_snw[0]", "$d_studyInput$d_reg$a_nwa[0]", "$d_studyInput$d_reg$a_noa[0]", $f_eda);

        Subnetworks::writeSubnets($rh_reg_subnets_nwa, $rh_reg_noa, $rh_reg_eda,
                             "$downloadDir/reg-subnetworks.nnf",
                             "$downloadDir/reg-subnetworks_attributes-nwa.txt",
                             "$downloadDir/reg-genes_attributes-noa.txt");

        unless(defined $itg){
            system "perl -I $d_htsnet/pipeline $d_htsnet/parsing/extractSubnetwork.pl -nnf $downloadDir/reg-subnetworks.nnf -od $downloadDir/subnetworks/";

            my $cwd = getcwd;
            chdir $downloadDir;
            system "zip -rq reg-subnetworks.zip subnetworks/
                    cp reg-subnetworks_attributes-nwa.txt reg-genes_attributes-noa.txt subnetworks/
                    zip -rq reg-results.zip subnetworks/
                    rm -r subnetworks/ reg-subnetworks.nnf";
            chdir $cwd;
        }
    }
    if (defined $itg){
        my $netDir = "$d_output/$study/itg";
        mkdir $netDir;
        my $downloadDir = "$d_output/$study/itg/downloads";
        mkdir $downloadDir;

        my @a_files;
        if(defined $outInCurrDir){
            @a_files = @a_meta;
        }
        else{
            @a_files = Utils::readInputDir("$d_studyInput$d_metaSubnets");
        }

        my @a_metaNwa = grep(/metasubnets.txt/, @a_files);
        my @a_noa     = grep(/meta_nodes.txt/,  @a_files);
        my @a_metaSnw = grep(/meta_edges.nnf/,  @a_files);
        my @a_nwa     = grep(/^subnets.txt/,    @a_files);

        ($rh_meta_nwa, $rh_meta_noa,
         $rh_meta_eda, $rh_subnets) = Metasubnet::readMetasubnets("$d_studyInput$d_metaSubnets$a_metaNwa[0]",
                                                                  "$d_studyInput$d_metaSubnets$a_noa[0]",
                                                                  "$d_studyInput$d_metaSubnets$a_metaSnw[0]",
                                                                  "$d_studyInput$d_metaSubnets$a_nwa[0]",
                                                                  $f_eda);
	Metasubnet::writeMetasubnets($downloadDir, $rh_meta_nwa, $rh_meta_noa, $rh_meta_eda, $rh_subnets);

        system "perl -I $d_htsnet/pipeline  $d_htsnet/pipeline/extractSubnetwork.pl -nnf $downloadDir/meta_edges.nnf -od $downloadDir/metasubnetworks/";

        my $cwd = getcwd;
        chdir $downloadDir;
        system "zip -rq metasubnetworks.zip metasubnetworks/
                cp metasubnets.txt meta_nodes.txt subnets.txt metasubnetworks/
                zip -rq itg-results.zip metasubnetworks/
                rm -r metasubnetworks/ meta_edges.nnf";
        chdir $cwd;
    }

    # GO Terms enrichment processing

    my($rh_int_go, $rh_reg_go, $rh_meta_go);

    if(defined $GO && defined $d_int){
        my(@a_files, $d_GO);
        if(defined $outInCurrDir){
            @a_files = @a_f_intGO;
            $d_GO = "";
        }
        else{
            $d_GO = "$d_studyInput$d_int" . "GO/";
            @a_files = Utils::readInputDir($d_GO);
        }

        my @a_int_noa = grep(/noa.txt$/, @a_files);
        my @a_int_nwa = grep(/nwa.txt$/, @a_files);

        $rh_int_go = Network::readGOterms("$d_GO$a_int_noa[0]",
                                           "$d_GO$a_int_nwa[0]",
                                           $rh_int_subnets_nwa, $rh_int_noa);
    }
    if(defined $GO && defined $d_reg){
        my(@a_files, $d_GO);
        if(defined $outInCurrDir){
            @a_files = @a_f_regGO;
        }
        else{
            $d_GO = "$d_studyInput$d_reg" . "GO/";
            @a_files = Utils::readInputDir($d_GO);
        }

        my @a_reg_noa = grep(/noa.txt$/, @a_files);
        my @a_reg_nwa = grep(/nwa.txt$/, @a_files);

        $rh_reg_go = Network::readGOterms("$d_GO$a_reg_noa[0]",
                                           "$d_GO$a_reg_nwa[0]",
                                           $rh_reg_subnets_nwa, $rh_reg_noa);
    }
    if(defined $GO && defined $d_metaSubnets){
        my(@a_files, $d_GO);
        if(defined $outInCurrDir){
            @a_files = @a_f_metaGO;
        }
        else{
            $d_GO = "$d_studyInput$d_metaSubnets" . "GO/";
            @a_files = Utils::readInputDir($d_GO);
        }

        my @a_meta_noa = grep(/noa.txt$/, @a_files);
        my @a_meta_nwa = grep(/nwa.txt$/, @a_files);

        $rh_meta_go = Network::readGOterms("$d_GO$a_meta_noa[0]",
                                           "$d_GO$a_meta_nwa[0]",
                                           $rh_meta_nwa, $rh_meta_noa);
    }

    if(scalar(@a_f_desc) == 1 and $i == 0){
        open(DESC, $a_f_desc[0]) || die "Cannot open $a_f_desc[0]:\n$!";
        while(<DESC>){
            $_ =~ s/[\r\n]+$/<br>\n/;
            $desc .= $_;
        }
        close DESC;
    }
    elsif(scalar(@a_f_desc) == scalar(@a_d_studies)){
        $desc = "";
        open(DESC, $a_f_desc[$i]) || die "Cannot open $a_f_desc[$i]:\n$!";
        while(<DESC>){
            $_ =~ s/[\r\n]+$/<br>\n/;
            $desc .= $_;
        }
        close DESC;
    }
    $i++;

    # Write HTML report

    Web::homePage($d_output, $study, $class, $project, $resHeading, $desc);

    # Writing analysis content

    if (defined $int && defined $reg) {
        #print "\n\nRunning HTML generating program for study $study...\n";

        if(scalar $rh_int_subnets_nwa >0) {
            Web::writeCategory($d_output, $d_htsnet, $d_rscript, $graphVizPath, $rh_datasets, $project, $study,
                               "int", $rh_int_subnets_nwa, $rh_int_subnets_nwa, $rh_int_noa,  $rh_int_eda,
                               $rh_int_go, $class, $sortByScore, $noHeatmap);
        }
        else {
            warn "No int subnetworks found - skipped";
        }

        if(scalar $rh_reg_subnets_nwa >0) {
            Web::writeCategory($d_output, $d_htsnet, $d_rscript, $graphVizPath, $rh_datasets, $project, $study,
                               "reg", $rh_reg_subnets_nwa, $rh_reg_subnets_nwa, $rh_reg_noa,   $rh_reg_eda,
                               $rh_reg_go, $class, $sortByScore, $noHeatmap);
        }
        else {
            warn "No reg subnetworks found - skipped";
        }

        if(scalar  $rh_subnets >0) {
            Web::writeCategory($d_output, $d_htsnet, $d_rscript, $graphVizPath, $rh_datasets, $project, $study,
                               "itg", $rh_subnets, $rh_meta_nwa, $rh_meta_noa,  $rh_meta_eda,
                               $rh_meta_go,  $class);
        }
        else {
            warn "No itg subnetworks found - skipped";
        }
    }
    elsif (defined $int){
        #print "\n\nRunning HTML generating program for $class...\n";
        if(scalar $rh_int_subnets_nwa >0) {
            Web::writeCategory($d_output, $d_htsnet, $d_rscript, $graphVizPath, $rh_datasets, $project, $study,
                               "int", $rh_int_subnets_nwa, $rh_int_subnets_nwa, $rh_int_noa, $rh_int_eda,
                               $rh_int_go, $class, $sortByScore, $noHeatmap, $resHeading, $noImage);
        }
        else {
            warn "No int subnetworks found - skipped";
        }
    }
    elsif (defined $reg) {
        #print "\n\nRunning HTML generating program for $class...\n";
        if(scalar $rh_reg_subnets_nwa >0) {
            Web::writeCategory($d_output,  $d_htsnet, $d_rscript, $graphVizPath, $rh_datasets, $project, $study,
                               "reg", $rh_reg_subnets_nwa, $rh_reg_subnets_nwa, $rh_reg_noa,   $rh_reg_eda,
                               $rh_reg_go, $class, $sortByScore, $noHeatmap, $resHeading, $noImage);
        }
        else {
            warn "No reg subnetworks found - skipped";
        }
    }

    # Writing Download page for $study

    Web::downloadPage($d_output, $study, $class, $project, $resHeading, $desc);

}

Web::indexPage($d_output, \@a_studies, $project, $resHeading);

if(defined $zip) {
    system "zip -rq  $d_output $d_output";
    system "rm -r $d_output";
}

# Time measurement system - ENDING  #TMS
my $date    = localtime(time);
my $time    = time;
my $period  = $time - $oldTime;

# Printing exit message
#print "\n$date: process complete\n";
#print "It took $period seconds to run\n";
#print "\n$0 done.\n\n"

