#!/usr/bin/perl
#
# HTS-Net project
#
# Copyright Maxime Garcia, Ghislain Bidaut, Claire Rioualen, Quentin Da Costa
# (2009-2017)
#
# maxime.u.garcia@gmail.com
# ghislain.bidaut@inserm.fr
# claire.rioualen@inserm.fr
# quentin.da-costa@inserm.fr
#
# http://iti.sourceforge.net
#
# This software is a computer program whose purpose is to superimpose
# interactome over several gene expression datasets to find
# discriminative subnetworks predicting metastatic relapse in breast
# cancer.
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty and the software's author, the holder of the
# economic rights, and the successive licensors have only limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading, using, modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean that it is complicated to manipulate, and that also
# therefore means that it is reserved for developers and experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and, more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod

=head1 Name
integrateInteractomeRegulomeSubnetworks.pl

=head1 Description
Integrate regulome and interactome subnetworks by clustering subnetworks sharing common nodes.
Obtained metasubnetworks are stored in nested network file along with attributes.

=head1 Parameters

=head2 Input files

    -od | --outputDir <Output directory>

    -innf | -interactomeNestedNetworkFile    <interactome-derived subnetworks file>
    -rnnf | -regulomeNestedNetworkFile       <regulome-derived subnetworks file>
    -inwa | -interactomeNetworkAttributeFile <interactome-derived subnetworks attributes file>
    -rnwa | -regulomeNetworkAttributeFile    <regulome-derived subnetworks attributes file>
    -inoa | -interactomeNodeAttributeFile    <interactome-derived subnetworks nodes attributes file>
    -rnoa | -regulomeNodeAttributeFile       <regulome-derived subnetworks nodes attributes file>
    -eda | -edgeAttributeFile                <edges attributes file>
    OR
    -id | --intDir <Output directory with interactome>
    -rd | --regDir <Output directory with regulome>

=head2 Required
    -th | --connectivityThreshold        <Connectivity threshold>
    -rnwf | --regulomeNetworkFile        <Regulome network file>
    -inwf | --interactomeNetworkFile     <Interactome network file>
    -gif | --geneInfoFile                <gene_info file>
    -gof | --gene2goFile                 <gene2go file>
    -taxon                               <taxon>
    -nwf | --networkFile                 <network file>

=head2 Optional
    --help | --usage | -u | -h

=head1 Usage
~/bin/iti-project/integrateSubnetworks.pl -innf ./runs-int/*.nnf -rnnf runs-reg/*.nnf -inwa runs-int/*nwa* -rnwa runs-reg/*nwa* -inoa runs-int/*noa* -rnoa runs-reg/*noa* -eda ./network-data/edge_attributes-eda.txt -od ./integration -th 2 -gif /home/data/annotations/stable/ncbi/Homo_sapiens.gene_info.gz -gof /home/data/annotations/stable/ncbi/Homo_sapiens.gene2go.gz -taxon 9606 -inwf network-data/interactome.txt -rnwf network-data/regulome-bidirectional\_interactions.txt

=head1 Authors
Claire Rioualen         claire.rioualen@inserm.fr
Ghislain Bidaut         ghislain.bidaut@inserm.fr
Quentin Da Costa        quentin.da-costa@inserm.fr

=cut

use strict;
use warnings;

use Subnetworks;
use Network;
use Metasubnet;
use Utils;

use Getopt::Long;

## Arguments management

my (
    $help,
    $f_int_nnf,
    $f_reg_nnf,
    $f_int_nwa,
    $f_reg_nwa,
    $f_int_noa,
    $f_reg_noa,
    $f_reg_eda,
    $f_regulomeNetwork,
    $f_interactomeNetwork,
    $f_gene2go,
    $taxonID,
    $d_output,
    $d_outputGO,
    $th,
    $gif,
    $d_int,
    $d_reg,
    $outInCurrDir
    );

GetOptions(
    'help|usage|u|h'                         => \$help,
    'interactomeNestedNetworkFile|innf=s'    => \$f_int_nnf,
    'regulomeNestedNetworkFile|rnnf=s'       => \$f_reg_nnf,
    'interactomeNetworkAttributeFile|inwa=s' => \$f_int_nwa,
    'regulomeNetworkAttributeFile|rnwa=s'    => \$f_reg_nwa,
    'interactomeNodeAttributeFile|inoa=s'    => \$f_int_noa,
    'regulomeNodeAttributeFile|rnoa=s'       => \$f_reg_noa,
    'intDir|id=s'                            => \$d_int,
    'regDir|rd=s'                            => \$d_reg,
    'edgeAttributeFile|eda=s'                => \$f_reg_eda,
    'gene2goFile|gof=s'                      => \$f_gene2go,
    'taxon=i'                                => \$taxonID,
    'regulomeNetworkFile|rnwf=s'             => \$f_regulomeNetwork,
    'interactomeNetworkFile|inwf=s'          => \$f_interactomeNetwork,
    'outputDir|od=s'                         => \$d_output,
    'connectivityThreshold|th=i'             => \$th,
    'geneInfoFile|gif=s'                     => \$gif,
    'singleOutput|so'                        => \$outInCurrDir
);

if(defined $d_int and defined $d_reg){
    $f_int_nnf = glob("$d_int/*.nnf");
    $f_reg_nnf = glob("$d_reg/*.nnf");
    $f_int_nwa = glob("$d_int/*nwa.txt");
    $f_reg_nwa = glob("$d_reg/*nwa.txt");
    $f_int_noa = glob("$d_int/*noa.txt");
    $f_reg_noa = glob("$d_reg/*noa.txt");
}

if(defined $outInCurrDir and
   defined $f_int_nnf and
   defined $f_reg_nnf and
   defined $f_int_nwa and
   defined $f_reg_nwa and
   defined $f_int_noa and
   defined $f_reg_noa and
   defined $f_reg_eda and
   defined $f_gene2go and
   defined $taxonID and
   defined $f_regulomeNetwork and
   defined $f_interactomeNetwork and
   defined $th and
   defined $gif) {

    $d_output = "";
    $d_outputGO = "";
}

elsif(defined $f_int_nnf and
      defined $f_reg_nnf and
      defined $f_int_nwa and
      defined $f_reg_nwa and
      defined $f_int_noa and
      defined $f_reg_noa and
      defined $f_reg_eda and
      defined $f_gene2go and
      defined $taxonID and
      defined $f_regulomeNetwork and
      defined $f_interactomeNetwork and
      defined $d_output and
      defined $th and
      defined $gif){

    if($d_output !~ m/\/$/) {$d_output .= "/";}
    mkdir $d_output;

    $d_outputGO = "$d_output/GO/";
    mkdir $d_outputGO;
}

else{

print "
NAME
$0

DESCRIPTION
- This script integrates regulome and interactome subnetworks by clustering subnetworks sharing common nodes.
- Obtained metasubnetworks are stored in nested network file along with attributes.

PARAMETERS

Input files

-od | --outputDir <Output directory>\n\tUse the given directory (full path) as output directory

-innf | -interactomeNestedNetworkFile <interactome-derived subnetworks file>
-rnnf | -regulomeNestedNetworkFile <regulome-derived subnetworks file>
-inwa | -interactomeNetworkAttributeFile <interactome-derived subnetworks attributes file>
-rnwa | -regulomeNetworkAttributeFile <regulome-derived subnetworks attributes file>
-inoa | -interactomeNodeAttributeFile <interactome-derived subnetworks node attributes file>
-rnoa | -regulomeNodeAttributeFile <regulome-derived subnetworks nodes attributes file>
-reda | -regulomeEdgeAttributeFile <regulome-derived subnetworks nodes attributes file>
OR
-id | --intDir <Output directory with interactome>
-rd | --regDir <Output directory with regulome>

Required

-th | --connectivityThreshold <Connectivity threshold>\n\tMinimum number of connections between interactome and regulome subnetworks
-rnwf | --regulomeNetworkFile  <Regulome network file>
-inwf | --interactomeNetworkFile <Interactome network file>
-gif | --geneInfoFile <gene_info file>\n\tFull path of local copy of gene_info file gzip-compressed (.gz) downloaded from NCBI
-gof | --gene2goFile <gene2go file>\n\tFull path of local copy of gene2go file gzip-compressed (.gz) downloaded from NCBI
-taxon <taxon> NCBI Taxon ID

EXAMPLE

$0 -innf ./runs-int/*.nnf -rnnf runs-reg/*.nnf -inwa runs-int/*nwa* -rnwa runs-reg/*nwa* -inoa runs-int/*noa* -rnoa runs-reg/*noa* -reda ./network-data/edge_attributes-eda.txt -od ./integration -th 2 -gif /home/data/annotations/stable/ncbi/Homo_sapiens.gene_info.gz -gof /home/data/annotations/stable/ncbi/Homo_sapiens.gene2go.gz -taxon 9606 -inwf network-data/interactome.txt -rnwf network-data/regulome-bidirectional\_interactions.txt

Please report any problem at <ghislain.bidaut\@inserm.fr>

";
    exit(0);
}


my ($rh_geneIDtoSymbol, $rh_geneSymbolToID) = Utils::readGeneInfo($gif);

###### Part 1 - Data import

my ($rh_int_nwa, $rh_int_noa, $rh_int_eda) = Subnetworks::readSubnets($f_int_nnf, $f_int_nwa, $f_int_noa, $f_reg_eda);
my ($rh_reg_nwa, $rh_reg_noa, $rh_reg_eda) = Subnetworks::readSubnets($f_reg_nnf, $f_reg_nwa, $f_reg_noa, $f_reg_eda);
my ($rh_trash1, $rh_trash2, $rh_tot_eda)   = Subnetworks::readSubnets(undef, undef, undef, $f_reg_eda);

# Retrieve \%($geneID => \%("$GO_ID - $GO_term" => 1)).
my $rh_gene2GOs = Utils::readGene2Go($f_gene2go, $taxonID);

# Retrieve global network
my $rh_regNetwork  = Utils::readHashFile($f_regulomeNetwork);
my $rh_intNetwork = Utils::readHashFile($f_interactomeNetwork);
my %h_network;

foreach my $k (keys %$rh_intNetwork) {
    $h_network{$k} = $$rh_intNetwork{$k};
}

foreach my $k (keys %$rh_regNetwork) {
    $h_network{$k} = $$rh_regNetwork{$k};
}

my ($totalGeneNb, $rh_GOccur) = Network::calcNetworkGlobalGOStatistics(\%h_network, $rh_gene2GOs);

print "\nInteractome subnetworks count = ". scalar keys(%{$rh_int_nwa}) ."\n";
print "Regulome subnetworks count = ". scalar keys(%{$rh_reg_nwa}) ."\n\n";

###### Part 2 - Subnet Integration to generate metasubnetworks

my($rh_meta_nwa, $rh_meta_noa, $rh_meta_eda, $rh_subnets, $rh_metaGOnwas, $rh_metaGOnoas) =
Metasubnet::metasubnets($rh_int_eda, $rh_int_nwa, $rh_int_noa, $rh_reg_eda, $rh_reg_nwa, $rh_reg_noa,
                        $th, $rh_tot_eda, $totalGeneNb, $rh_GOccur, $rh_gene2GOs);


if(keys(%$rh_meta_nwa)>0) {
    Metasubnet::writeMetasubnets($d_output, $rh_meta_nwa, $rh_meta_noa, $rh_meta_eda, $rh_subnets);
    
    print scalar(keys(%$rh_meta_nwa))." metasubnetworks were identified with connection threshold = $th\n";
    
    writeGO($rh_metaGOnwas, $rh_metaGOnoas, "$d_outputGO/meta_GOs-nwa.txt", "$d_outputGO/gene_GOs-noa.txt");
    
    #print "$0 done.\n";
}
else {
    print "No metasubnetworks were identified with connection threshold = $th\n";
    exit(1);
}
   

sub writeGO{
    my $rh_GONwas = shift;
    my $rh_GONoas = shift;
    my $f_GONwas = shift;
    my $f_GONoas = shift;

    open(GENE_GO, ">$f_GONoas") || die "Cannot open $f_GONoas:\n$!";
    print GENE_GO "# Gene_ID\tGO_ID\tGO_term\n";

    for my $geneID(keys %$rh_GONoas){
        for my $GOID(keys %{$$rh_GONoas{$geneID}}){

            my @GO = split(" - ", $GOID);
            my $GOIDShort = $GO[0];
            my $term = $GO[1];

            print GENE_GO "$geneID\t$GOIDShort\t$term\n";
        }
    }

    close(GENE_GO);

    open(SNW_GO, ">$f_GONwas") || die "Cannot open $f_GONwas:\n$!";
    print SNW_GO "# Network\tGO_ID\tGO_term\tCombined_subnet_GO_occurrence\tTotal_GO_occurrence\tCombined_subnet_gene_number\tAnnotated_gene_number\tEnrichment_ratio\tHypergeometric_test\tCorrected_p-value\n";

    for my $snwID(keys %$rh_GONwas){
        for my $GOID(keys %{$$rh_GONwas{$snwID}}){

            my @GO = split(" - ", $GOID);
            my $GOIDShort = $GO[0];
            my $term = $GO[1];

            print SNW_GO "$snwID\t$GOIDShort\t$term"
                . "\t$$rh_GONwas{$snwID}{$GOID}{'Subnet_GO_occurrence'}\t$$rh_GONwas{$snwID}{$GOID}{'Total_GO_occurrence'}"
                . "\t$$rh_GONwas{$snwID}{$GOID}{'Subnet_gene_number'}\t$$rh_GONwas{$snwID}{$GOID}{'Annotated_gene_number'}"
                . "\t$$rh_GONwas{$snwID}{$GOID}{'Enrichment_ratio'}\t$$rh_GONwas{$snwID}{$GOID}{'Hypergeometric_test'}"
                . "\t$$rh_GONwas{$snwID}{$GOID}{'Corrected_p-value'}\n";
        }
    }

    close(SNW_GO);
}
