#!/usr/bin/perl
#
# ITI project
#
# Copyright Maxime Garcia, Ghislain Bidaut, Claire Rioualen, Quentin Da Costa
# (2009-2017)
#
# maxime.garcia@inserm.fr
# ghislain.bidaut@inserm.fr
# http://bioinformatique.marseille.inserm.fr/iti/
#
# This software is a computer program whose purpose is to superimpose
# interactome  over  several  gene  expression  datasets  to  find
# discriminative subnetworks predicting metastatic relapse in breast
# cancer.
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME
intersectSubnetworksAndGenesClassificationList.pl - This script intersects lists of subnetworks and genes produced by classification routines


=head1  SYNOPSIS
intersectLists.pl	-sl <subnetworks files lists>
													-gl <gene files lists> [only one list is mandatory]
													-o  <Output directory>
													-p  <Overlap percentage>
													-t  <transcriptome directory>
													-s  <Suffix string>

=head1 DESCRIPTION
intersectSubnetworksClassificationList.pl intersects lists of subnets and generates a new list

=head1 SUBROUTINES
hashValueDescendingScore

=head1 Author
Maxime Garcia maxime.garcia@inserm.fr
Ghislain Bidaut ghislain.bidaut@inserm.fr

=cut

use strict;
use warnings;

use Subnetworks;
use Transcriptome;
use Utils;

#GLOBAL PARAMETERS
my $VERSION = 1.0;

########################
#     MAIN SECTION     #
########################

if ( $#ARGV < 1 ) {
	print "Usage: $0
	-sl     <subnetworks annotation files lists>
	-gl     <gene lists> [only one type of list is mandatory]
	-o      <Output directory>
	-t      <transcriptome directory>
	-c      <condition directory>
	-r      <resourcerer directory>
	-p      <Minimal overlap to merge two subnetworks> [specify value between 0.0 and 1.0]
	-s      <Suffix string>
	-abs    <uses absolute value for subnetwork score>
	-noabs  <Does not use absolute value for subnetwork score>
	-anti   <uses anticorrelation value for subnetwork score>
	-normal <Does not use anticorrelation value for subnetwork score>
	-nrand  <number of random subnetworks to generate>\n";
	exit;
}

my @a_subnetworkListFiles;
my @a_geneListFiles;

my ($outputDirectory,
	$conditionDirectory,
	$resourcererDirectory,
	$transcriptomeDirectory,
	$interactomeDirectory,
	$geneInfoFile,
	$overlapPercentage,
	$suffix,
	$useAbsoluteValue,
	$antiCorrelation);


for ( my $i = 0; $i < (scalar @ARGV); $i++ ) {
	if ( $ARGV[$i] eq "-sl" ) {
		for ( my $j = $i+1; $ARGV[$j] =~ m/^[^-]/ && $j < (scalar @ARGV) ; $j++ ) { push(@a_subnetworkListFiles, "$ARGV[$j]"); }
	}
	if ( $ARGV[$i] eq "-gl" ) {
		for ( my $j = $i+1; $ARGV[$j] =~ m/^[^-]/ && $j < (scalar @ARGV) ; $j++ ) { push(@a_geneListFiles, "$ARGV[$j]"); }
	}
	if ( $ARGV[$i] eq "-o"      ) { $outputDirectory        = "$ARGV[$i+1]"; }
	if ( $ARGV[$i] eq "-c"      ) { $conditionDirectory     = "$ARGV[$i+1]"; }
	if ( $ARGV[$i] eq "-a"      ) { $resourcererDirectory   = "$ARGV[$i+1]"; }
	if ( $ARGV[$i] eq "-t"      ) { $transcriptomeDirectory = "$ARGV[$i+1]"; }
	if ( $ARGV[$i] eq "-i"      ) { $interactomeDirectory   = "$ARGV[$i+1]"; }
	if ( $ARGV[$i] eq "-p"      ) { $overlapPercentage      = "$ARGV[$i+1]"; }
	if ( $ARGV[$i] eq "-s"      ) { $suffix                 = "$ARGV[$i+1]"; }
	if ( $ARGV[$i] eq "-gi"     ) { $geneInfoFile           = "$ARGV[$i+1]"; }
	if ( $ARGV[$i] eq "-abs"    ) { $useAbsoluteValue       = 1; }
	if ( $ARGV[$i] eq "-anti"   ) { $antiCorrelation        = 1; }
	if ( $ARGV[$i] eq "-noabs"  ) { $useAbsoluteValue       = 0; }
	if ( $ARGV[$i] eq "-normal" ) { $antiCorrelation        = 0; }
}

if(scalar(@a_subnetworkListFiles)==0 && scalar(@a_geneListFiles)==0) { die "No input file found"; }
if(!defined $outputDirectory )                                       { die "No output directory defined"; }
if(!defined $overlapPercentage )                                     { $overlapPercentage = 0.8; }
if(!defined $suffix )                                                { die "No suffix defined"; }
if(!defined $interactomeDirectory)                                   { die("No interactome directory defined"); }
if(!defined $transcriptomeDirectory)                                 { die("No transcriptome directory defined"); }
if(!defined $geneInfoFile)                                           { die("No gene_info file specified"); }
if(!defined $useAbsoluteValue )                                      { die("Absolute Value use not specified ( -abs or -noabs )"); }
if(!defined $antiCorrelation )                                       { die("Anti-correlation use not specified ( -anti or -normal )"); }
if(!defined $conditionDirectory)                                     { die "No condition directory defined"; }

print "Output directory        = $outputDirectory\n";
print "Resourcerer directory   = $resourcererDirectory\n";
print "Transcriptome directory = $transcriptomeDirectory\n";
print "Interactome directory   = $interactomeDirectory\n";
print "gene_info File          = $geneInfoFile\n";
print "Condition directory     = $conditionDirectory\n";
if(scalar(@a_subnetworkListFiles)) {
    print "Minimal overlap to merge two subnetworks      = $overlapPercentage\n";
}
print "Suffix string           = $suffix\n";

if($useAbsoluteValue) { print "Absolute Value for Pearson correlation is USED\n"; }
else                  { print "Absolute Value for Pearson correlation is NOT USED\n"; }
if($antiCorrelation)  { print "Detecting profiles ANTI-correlated with clinical condition\n"; }
else                  { print "Detecting profiles normally-correlated with clinical condition\n"; }


print "\n";

my $subnetworkDirectory = "$outputDirectory/subnetworks";

mkdir $outputDirectory;
mkdir $subnetworkDirectory;

my $rh_dataset                                   = Transcriptome::readTranscriptomeDir($transcriptomeDirectory, $resourcererDirectory, $conditionDirectory);
my $rh_projection                                = projectsTranscriptomeData($rh_dataset);
my ($rh_interactome, $rh_interactomeAnnotations) = readInteractomeDir($interactomeDirectory); #TODO Put in a module
my ($h_geneIDtoSymbol, $h_geneSymbolToID)        = Utils::readGeneInfo($geneInfoFile);

my %subNetworkProfile;
my %subNetworkProfileSize;

if(scalar(@a_subnetworkListFiles)>0) {
    my ($rh_Subnetwork, $name) = readSubNetworkList(\@a_subnetworkListFiles);
    my $rh_SubnetworkScore = calcSubNetworkAverageScore($rh_Subnetwork);
    my $rh_cluster = crossSubnetworksList($rh_Subnetwork, $overlapPercentage);

    print "Number of clusters = ". scalar (keys %$rh_cluster) ."\n";

    my $rh_SubnetworkKept = filterSubnetworkOnHighestScore($rh_cluster, $rh_SubnetworkScore);

    print "Number of subnets kept  = ". scalar (keys %$rh_SubnetworkKept) ."\n";

    my $fileName =  "$outputDirectory/subnets-kept-classification-$suffix.txt";
    printSubnetorksByDecreasingScore($rh_SubnetworkKept, $rh_SubnetworkScore, $fileName, $overlapPercentage,\@a_subnetworkListFiles);
}


if(scalar(@a_geneListFiles)>0) {
	my ($rh_GeneList, $name) = readGeneLists(\@a_geneListFiles);
	my $rh_crossedGeneLists = crossGeneList($rh_GeneList);
	my $rh_geneIDAverageScore = calcAverageScore($rh_GeneList);
	my $fileName =  "$outputDirectory/genes-kept-classification-$suffix.txt";
	printGeneListByDecreasingScore($rh_GeneList, $rh_crossedGeneLists, $rh_geneIDAverageScore, $fileName, \@a_geneListFiles);
}

sub readSubNetworkList {
	my $ra_subnetworkListFiles = shift;

	print "\nReading subnetworks lists :\n";

	my $name = '';
	my %h_Subnetwork;
	my $subnetworkID=0;
	foreach(my $i=0; $i< scalar(@$ra_subnetworkListFiles); $i++)  {

		my $list = $$ra_subnetworkListFiles[$i];

		# determining output name by aggregating list files names

		my $temp = $list;
		$temp =~ s/.txt//;
		my @tmp = split(/\_/, $temp);
		$name .= $tmp[scalar(@tmp)-1]."-";

		open(LISTFILE, "$list") || die "Cannot open $list : $!";
		my @a_dataTemp = <LISTFILE>;
		close LISTFILE;
		chomp @a_dataTemp;

		print "List $list\t";
		my $numberOfSubnetworks = 0;
		for ( my $j = 0; $j < (scalar @a_dataTemp); $j++ ) {
			if ( $a_dataTemp[$j] =~ m/^[^\#]/ && $a_dataTemp[$j] =~ m/annot.txt/ ) {

				my @tmp = split(/\t/, $a_dataTemp[$j]);
				my $temp2 = $tmp[0];
				$subnetworkID++;

				my $file = regenerateSubnetworkFile($subnetworkID, $tmp[0], $rh_dataset, \%subNetworkProfile,  \%subNetworkProfileSize, $subnetworkDirectory, $useAbsoluteValue, $antiCorrelation);

				$h_Subnetwork{$list}{$file} = Subnetworks::readSubnetworkFile($file);

				$numberOfSubnetworks++;
			}
		}
		print "contains $numberOfSubnetworks\tSubnetworks\n";
	}
	print "There are $subnetworkID subnetworks in total\n";
	return \%h_Subnetwork, $name;
}

sub calcSubNetworkAverageScore {
    my $rh_Subnetwork = shift;

    print "\nCalculate all scores, and assign them to the h_SubnetworkScore hash\n";

    my %h_SubnetworkScore;

    for my $list(sort keys %$rh_Subnetwork) {

	for my $file(sort keys %{$$rh_Subnetwork{$list}}) {

	    my $rh_curSubnetwork = $$rh_Subnetwork{$list}{$file};

	    my %rh_score = %{$rh_curSubnetwork->{'ScoreDataset'}};
	    my @a_score;
	    while( my($key,$value) = each(%rh_score)) { push(@a_score, $value); }

	    my($averageScore, $standardDeviation) = Utils::avstd(@a_score);
	    $h_SubnetworkScore{$file} = $averageScore;
	}
    }

    return \%h_SubnetworkScore;
}

sub crossSubnetworksList {
    my $rh_Subnetwork = shift;
    my $overlapPercentage = shift;

    my %h_cluster;
    my %h_SubnetworkFlagged;

    my @a_list = sort keys %$rh_Subnetwork;

    print "\nCompare all subnetworks two by two\n";

	if ( scalar(@a_list) <= 1 ) {
		die "Only one list to intersect with itself. Check your files\n";
	}

    for(my $i=0; $i<scalar(@a_list); $i++) {

		my $list1 = $a_list[$i];

		my @a_subnetList1 = sort keys %{$$rh_Subnetwork{$list1}};

		for(my $k=0; $k< scalar(@a_subnetList1); $k++)  {

			my $file1 = $a_subnetList1[$k];
			my $rh_subnetwork1 = $$rh_Subnetwork{$list1}{$file1};

			for(my $j=$i+1; $j<scalar(@a_list); $j++)  {

				my $list2 = $a_list[$j];

				my @a_subnetList2 = sort keys %{$$rh_Subnetwork{$list2}};
				my $maxOverlap = $overlapPercentage;
				my $maxSubnet = undef;

				for(my $l=0; $l< scalar(@a_subnetList2); $l++)  {

					my $file2 = $a_subnetList2[$l];

					my $rh_subnetwork2 = $$rh_Subnetwork{$list2}{$file2};

					my $netHash1    = $rh_subnetwork1->{'BinaryInteraction'};
					my $netHash2    = $rh_subnetwork2->{'BinaryInteraction'};
					my $overlap1    = Subnetworks::overlap($netHash1, $netHash2);
					my $overlap2    = Subnetworks::overlap($netHash2, $netHash1);

					my $overlap     = Utils::max($overlap1, $overlap2);

					if($overlap >= $maxOverlap && !$h_SubnetworkFlagged{$list2}{$file2}) {
						$maxOverlap = $overlap;
						$maxSubnet  = $file2;
					}
				}
				#THIS THING IS PROBLEMATIC, BECAUSE HERE WE KEEP THE LAST SUBNETWORK FOR EACH LIST THAT IS OVERLAPPING OVER THE SUBNERWORK $file1 (AND NOT THE MORE OVERLAPPING)...
				if(defined $maxSubnet) {
					push(@{$h_cluster{$file1}}, $maxSubnet);
					if (scalar @{$h_cluster{$file1}} == 1) { unshift(@{$h_cluster{$file1}}, $file1);}
					$h_SubnetworkFlagged{$list2}{$maxSubnet} = 1;
				}
			}
		}
    }
    return \%h_cluster;
}

sub filterSubnetworkOnHighestScore {
    my $rh_cluster = shift;

    my $rh_SubnetworkScore = shift;

    my %h_SubnetworkKept;

    my @a_list = sort keys %$rh_SubnetworkScore;

    foreach my $file(sort keys %$rh_cluster) {

		my $maxScore = $$rh_SubnetworkScore{$file};
		my $maxNet = undef;

		for (my $i=0; $i<scalar @{$$rh_cluster{$file}}; $i++) {
			my $subnet = @{$$rh_cluster{$file}}[$i];
			if($$rh_SubnetworkScore{$subnet}>=$maxScore) {
				$maxScore = $$rh_SubnetworkScore{$subnet};
				$maxNet   = $subnet;
			}
		}

		if(defined $maxNet) {
			$h_SubnetworkKept{$maxNet} = $maxScore;
		}
    }
    return \%h_SubnetworkKept;
}

sub printSubnetorksByDecreasingScore {

	my $rh_SubnetworkKept  = shift;
	my $rh_SubnetworkScore = shift;
	my $fileName           = shift;
	my $overlapPercentage  = shift;
	my $ra_subnetworkList  = shift;

	my %h_subnetScore2 = %{$rh_SubnetworkKept};

	print "Print subnetworks sorted by decreasing average score\n";

	open(SUBNETSKEPT, ">$fileName") || die "Cannot open $fileName : $!";

	print SUBNETSKEPT "#subnetworks sorted by decreasing average score\n";
	print SUBNETSKEPT "#subnetworks kept only if present in all input files with an overlap=$overlapPercentage\n";
	for my $file (@$ra_subnetworkList) { print SUBNETSKEPT "#$file\n"; }
	foreach my $subnetworkNumber (sort {  $h_subnetScore2{$b} <=>  $h_subnetScore2{$a}; } (keys(%h_subnetScore2))) {
		print SUBNETSKEPT "$subnetworkNumber\t$$rh_SubnetworkScore{$subnetworkNumber}\n";
	}

	close SUBNETSKEPT;
}

sub readGeneLists {

	my $ra_geneListFiles = shift;

	print "\nRead gene lists :\n";

	my $name = '';
	my %h_GeneLists;

	foreach(my $i=0; $i< scalar(@$ra_geneListFiles); $i++)  {

	my $list = $$ra_geneListFiles[$i];

	print "List : $list\n";

	# determining output name by aggregating list files names

	my $temp = $list;
	$temp =~ s/.txt//;
	my @tmp = split(/\_/, $temp);
	$name .= $tmp[scalar(@tmp)-1]."-";

	open(FILE, "$list") || die "Cannot open $list : $!";
	my @a_dataTemp = <FILE>;
	close FILE;
	chomp @a_dataTemp;

	foreach(@a_dataTemp) {
		if ($_ =~ m/^[^\#]/ ) {
		my @a_rowTemp = split(/\t/,$_);
		my $geneID = $a_rowTemp[0];
		my $score = $a_rowTemp[1];
		$h_GeneLists{$list}{$geneID} = $score;
		}
	}

	print "list contains ". scalar(keys(%{$h_GeneLists{$list}})) . " genes\n";
	}

	return \%h_GeneLists, $name;
}


sub calcAverageScore {
	my $rh_GeneList = shift;

	my @fileList = sort keys %$rh_GeneList;

	my %h_geneIDAverageScore;

	for my $file(@fileList) {
	for my $geneID(sort keys %{$$rh_GeneList{$file}}) {
		$h_geneIDAverageScore{$geneID} = 0;
	}
	}

	my @geneIDList = sort keys %h_geneIDAverageScore;

	for my $geneID(@geneIDList) {

	my @score=();
	for my $file(@fileList) {

		if(defined $$rh_GeneList{$file}{$geneID}) {
		push(@score, $$rh_GeneList{$file}{$geneID});
		}
	}

	my($av, $std) = Utils::avstd(@score);

	$h_geneIDAverageScore{$geneID} = $av;
	}

	return \%h_geneIDAverageScore;
}


sub crossGeneList {
	my $rh_GeneList = shift;

	my %h_crossedGeneLists;

	for my $file(sort keys %$rh_GeneList) {
		for my $geneID(sort keys %{$$rh_GeneList{$file}}) {
			if(defined $h_crossedGeneLists{$geneID}) { $h_crossedGeneLists{$geneID}++; }
			else                                     { $h_crossedGeneLists{$geneID} = 1; }
		}
	}
	return \%h_crossedGeneLists;
}


sub printGeneListByDecreasingScore {

	my $rh_GeneList = shift;
	my $rh_crossedGeneLists = shift;
	my $rh_geneIDAverageScore = shift;
	my $fileName= shift;
	my $ra_geneListFiles = shift;

	open(GENEID, ">$fileName") || die "Cannot open $fileName : $!";

	print GENEID "#genes sorted by frequency and decreasing average score\n";
	for my $file (@$ra_geneListFiles) { print GENEID "#$file\n"; }
	my %h_subcrossedList;

	while(my($geneID, $frequency)=each (%$rh_crossedGeneLists)) {
	push(@{$h_subcrossedList{$frequency}}, $geneID);
	}

	for(my $index = scalar keys %$rh_GeneList; $index>=1; $index--) {
	my %h_subGeneIDHash = ();
	foreach my $geneID(@{$h_subcrossedList{$index}}) {
		$h_subGeneIDHash{$geneID} = $$rh_geneIDAverageScore{$geneID};
	}

	foreach my $geneIDNumber (sort {  $h_subGeneIDHash{$b} <=>  $h_subGeneIDHash{$a}; } (keys(%h_subGeneIDHash))) {
		print GENEID "$geneIDNumber\t$index\t$$rh_geneIDAverageScore{$geneIDNumber}\n";
	}
	}

	close GENEID;
}

sub projectsTranscriptomeData { #TODO Put in transcriptome.pm as a function
	my $rh_dataset = shift;

	my %h_projection;

	print "perform projection\n";

	my @experimentNumber;
	foreach my $label(keys %$rh_dataset) {
		my $ra_experience = $$rh_dataset{$label}{'expe'};
		push(@experimentNumber, scalar @$ra_experience);
	}

	my $maxNumberofExperiments = Utils::max(@experimentNumber);

	print "maxmimum number of experiments =  $maxNumberofExperiments\n";

	print "measured correlations are normalised by sqrt(nconds_dataset)/sqrt(max_nconds_dataset)\n";

	foreach my $label(sort keys %$rh_dataset) {

		my $rh_data = $$rh_dataset{$label}{'collapseddata'};
		my $ra_experience = $$rh_dataset{$label}{'expe'};
		my $rh_conditions = $$rh_dataset{$label}{'cond'};
		my $nconds = scalar @$ra_experience;

		print "current Dataset = $label  ($nconds conditions)\n";

		my %projection;

		foreach my $geneID(keys %$rh_data) {

			my $correlation;
			my $n=0;
			my (@vect1, @vect2);
			foreach my $exp(@$ra_experience) {
				push(@vect1, $$rh_conditions{$exp}+1);
				push(@vect2, $$rh_data{$geneID}{$exp});
			}

			if($useAbsoluteValue == 1) { $correlation = abs(Utils::pearson(\@vect1, \@vect2)); }
			else                       { $correlation = Utils::pearson(\@vect1, \@vect2); }
			if($antiCorrelation == 1)  { $correlation = -Utils::pearson(\@vect1, \@vect2); }
			else                       { $correlation = Utils::pearson(\@vect1, \@vect2); }

			$correlation*=sqrt($nconds)/sqrt($maxNumberofExperiments);

			#print "proj value = $correlation\n";
			$projection{$geneID} = $correlation;
		}
		$h_projection{$label} = \%projection;
	}
	return \%h_projection;
}


sub regenerateSubnetworkFile {
	my $subnetworkID             = shift;
	my $subnetworkAnnotationFile = shift;
	my $datasetsHash             = shift;
	my $subNetworkProfile        = shift;
	my $subNetworkProfileSize    = shift;
	my $subnetworkDirectory      = shift;
	my $useAbsoluteValue         = shift;
	my $antiCorrelation          = shift;

	my @label = keys(%$datasetsHash);

	my $subnetworkInteractionFile = $subnetworkAnnotationFile;
	$subnetworkInteractionFile =~ s/-subnet-annot.txt/-subnet.txt/;

	my %rh_binaryInteraction;

	open(SUBNET, "$subnetworkInteractionFile") || die "Cannot open $subnetworkInteractionFile: $!";
	my @data = <SUBNET>;
	chomp @data;

	foreach(@data) {
		my @row = split(/\t/, $_);

		if(!defined $rh_binaryInteraction{$row[0]}) {
			my @ar;
			push(@ar, $row[1]);
			$rh_binaryInteraction{$row[0]} = \@ar;
		}
		else { push(@{$rh_binaryInteraction{$row[0]}}, $row[1]); }
	}
	close SUBNET;

	my %subnetwork;
	$subnetwork{'BinaryInteraction'}       = \%rh_binaryInteraction;
	$subnetwork{'ID'}                      = $subnetworkID;
	$subnetwork{'path'}                    = $subnetworkInteractionFile;

#	my %subNetwork             = %{$$net{'Nodes'}};
	my $outFile                = "$subnetworkDirectory/$subnetworkID-subnet.txt";

#	printSubnetFile ($id,\%subnetwork{'BinaryInteraction'}, \%subNetwork, $outFile);

#ça c'est la partie à mettre en fonction dans subnet.pm
	my @a_node;

	open(OUT, ">$outFile") || die "Cannot open $outFile : $!";
	my %h_geneInSubnetwork;
	for my $id(keys %rh_binaryInteraction) {
	    foreach my $g2 (@{$rh_binaryInteraction{$id}}) {
			if ( !defined $h_geneInSubnetwork{$id} ) { $h_geneInSubnetwork{$id} =1 };
			if ( !defined $h_geneInSubnetwork{$g2} ) { $h_geneInSubnetwork{$g2} =1 };
			print OUT "$id\t$g2";
			my $gi1 = Utils::min($id, $g2);
			my $gi2 = Utils::max($id, $g2);
			if(defined $$rh_interactomeAnnotations{"$gi1"."-"."$gi2"}) {
			    my @formatted = ();
			    for my $interactome (sort keys %{$$rh_interactomeAnnotations{"$gi1"."-"."$gi2"}}) {
					push(@formatted, "$interactome\:\:" . join (';', @{$$rh_interactomeAnnotations{"$gi1"."-"."$gi2"}{$interactome}}));
			    }
			    print OUT "\t" .join('||', @formatted);
			}
		print OUT "\n";
	    }
	}
	close OUT;
	my %score;

	$outFile = "$subnetworkDirectory/$subnetworkID-subnet-annot.txt";

#ça on met en fonction dans subnet.pm
#	printSubnetAnnotFile ($id,\%subnetwork{'BinaryInteraction'}, \%subNetwork, $outFile);

	open(OUT, ">$outFile") || die "Cannot open $outFile : $!";
	print OUT "ID\tSymbol";
	for my $label(sort keys %$rh_dataset) { print OUT "\tprojvalue-$label"; }
	for my $label(sort keys %$rh_dataset) { print OUT "\tscore-$label"; }
	print OUT "\n";

	foreach my $node(keys %h_geneInSubnetwork) {
			foreach my $label(keys %$rh_dataset) {
			if(defined $$rh_projection{$label}{$node}) {
				my $rh_data = $$rh_dataset{$label}{'collapseddata'};
				my $ra_experience = $$rh_dataset{$label}{'expe'};
				my $rh_conditions = $$rh_dataset{$label}{'cond'};
				my $rh_ranconditions = $$rh_dataset{$label}{"rancond"};
				if(defined $subNetworkProfileSize->{$label}) {
					for(my $i=0; $i<scalar @$ra_experience; $i++) {
						my $exp = $$ra_experience[$i];
						if(defined  $$rh_data{$node}{$exp}) {
							$subNetworkProfile->{$label}{$exp} += $$rh_data{$node}{$exp};
							$subNetworkProfileSize->{$label}{$exp}++;
						}
					}
				}
				else {
					for(my $i=0; $i<scalar @$ra_experience; $i++) {
						my $exp = $$ra_experience[$i];
						if(defined $$rh_data{$node}{$exp}) {
							$subNetworkProfile->{$label}{$exp} = $$rh_data{$node}{$exp};
							$subNetworkProfileSize->{$label}{$exp} = 1;
						}
						else {
							$subNetworkProfile->{$label}{$exp} = undef;
							$subNetworkProfileSize->{$label}{$exp} = 0;
						}
					}
				}
			} # if defined projection
		} # foreach label
	}

	for my $label(keys %$rh_dataset) { $score{$label} = networkScore($rh_dataset, \%subNetworkProfile, \%subNetworkProfileSize, $label); }

	for my $id(keys %h_geneInSubnetwork) {
		if(defined $h_geneIDtoSymbol->{$id}) {
			my $sym = $h_geneIDtoSymbol->{$id};
			print OUT "$id\t$sym";
		}
		else { print OUT "$id\t"; }
		for my $label(sort keys %$rh_dataset) {
			if(defined $$rh_projection{$label}{$id}) {
				my $printedValue;
				if($useAbsoluteValue == 1) {
					$printedValue = abs($$rh_projection{$label}{$id});
				}
				else {
					$printedValue = $$rh_projection{$label}{$id};
				}
				if($antiCorrelation == 1) {
					$printedValue = -$$rh_projection{$label}{$id};
				}
				else {
					$printedValue = $$rh_projection{$label}{$id};
				}
				print OUT "\t$printedValue";
			}
			else { print OUT "\tundef"; }
		}
		for my $label(sort keys %$rh_dataset) {
			if (defined $score{$label}) {
				print OUT "\t$score{$label}";
			}
			else { print OUT "\tundef"; }
		}
		print OUT "\n";
	}
	close OUT;
	return $outFile;
}

sub readInteractomeDir {
    my $interactomeDirectory = shift;
    my @dirContent = Utils::readInputDir($interactomeDirectory);
    my %h_combinedInteractome;
    my %h_combinedInteractomeAnnotations;

    print "Interactome directory content =  @dirContent\n\n";

    for(my $i=0; $i< scalar @dirContent; $i++) {
	my $curFile = $dirContent[$i];
	print "curfile = $curFile\n";

	my ($rh_interactome, $rh_interactomeAnnotations, $interactomeID) = parseInteractome("$interactomeDirectory/$curFile"); #TODO Put in a module

	while(my($keys, $values) = each(%$rh_interactome)) {

	    if(defined $h_combinedInteractome{$keys}) {
		for my $v(@$values) { push(@{$h_combinedInteractome{$keys}}, $v) if(!Utils::in_array($v, $h_combinedInteractome{$keys})); }
	    }
	    else { $h_combinedInteractome{$keys} = $values; }

	    my $geneID1 = $keys;
	    for my $geneID2(@$values) {

		my $gi1 = Utils::min($geneID1, $geneID2);
		my $gi2 = Utils::max($geneID1, $geneID2);

		foreach my $a(@{$$rh_interactomeAnnotations{"$gi1"."-"."$gi2"}}) {
		    if(!Utils::in_array($a, \@{$h_combinedInteractomeAnnotations{"$gi1"."-"."$gi2"}{$interactomeID}})) {
			push(@{$h_combinedInteractomeAnnotations{"$gi1"."-"."$gi2"}{$interactomeID}}, $a);
		    }
		}
	    }
	}
    }

    print "total number of individual node(s): ". scalar (keys %h_combinedInteractome) ."\n";
    my $ninter = 0;

    while(my($k,$v) =each(%h_combinedInteractome)) { $ninter += scalar @$v; }

    $ninter/=2;

    print "$ninter unique interaction(s) found in combined interactome\n\n";

    return \%h_combinedInteractome, \%h_combinedInteractomeAnnotations;
}

sub parseInteractome {

    my $file = shift;

    print "reading interactome file $file\n";
    if(!open(FILE, "<$file")) { die("cannot open $file : $!"); }

    my @data = <FILE>;
    chomp @data;
    close(FILE);

    my @tmp1 = split(/\//, $file);
    my @tmp2 = split(/\_/, $tmp1[-1]);

    my $interactomeID = $tmp2[0];

    print "Interactome ID = $interactomeID\n";

    my %h_interactome;
    my %h_interactomeAnnotations;

    for(my $l=0; $l<scalar @data; $l++) {

	my @row=split(/\t/, $data[$l]);
	my $geneID1 = $row[1];
	my $geneID2 = $row[3];
	my @annot = split(';', $row[4]);

	next if($geneID1 eq $geneID2);

	if(!defined $h_interactome{$geneID1}) {

	    my @ar;
	    push(@ar, $geneID2);
	    $h_interactome{$geneID1} = \@ar;
	}

	else { push(@{$h_interactome{$geneID1}}, $geneID2) if(!Utils::in_array($geneID2, $h_interactome{$geneID1})); }

	if(!defined $h_interactome{$geneID2}) {
	    my @ar;
	    push(@ar, $geneID1);
	    $h_interactome{$geneID2} = \@ar;
	}

	else { push(@{$h_interactome{$geneID2}}, $geneID1) if(!Utils::in_array($geneID1, $h_interactome{$geneID2}));  }

	my $gi1 = Utils::min($geneID1, $geneID2);
	my $gi2 = Utils::max($geneID1, $geneID2);

	foreach my $an(@annot) {
	    if(!Utils::in_array($an, \@{$h_interactomeAnnotations{"$gi1"."-"."$gi2"}})) {
		push(@{$h_interactomeAnnotations{"$gi1"."-"."$gi2"}}, $an);
	    }
	}

    }
    print "number of individual node(s) = " . (scalar keys %h_interactome) ."\n";
    my $ninter = 0;
    while(my($k,$v) =each(%h_interactome)) { $ninter += scalar @$v; }
    $ninter/=2;
    print "$ninter unique interaction(s) found\n\n";

    return \%h_interactome, \%h_interactomeAnnotations, $interactomeID;
}

# return network score
sub networkScore { #TODO Put in subnet.pm

	my $rh_datasets           = shift;
	my $subNetworkProfile     = shift;
	my $subNetworkProfileSize = shift;
	my $label                 = shift;

	#print "networkscore - ";
	#print "current subnetwork = \n";
	#foreach(keys %$subNetwork) { print "$_ "; }
	if(defined $subNetworkProfileSize->{$label} && $subNetworkProfileSize->{$label}>0) { # subnet non vide
		my $score;
		my $ra_experience = $$rh_datasets{$label}{'expe'};
		my $rh_conditions = $$rh_datasets{$label}{'cond'};
		my (@vect1, @vect2);

		for(my $i=0; $i<scalar (@$ra_experience); $i++) {
			my $exp = $$ra_experience[$i];

			if(defined $subNetworkProfile->{$label}{$exp}) { $vect1[$i] = $subNetworkProfile->{$label}{$exp}/$subNetworkProfileSize->{$label}{$exp}; }
			else                                           { $vect1[$i] = undef; }
		}
		for my $exp(@$ra_experience) { push(@vect2, $$rh_conditions{$exp}+1); }
		if($antiCorrelation == 1)  { $score = -Utils::pearson(\@vect1, \@vect2); }
		else                       { $score = Utils::pearson(\@vect1, \@vect2); }
		if($useAbsoluteValue == 1) { $score = abs(Utils::pearson(\@vect1, \@vect2)); }
		else                       { $score = Utils::pearson(\@vect1, \@vect2); }
		return $score;
	}
	else { return undef; }
}
