#!/usr/bin/perl
#
# HTS-Net project
#
# Copyright Maxime Garcia, Ghislain Bidaut, Claire Rioualen, Quentin Da Costa
# (2009-2017)
#
# maxime.u.garcia@gmail.com
# ghislain.bidaut@inserm.fr
# claire.rioualen@inserm.fr
# quentin.da-costa@inserm.fr
#
# http://iti.sourceforge.net
#
# This software is a computer program whose purpose is to superimpose
# interactome over several gene expression datasets to find
# discriminative subnetworks predicting metastatic relapse in breast
# cancer.
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty and the software's author, the holder of the
# economic rights, and the successive licensors have only limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading, using, modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean that it is complicated to manipulate, and that also
# therefore means that it is reserved for developers and experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and, more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod

=head1 Name
intersectSubnetworks-getGO-computePVal.pl

=head1 Synopsis

=head2 Multiple output directories
intersectSubnetworks-getGO-computePVal.pl
        -od | --outputDir <Output directory>

=head2 Output files in current directory
intersectSubnetworks-getGO-computePVal.pl
        -so | --singleOutput

        -nnf    | --subnetworkFile       <Real subnetwork file>
        -nwaf   | --networkAttributeFile <Real network attribute file (network scores)>
        -randsf | --randomScoreFiles     <Random score file 1 Random score file 2 Random score file 3>
        -noaf   | --geneScoreFile        <Real node attribute file (gene scores)>

=head2 Required
        -htsnet | --htsnetPath                 <HTS-Net directory>

        -gof | --gene2go                 <gene2go file>
        -nwf | --networkFile             <Interaction network file>

        -pVal            <Maximum p-value(s) of subnetworks to keep>
        -c | --consensus <Consensus number of datasets>
        -taxon           <Taxon ID>

=head2 Optional
        -lf   | --listFiles         <Kept subnet file 1 Kept subnet file 2 Kept subnet file 3>
        -igf  | --interestGeneFiles <Lists of genes of interest>

        -ov    | -mo | -ot    <Overlap threshold>
        -merge | --snwMerging
        -s     | --suffix     <Suffix string>

        -h | --help | -u | --usage

=head1 Examples
perl intersectSubnetworks-getGO-computePVal.pl -htsnet ~/workspace/local-htsnet/ -od ~/workspace/data/htsnet_test/test/ -pVal 1e-3 -consensus 1 -gof /home/data/annotations/stable/ncbi/Homo_sapiens.gene2go.gz -nwf ~/workspace/data/network_data-test/network.txt -taxon 9606
perl intersectSubnetworks-getGO-computePVal.pl --singleOutput --listFiles subnets-kept_dist1-normal.txt subnets-kept_dist2-gmd.txt subnets-kept_dist3-normal.txt -nwf real-subnets.nnf -nwaf real-subnet-scores-nwa.txt -randsf scores_dist1-normal.txt scores_dist2-normal.txt scores_dist3-normal.txt -pVal 0.05 -consensus 1 -gof /home/data/annotations/stable/ncbi/Homo_sapiens.gene2go.gz -nwf ~/workspace/data/network_data-test/network.txt -taxon 9606

=head1 Description
=over

=item This script intersects lists of subnetworks made by subnetworksStatisticalValidation.pl and generates a new list;
=item It finds the Gene Ontologies IDs of functions in which genes of selected subnetworks are involved;
=item It computes the p-value of each kept subnetwotk.

=back

=head1 Subroutines
=over

=item hashValueDescendingScore
=item hashValueDescendingGene

=back

=head1 Authors
Ghislain Bidaut     ghislain.bidaut@inserm.fr
Quentin Da Costa    quentin.da-costa@inserm.fr
Maxime Garcia       maxime.u.garcia@gmail.com
Claire Rioualen     claire.rioualen@inserm.fr

=cut


## Modules ##

use strict;
use warnings;

use Getopt::Long;
use File::Path;
use IO::Uncompress::Gunzip qw(gunzip $GunzipError);

use Subnetworks;
use Network;
use Utils;


## Arguments management ##

my ($help,         $d_output,
    $d_htsnet,  $d_rscript,        $suffix,
    $taxonID,      $f_gene2go,
    $outInCurrDir, @a_pVal,
    $consensus,    $f_nnf,
    $f_nwa,        $f_noa,
    $f_network,    $f_totalEda, $maxOvl,
    @a_f_list,     @a_f_randomNwa,
    @a_f_interestGenes, $snwMerging,
    $useAbsVal, $useAnti,
    $f_pVal);

GetOptions(
    'help|usage|u|h'               => \$help,
    'gene2goFile|gof=s'            => \$f_gene2go,
    'htsnetPath|htsnet=s'                => \$d_htsnet,
    'rscriptDir|rsd=s'             => \$d_rscript,
    'outputDir|od=s'               => \$d_output,
    'networkFile|nwf=s'            => \$f_network,
    'edgeAttFile|edaf=s'           => \$f_totalEda,
    'interestGeneFiles|igf=s{,}'   => \@a_f_interestGenes,
    'taxon=i'                      => \$taxonID,
    'suffix|s=s'                   => \$suffix,
    'listFiles|lf=s{,}'            => \@a_f_list,
    'pVal=f{,}'                    => \@a_pVal,
    'pValFile|pvf=s'               => \$f_pVal,
    'consensus=i'                  => \$consensus,
    'ov|mo|ot=f'                   => \$maxOvl,
    'snwMerging|merge'             => \$snwMerging,
    'abs'                          => \$useAbsVal,
    'anti'                         => \$useAnti,
    'singleOutput|so'              => \$outInCurrDir,
    'subnetworkFile|nnf=s'         => \$f_nnf,
    'networkAttributeFile|nwaf=s'  => \$f_nwa,
    'randomScoreFiles|randsf=s{,}' => \@a_f_randomNwa,
    'geneScoreFile|noaf=s'         => \$f_noa);

if(defined $d_rscript) {
    if($d_rscript !~ m/\/$/){$d_rscript .= "/";}
}
else {
    $d_rscript = "";
}
if(!defined $maxOvl || $maxOvl <= 0 || $maxOvl > 1){
    $maxOvl = "Unused";
}

my ($d_lists, $d_GO);

if(defined $outInCurrDir  and
   scalar(@a_f_list)      and
   defined $f_nnf         and
   defined $f_nwa         and
   scalar(@a_f_randomNwa) and
   defined $f_noa         and
   defined $f_gene2go     and
   defined $f_network     and
   defined $d_htsnet         and
   defined $taxonID       and
   (scalar(@a_pVal) || defined $f_pVal)       and
   defined $consensus     and
   !(defined $snwMerging and !defined $f_totalEda)){

    ($d_output, $d_lists, $d_GO) = ("", "", "");

    print "$0:

    HTS-Net directory                       = $d_htsnet

    Files listing subnetworks               = @a_f_list
    Subnetwork file                         = $f_nnf
    Network attribute file (network scores) = $f_nwa
    Random score files                      = @a_f_randomNwa
    Node attribute file (gene scores)       = $f_noa

    Gene to GO file                         = $f_gene2go
    Network file                            = $f_network
    Taxon                                   = $taxonID

    Maximum p-value of subnetworks to keep  = @a_pVal
    Consensus number of datasets            = $consensus
    Maximum overlap                         = $maxOvl
\n\n";
}

elsif(defined $d_output  and
      defined $f_gene2go and
      defined $f_network and
      defined $d_htsnet     and
      defined $taxonID   and
      (scalar(@a_pVal) || defined $f_pVal)    and
      defined $consensus and
      !(defined $snwMerging and !defined $f_totalEda)) {

    if($d_output !~ m/\/$/) {$d_output .= "/";}
    $d_lists = $d_output . "lists/";
    $d_GO    = $d_output . "GO/";
    rmtree($d_GO, 0, 0);
    mkdir $d_GO;

    # Retrieve full path of kept subnetwork files, network file and network attribute files.
    @a_f_list      = glob("$d_lists"  . "subnets-kept_dist*") unless(scalar(@a_f_list));
    $f_nnf         = glob("$d_output" . "subnets/*.nnf");
    $f_nwa         = glob("$d_output" . "subnets/*nwa.txt");
    @a_f_randomNwa = glob("$d_output" . "scores/*dist*");
    $f_noa         = glob("$d_output" . "subnets/*noa.txt");

    print "$0:

    HTS-Net directory                       = $d_htsnet
    Output directory                        = $d_output

    Files listing subnetworks               = @a_f_list
    Subnetwork file                         = $f_nnf
    Network attribute file (network scores) = $f_nwa
    Random score files                      = @a_f_randomNwa
    Node attribute file (gene scores)       = $f_noa

    Gene to GO file                         = $f_gene2go
    Network file                            = $f_network
    Taxon                                   = $taxonID

    Maximum p-value of subnetworks to keep  = @a_pVal
    Consensus number of datasets            = $consensus
    Maximum overlap                         = $maxOvl
\n\n";
}

else{
    print "
NAME
intersectSubnetworks-getGO-computePVal.pl

DESCRIPTION
- This script intersects lists of subnetworks made by subnetworksStatisticalValidation.pl and generates a new list;
- It finds the Gene Ontologies IDs of functions in which genes of selected subnetworks are involved;
- It computes the p-value of each kept subnetwotk.

PARAMETERS

Multiple output directories
-od | --outputDir <Output directory>\n\tUse the given directory (full path) as output directory which must be the same than the one used in subnetworksStatisticalValidation.pl

Output files in current directory
-so | --singleOutput

-nnf | --subnetworkFile <Real subnetwork file>
-nwaf | --networkAttributeFile <Real network attribute file>\n\tScores of real subnetworks
-randsf | --randomScoreFiles <Random score file 1 Random score file 2 Random score file 3>\n\tScores of random subnetworks
-noaf | --geneScoreFile <Real node attribute file>\n\tGene scores of random subnetworks

Required
-htsnet | --htsnetPath <HTS-Net directory>\n\tHTSNET directory, where pipeline/ and R/ directories must be
-gof | --gene2goFile <gene2go file>\n\tFull path of local copy of gene2go file gzip-compressed (.gz) downloaded from NCBI
-nwf | --networkFile <Interaction network file>\n\tInteractome and/or regulome
-pVal <Maximum p-value(s) of subnetworks to keep>
-c | --consensus <Consensus number of datasets>
-taxon <Taxon ID>\n\tTaxonomic ID (TAXID) of the studied species as defined in the Taxonomy database at NCBI (http://www.ncbi.nlm.nih.gov/taxonomy/)

Optional
-lf | --listFiles <Kept subnet file 1 Kept subnet file 2 Kept subnet file 3>\n\tList files of subnetworks you want to intersect and which gene GO functions will be retrieved and p-values computed
-igf | --interestGeneFiles <Lists of genes of interest>\n\tLists of genes of interest (gene IDs) to highlight in visualization (gene IDs in the first column of files, annotations in the second column (not required))
-ov | -mo | -ot <Overlap threshold>\n\tUse the given number in ]0; 1] as identity threshold from which 2 subnetworks are filtered/merged
-merge | --snwMerging\n\tCombine overlapping subnetworks instead of filtering them
-s | --suffix <Suffix string>\n\tWrite the given string in created files so that you can identify from which run they come
-rsd | --rscriptDir <Rscript directory>

-h | --help | -u | --usage\n\tDisplay this user guide

EXAMPLES
$0 -htsnet ~/workspace/local-htsnet/ -od ~/workspace/data/htsnet_test/test/ -pVal 1e-3 -consensus 1 -gof /home/data/annotations/stable/ncbi/Homo_sapiens.gene2go.gz -nwf ~/workspace/data/network_data-test/network.txt -taxon 9606
$0 --singleOutput -lf subnets-kept_dist1-normal.txt subnets-kept_dist2-gmd.txt subnets-kept_dist3-normal.txt -nwf real-subnets.nnf -nwaf real-subnet-scores-nwa.txt -randsf scores_dist1-normal.txt scores_dist2-normal.txt scores_dist3-normal.txt -pVal 0.05 -consensus 1 -gof /home/data/annotations/stable/ncbi/Homo_sapiens.gene2go.gz -nwf ~/workspace/data/network_data-test/network.txt -taxon 9606

Please report every problem at <ghislain.bidaut\@inserm.fr>

";
    exit(0);
}

$d_htsnet =~ s/\/+$//;

if(defined $f_pVal) {
     open(FILE, "$f_pVal") || die "Cannot open $f_pVal: $!";
     while(defined (my $l =<FILE>)) {
	   next if ($l =~ m/^#/);
	   $l =~ s/[\r\n]+$//;
	   my $pVal = $l;
	   print "**** p-value = $pVal\n";
	   push(@a_pVal, $pVal);
     }
     close FILE;
     
}

if(scalar(@a_pVal) == 1 and scalar(@a_f_list) > 1){
    my $pVal = $a_pVal[0];
    for my $f(@a_f_list){
        push(@a_pVal, $pVal);
    }
}

die "\nA p-value is missing" if(scalar(@a_pVal) < scalar(@a_f_list));


## Main program ##

my $oldTime = time;

#print "\nRead subnetwork lists...\n";

# Counting the occurrence of each subnetwork in all lists.
my $i = 0;
my %h_subnetOccurrence;
for my $file (sort @a_f_list) {
    print "Current file: $file\n";

    open(FILE, $file) || die "Cannot open $file:\n$!";

    my $isTheGoodList = 0;
    while (my $subnet = <FILE>) {
        next if($subnet =~ m/^#/);
        $subnet =~ s/[\r\n]+$//;
        if ($subnet =~ m/^>START/) {
            last if $isTheGoodList;
            my @a_start = split(/\|/, $subnet);
            my $p = $a_start[1];
            my $c = $a_start[2];
            if ($p == $a_pVal[$i] and $c == $consensus){
                $isTheGoodList = 1;
            }
        }
        elsif ($isTheGoodList) {
            $h_subnetOccurrence{$subnet}++;
        }
    }

    close FILE;
    $i++;
}
die "\nNo subnetwork found with selected p-value(s) (@a_pVal) and consensus number ($consensus)"
    if(scalar(keys(%h_subnetOccurrence)) == 0);

print "\nSearch subnetworks present in all lists...\n";
my @a_subnetInAllLists;
my $listNb = scalar @a_f_list;
while (my ($subnetID, $occurrence) = each(%h_subnetOccurrence)) {
    if ($occurrence == $listNb) {
        push(@a_subnetInAllLists, $subnetID);
    }
}
my $selectedSubnetNb = scalar(@a_subnetInAllLists);

die "\nNo subnetworks were found with specified p-values and overlap\n" if($selectedSubnetNb == 0);
print "Number of subnetworks = $selectedSubnetNb\n\n";

# Get:
# Subnet / Network attribute hash:
# \%rh_subnets_nwa($subnetwork_ID => \%('Edges' => \@("$gene1-$gene2"),
#                                       'Nodes' => \%($gene_ID => 1))),
#                                    \%('Scores' => \%($datasetLabel => $subnetwork_score))).
my($rh_subnets_nwa, $rh_nodeAtts,
   $rh_edgeAtts) = Subnetworks::readSubnets($f_nnf, $f_nwa, $f_noa, undef, \@a_subnetInAllLists);

my $rh_network  = Utils::readHashFile($f_network);

#print "\n********* Processes nodes degree\n\n";
for my $n (keys %$rh_network){
    my $degree = scalar(@{$$rh_network{$n}});
    $$rh_nodeAtts{$n}{'Gene_degree'} = $degree;
}

my($rh_keptSubnets_nwa, $rh_avScore);
if($maxOvl ne "Unused"){
    if(defined $snwMerging){
        my $rh_scores_nwa;
        ($rh_keptSubnets_nwa, $rh_avScore, $rh_scores_nwa) =
            mergeOverlappingSnws(\@a_subnetInAllLists, $rh_subnets_nwa, $rh_nodeAtts,
                                 $rh_network, $maxOvl, $useAbsVal, $useAnti);

        my($r1, $r2);   # Unused
        # Retrieving all interaction types
        ($r1, $r2, $rh_edgeAtts) = Subnetworks::readSubnets(undef, undef, undef, $f_totalEda);

        mkdir "$d_output/merged_subnets/";
        $f_nwa = "$d_output" . "merged_subnets/mergedSnw-scores-nwa.txt";
        Subnetworks::writeSubnets($rh_scores_nwa, undef, undef,
                             undef, $f_nwa, undef);
    }
    else{
        ($rh_keptSubnets_nwa, $rh_avScore) = removeOverlappingSnws(\@a_subnetInAllLists, $rh_subnets_nwa, $maxOvl);
    }
}
else{
    print "All subnetworks kept as they are\n";
    $rh_keptSubnets_nwa = $rh_subnets_nwa;
    my @a_keptSubnetworks = keys %$rh_keptSubnets_nwa;
    my @a_datasets = keys %{$$rh_keptSubnets_nwa{$a_keptSubnetworks[0]}{'Scores'}};
    $rh_avScore = getAverageScores(\@a_keptSubnetworks, $rh_keptSubnets_nwa, \@a_datasets);

    for my $subnetID (@a_keptSubnetworks) {
        if (scalar(@a_datasets) > 1) {
            $$rh_keptSubnets_nwa{$subnetID}{'Average_score'} = $$rh_avScore{$subnetID};
        }
    }
}
my @a_keptSubnetworks = keys %$rh_keptSubnets_nwa;

my %h_keptNodeAtts;
for my $subnetID (@a_keptSubnetworks) {
    my $rh_nodes = $$rh_keptSubnets_nwa{$subnetID}{'Nodes'};
    for my $geneID (keys %$rh_nodes) {
        $h_keptNodeAtts{$geneID} = $$rh_nodeAtts{$geneID};
    }
}

print "\nWriting lists/subnets-kept.txt...\n";

my $headFix = "";
if(defined $suffix){
    $headFix = "# $suffix\n";
}

my $runInfos;
$i = 0;
for my $file (sort @a_f_list) {
    $runInfos .= "# -> $file
#    p-value:
#    $a_pVal[$i]
#    Number of experiments to keep a subnet:
#    $consensus\n";

    $i++;
}

my $date = localtime(time);

open(SUBNETSKEPT, ">$d_lists" . "subnets-kept.txt")
  || die "Cannot open subnets-kept.txt:\n$!";

print SUBNETSKEPT $headFix;
print SUBNETSKEPT "# $date\n";
print SUBNETSKEPT "# Subnetworks kept after intersection from:\n";
print SUBNETSKEPT $runInfos;

for my $subnetID (@a_keptSubnetworks) {
    print SUBNETSKEPT "$subnetID\n";
}
close SUBNETSKEPT;


for my $file(@a_f_randomNwa){
    # System call of the R script which computes the p-value for each score of subnetworks kept, by dataset and distribution.
    #print "\nR system call: $d_htsnet/R/computePValues.R $f_nwa $file $d_output\n\n";
    system("$d_rscript" . "Rscript $d_htsnet/R/computePValues.R $f_nwa $file $d_output");
}
#print "\n";

my @a_pValFiles = glob("$d_lists" . "p-values_dist*");

for my $file(@a_pValFiles){
    open(FILE, $file) || die "Cannot open $file:\n$!";

    my(@a_nwaHeads, %h_indexes);
    my $isHeader = 1;
    while(<FILE>){
        next if($_ =~ m/^#/);
        $_ =~ s/[\r\n]+$//;
        my @a_row = split(/\t/, $_);

        if($isHeader){
            @a_nwaHeads = @a_row;

            for (my $i = 1; $i < scalar(@a_nwaHeads); $i++){
                if($a_nwaHeads[$i] =~ m/^p_value/){
                    push(@{$h_indexes{'p-values'}}, $i);
                    $a_nwaHeads[$i] =~ s/^p_value-//;
                }
                elsif($a_nwaHeads[$i] =~ m/^CNV_p_value/){
                    push(@{$h_indexes{'CNV_p-values'}}, $i);
                    $a_nwaHeads[$i] =~ s/^CNV_p_value-//;
                }
            }
            $isHeader = 0;
        }
        else{
            my $subnetID = $a_row[0];
            for my $key(keys %h_indexes){
                for my $i (@{$h_indexes{$key}}){
                    my $head = $a_nwaHeads[$i];
                    my $nwa  = $a_row[$i];

                    $$rh_keptSubnets_nwa{$subnetID}{$key}{$head} = $nwa;
                }
            }
        }
    }
    close FILE;
}


my @a_snwSortedByAv = sort hashValueDescendingScore (keys(%$rh_avScore));

my %h_geneOccurrence;
# Counting the occurrence of each gene in subnetworks and retrieving the best average score of subnetworks.
for my $subnetID (@a_snwSortedByAv) {
    my $rh_nodes = $$rh_keptSubnets_nwa{$subnetID}{'Nodes'};

    for my $geneID (keys %$rh_nodes) {
        $h_geneOccurrence{$geneID}++;
        $h_keptNodeAtts{$geneID}{'Gene_frequency'}++;

        if ($h_geneOccurrence{$geneID} == 1) {
            $h_keptNodeAtts{$geneID}{'Best_subnetwork_average_score'} = $$rh_avScore{$subnetID};
        }
    }
}

for my $file(@a_f_interestGenes){
    #print "Reading list of interest genes: $file...\n\n";

    open(FILE, "<$file") || die "Cannot open $file: $!";
    my %h_interestGeneIDs;
    while(<FILE>){
        next if ($_ =~ m/^#/);
        $_ =~ s/[\r\n]+$//;
        my @row = split('\t', $_);
        if(defined $row[1]){
            $h_interestGeneIDs{$row[0]} = $row[1];
        }
        else{
            $h_interestGeneIDs{$row[0]} = 'Yes';
        }
    }
    close(FILE);

    my $fileName = Utils::getName($file);
    for my $geneID (keys %h_keptNodeAtts) {
        if(defined $h_interestGeneIDs{$geneID}){
            $h_keptNodeAtts{$geneID}{'Highlights'}{$fileName} = $h_interestGeneIDs{$geneID};
        }
        else{
            $h_keptNodeAtts{$geneID}{'Highlights'}{$fileName} = undef;
        }
    }
}


Subnetworks::writeSubnets($rh_keptSubnets_nwa, \%h_keptNodeAtts, $rh_edgeAtts,
                     $d_output . "finalSubnets.nnf",
                     $d_output . "finalNwAttributes-nwa.txt",
                     $d_output . "finalNodeAttributes-noa.txt");

# Retrieve \%($geneID => \%("$GO_ID - $GO_term" => 1)).
my $rh_gene2GOs = Utils::readGene2Go($f_gene2go, $taxonID);

my ($totalGeneNb, $rh_GOccur) = Network::calcNetworkGlobalGOStatistics($rh_network, $rh_gene2GOs);

print "Looking for all GO functions of all proteins for each subnetwork and writing subnet-GOs and gene-GOs files...\n";

open(SNW_GO, ">$d_GO" . "subnet-GOs-nwa.txt")
  || die "Cannot open subnet-GOs-nwa.txt:\n$!";
print SNW_GO $headFix;
print SNW_GO "# $date\n";

for my $subnetID (@a_snwSortedByAv) {

    #print "-> $subnetID\n";

    my @a_geneIDs = keys %{$$rh_keptSubnets_nwa{$subnetID}{'Nodes'}};
    my $snwGeneNb = scalar(@a_geneIDs);

    my (%h_snwGOccur, %h_GO2Nbs);
    # Counting the occurrence of each GO and their sum in the current subnetwork.
    for my $geneID (@a_geneIDs) {
        next if (!defined $$rh_gene2GOs{$geneID});

        for my $GO (keys %{$$rh_gene2GOs{$geneID}}) {
            $h_snwGOccur{$GO}++;
        }
    }

    # Hypergeometric test for each GO.
    for my $GO(keys %h_snwGOccur) {
        my $snwGOccur      = $h_snwGOccur{$GO};
        my $totGenesWithGO = $$rh_GOccur{$GO};
        my $hypergeom      = Utils::hypergeometricTest($snwGOccur, $totGenesWithGO, $snwGeneNb, $totalGeneNb);
        my $enrichRatio    = ($snwGOccur / $snwGeneNb) / ($totGenesWithGO / $totalGeneNb);

        @{$h_GO2Nbs{$GO}} = ($snwGOccur, $totGenesWithGO, $hypergeom, $enrichRatio);
    }

    print SNW_GO "# Network\tGO_ID\tGO_term\tSubnet_GO_occurrence\tTotal_GO_occurrence\tSubnet_gene_number\tAnnotated_gene_number\tEnrichment_ratio\tHypergeometric_test\tBonf_corrected_p-value\tBH_correction\n";

    my $i = 1;                      # Rank
    my $m = scalar(keys %h_GO2Nbs); # Number of tests
    my $bhFDR = 0.05;
    my $signif = "*";
    # print "Number of tests = $m  FDR = $bhFDR\n";
    for my $GO (sort {$h_GO2Nbs{$a}[2] <=> $h_GO2Nbs{$b}[2]} keys(%h_GO2Nbs)) {
        my @a_GO   = split(/\ -\ /, $GO);
        my $GOID   = $a_GO[0];
        my $GOTerm = $a_GO[1];

        my @a_subNb_totNb_hyper_ratio = @{$h_GO2Nbs{$GO}};

        # Correction de Bonferroni
        my $correctedpval =  Utils::min(1.0, $a_subNb_totNb_hyper_ratio[2] * $m);

        # Benjamini-Hochberg multiple testing correction (FDR control)
        if($signif eq "*"){
            my $qVal = ($i/$m)*$bhFDR;
            if($a_subNb_totNb_hyper_ratio[2] <= $qVal) {
                $signif = "*";
            }
            else{
                $signif = "-";
            }
            $i++;
        }

        print SNW_GO "$subnetID\t$GOID\t$GOTerm\t$a_subNb_totNb_hyper_ratio[0]\t$a_subNb_totNb_hyper_ratio[1]\t$snwGeneNb\t$totalGeneNb\t"
                 . sprintf("%.3f", $a_subNb_totNb_hyper_ratio[3]) . "\t"
                 . sprintf("%.2e", $a_subNb_totNb_hyper_ratio[2]) . "\t"
                 . sprintf("%.2e", $correctedpval)
                 . "\t$signif\n";
    }
}
close SNW_GO;

open(GENE_GO, ">$d_GO" . "gene-GOs-noa.txt")
  || die "Cannot open gene-GOs-noa.txt:\n$!";
print GENE_GO $headFix;
print GENE_GO "# $date\n";
print GENE_GO "# Genes are ordered by descending frequency
# Gene_ID\tGO_ID\tGO_term
";

for my $geneID (sort hashValueDescendingGene (keys(%h_geneOccurrence))){
    if (defined $$rh_gene2GOs{$geneID}) {
        for my $GO (keys %{$$rh_gene2GOs{$geneID}}) {
            my @a_GO   = split(/\ -\ /, $GO);
            my $GOID   = $a_GO[0];
            my $GOTerm = $a_GO[1];
            print GENE_GO "$geneID\t$GOID\t$GOTerm\n";
        }
    }
}
close GENE_GO;

# Time measurement system - ENDING  #TMS
$date    = localtime(time);
my $time    = time;
my $period  = $time - $oldTime;

# Printing exit message
#print "\n$date: process complete\n";
#print "It took $period seconds to run\n";
#print "\n$0 done.\n\n";


## Subroutines ##

sub hashValueDescendingScore     {$$rh_avScore{$b} <=> $$rh_avScore{$a};}
sub hashValueDescendingGene      {$h_geneOccurrence{$b} <=> $h_geneOccurrence{$a};}

sub removeOverlappingSnws {
    my $ra_snwIDs = shift;
    my $rh_subnets_nwa = shift;
    my $maxOvl = shift;

    print "Removing overlapping subnetworks...\n";

    my @a_subnetIDs = keys %$rh_subnets_nwa;
    my @a_scoreKeys = keys %{$$rh_subnets_nwa{$a_subnetIDs[0]}{'Scores'}};

    my %h_rejectedSubnetworks;
    for (my $i = 0; $i < (scalar(@$ra_snwIDs) - 1); $i++) {
        my $subnetIDA = $$ra_snwIDs[$i];
        my $rh_nodesA = $$rh_subnets_nwa{$subnetIDA}{'Nodes'};

        my @a_scoreA;
        for my $scoreKey(@a_scoreKeys){
            if(defined $$rh_subnets_nwa{$subnetIDA}{'Scores'}{$scoreKey}){
                push(@a_scoreA, $$rh_subnets_nwa{$subnetIDA}{'Scores'}{$scoreKey});
            }
        }
        my ($averageScoreA, $standardDeviation) = Utils::avstd(@a_scoreA);

        for (my $j = $i + 1; $j < scalar(@$ra_snwIDs); $j++) {
            my $subnetIDB = $$ra_snwIDs[$j];
            my $rh_nodesB = $$rh_subnets_nwa{$subnetIDB}{'Nodes'};

            my $overlapA  = Subnetworks::overlap($rh_nodesB, $rh_nodesA);
            my $overlapB  = Subnetworks::overlap($rh_nodesA, $rh_nodesB);

            my @a_scoreB;
            for my $scoreKey(@a_scoreKeys){
                if(defined $$rh_subnets_nwa{$subnetIDB}{'Scores'}{$scoreKey}){
                    push(@a_scoreB, $$rh_subnets_nwa{$subnetIDB}{'Scores'}{$scoreKey});
                }
            }
            my ($averageScoreB, $standardDeviation) = Utils::avstd(@a_scoreB);

            if ($overlapA >= $maxOvl and $overlapB >= $maxOvl) {
                if ($averageScoreA >= $averageScoreB) {
                    #printf("Overlap = %.3f: $subnetIDB rejected (score = $averageScoreB)\n", $overlapB) unless(defined $h_rejectedSubnetworks{$subnetIDB});
                    $h_rejectedSubnetworks{$subnetIDB}++;
                }
                else {
                    #printf("Overlap = %.3f: $subnetIDA rejected (score = $averageScoreA)\n", $overlapA) unless(defined $h_rejectedSubnetworks{$subnetIDA});
                    $h_rejectedSubnetworks{$subnetIDA}++;
                }
            }

            elsif($overlapA >= $maxOvl and $averageScoreA <= $averageScoreB){
                #printf("$subnetIDA removed (overlap = %.3f)\n", $overlapA) unless(defined $h_rejectedSubnetworks{$subnetIDA});
                $h_rejectedSubnetworks{$subnetIDA}++;
            }

            elsif($overlapB >= $maxOvl and $averageScoreB <= $averageScoreA){
                #printf("$subnetIDB removed (overlap = %.3f)\n", $overlapB) unless(defined $h_rejectedSubnetworks{$subnetIDB});
                $h_rejectedSubnetworks{$subnetIDB}++;
            }
        }
    }

    my @a_keptSubnetworks;
    for my $subnetID (@$ra_snwIDs) {
        push(@a_keptSubnetworks, $subnetID) unless(defined $h_rejectedSubnetworks{$subnetID});
    }

    if(scalar(@a_keptSubnetworks) == 0){
        my @a_sortedRejectedSnw = sort{$h_rejectedSubnetworks{$a} <=> $h_rejectedSubnetworks{$b}} keys(%h_rejectedSubnetworks);
        my $rejMin = $h_rejectedSubnetworks{$a_sortedRejectedSnw[0]};
        warn "\nWarning: All subnetworks rejected! Subnetworks rejected $rejMin time(s) will be kept.\n";

        for my $subnetID(@a_sortedRejectedSnw){
            last if($h_rejectedSubnetworks{$subnetID} > $rejMin);
            push(@a_keptSubnetworks, $subnetID);
        }
    }

    print "Number of subnetworks kept = " . scalar(@a_keptSubnetworks) . "\n";

    my $rh_avScore = getAverageScores(\@a_keptSubnetworks, $rh_subnets_nwa, \@a_scoreKeys);

    my %h_keptSubnets_nwa;
    for my $subnetID (@a_keptSubnetworks) {
        $h_keptSubnets_nwa{$subnetID} = $$rh_subnets_nwa{$subnetID};

        if (scalar(@a_scoreKeys) > 1) {
            $h_keptSubnets_nwa{$subnetID}{'Average_score'} = $$rh_avScore{$subnetID};
        }
    }

    return \%h_keptSubnets_nwa, $rh_avScore;
}

sub getAverageScores {
    my $ra_subnetIDs = shift;
    my $rh_subnets_nwa = shift;
    my $ra_scoreKeys = shift;

    my %h_avScore;
    for my $subnetID (@$ra_subnetIDs) {
        my @a_score;
        for my $scoreKey(@$ra_scoreKeys){
            if(defined $$rh_subnets_nwa{$subnetID}{'Scores'}{$scoreKey}){
                push(@a_score, $$rh_subnets_nwa{$subnetID}{'Scores'}{$scoreKey});
            }
        }

        my ($averageScore, $standardDeviation) = Utils::avstd(@a_score);
        $h_avScore{$subnetID} = $averageScore;
    }

    return \%h_avScore;
}

sub mergeOverlappingSnws {
    my $ra_snwIDs = shift;
    my $rh_subnets_nwa = shift;
    my $rh_nodeAtts = shift;
    my $rh_network = shift;
    my $maxOvl = shift;
    my $useAbsVal = shift;
    my $useAnti = shift;

    print "Merging overlapping subnetworks...\n";

    my @a_subnetIDs = keys %$rh_subnets_nwa;
    my @a_datasets = keys %{$$rh_subnets_nwa{$a_subnetIDs[0]}{'Scores'}};
    my $rh_avScore = getAverageScores(\@a_subnetIDs, $rh_subnets_nwa, \@a_datasets);

    my(%h_snwsToMerge, %h_connectors, %h_combinedSnws, %h_alreadyMerged, %h_comb2snws, %h_scores_nwa);

    for (my $i = 0; $i < (scalar(@$ra_snwIDs) - 1); $i++) {
        my $subnetIDA = $$ra_snwIDs[$i];
        my $rh_nodesA = $$rh_subnets_nwa{$subnetIDA}{'Nodes'};

        for (my $j = $i + 1; $j < scalar(@$ra_snwIDs); $j++) {
            my $subnetIDB = $$ra_snwIDs[$j];
            my $rh_nodesB = $$rh_subnets_nwa{$subnetIDB}{'Nodes'};

            my($overlapA, $r) = overlap($rh_nodesB, $rh_nodesA);
            my($overlapB, $ra_connectors) = overlap($rh_nodesA, $rh_nodesB);

            if(($overlapA >= $maxOvl and $overlapB >= ($maxOvl * 3/4)) ||
               ($overlapB >= $maxOvl and $overlapA >= ($maxOvl * 3/4))){
                #printf("Overlaps = %.3f | %.3f: $subnetIDA and $subnetIDB merged\n", $overlapA, $overlapB);
                $h_snwsToMerge{$subnetIDA}{$subnetIDB} = 1;
                $h_snwsToMerge{$subnetIDB}{$subnetIDA} = 1;
            }

            $h_connectors{$subnetIDA}{$subnetIDB} = $ra_connectors;
            $h_connectors{$subnetIDB}{$subnetIDA} = $ra_connectors;
        }
    }

    my $index = 1;
    for my $snw(keys %h_snwsToMerge){
        next if(defined $h_alreadyMerged{$snw});

        my $combSnwID = "comb_snw-$index";
        #print "Combined subnetwork: $index\n";

        $h_combinedSnws{$combSnwID}{'Merged_snw_best_score'} = $$rh_avScore{$snw};

        mergeSubnetworks($combSnwID, \%h_combinedSnws, \%h_comb2snws, $snw, \%h_snwsToMerge,
                         $rh_subnets_nwa, $rh_avScore, \%h_alreadyMerged);

        $index++;
    }

    for my $combSnwID(keys %h_combinedSnws){
        my @a_merged = @{$h_comb2snws{$combSnwID}};
        for (my $i = 0; $i < (scalar(@a_merged) - 1); $i++) {
            my $subnetIDA = $a_merged[$i];

            for (my $j = $i + 1; $j < scalar(@a_merged); $j++) {
                my $subnetIDB = $a_merged[$j];

                for my $conn(@{$h_connectors{$subnetIDA}{$subnetIDB}}){
                    push(@{$h_combinedSnws{$combSnwID}{'Recurrent_nodes'}}, $conn)
                        unless(defined $h_combinedSnws{$combSnwID}{'Recurrent_nodes'} and
                               Utils::in_array($conn, $h_combinedSnws{$combSnwID}{'Recurrent_nodes'}));
                }
            }
        }

        my $rate = 100 * scalar(@{$h_combinedSnws{$combSnwID}{'Recurrent_nodes'}})
                   / scalar(keys %{$h_combinedSnws{$combSnwID}{'Nodes'}});
        $rate = sprintf("%.0f", $rate);
        #print "$combSnwID: $rate% of nodes present in different merged subnetworks\n";

        my %h_geneScores;
        for my $dataset(@a_datasets){
            for my $node(keys %{$h_combinedSnws{$combSnwID}{'Nodes'}}){
                my $sc;
                if(defined $useAbsVal) {
                    $sc = abs($$rh_nodeAtts{$node}{'Correlations'}{$dataset});
                }
                elsif(defined $useAnti)  {
                    $sc = -$$rh_nodeAtts{$node}{'Correlations'}{$dataset};
                }
                else{
                    $sc = $$rh_nodeAtts{$node}{'Correlations'}{$dataset};
                }
                $h_geneScores{$dataset}{$node} = $sc;
            }

            $h_combinedSnws{$combSnwID}{'Scores'}{$dataset} =
                Subnetworks::networkScore(undef, undef, undef,
                \%h_geneScores, 'averageCorrelation', $dataset);
        }
    }
#    print "\n";

    Subnetworks::includeEdges(\%h_combinedSnws, $rh_network);

    for my $combSnwID(keys %h_combinedSnws){

        $h_scores_nwa{$combSnwID}{'Scores'} = $h_combinedSnws{$combSnwID}{'Scores'};
        $h_scores_nwa{$combSnwID}{'Nodes'}  = $h_combinedSnws{$combSnwID}{'Nodes'};
        $h_scores_nwa{$combSnwID}{'Edges'}  = $h_combinedSnws{$combSnwID}{'Edges'};
    }

    for my $subnetID(@$ra_snwIDs){
        next if(defined $h_snwsToMerge{$subnetID});

        #print "Not combined subnetwork: $subnetID\n";

        $h_combinedSnws{$subnetID} = $$rh_subnets_nwa{$subnetID};
        $h_combinedSnws{$subnetID}{'Merged_snw_best_score'} = $$rh_avScore{$subnetID};
        $h_combinedSnws{$subnetID}{'Recurrent_nodes'} = undef;
        $h_combinedSnws{$subnetID}{'Merged_snw_seeds'} = undef;

        $h_scores_nwa{$subnetID}{'Scores'} = $$rh_subnets_nwa{$subnetID}{'Scores'};
        $h_scores_nwa{$subnetID}{'Nodes'}  = $$rh_subnets_nwa{$subnetID}{'Nodes'};
        $h_scores_nwa{$subnetID}{'Edges'}  = $$rh_subnets_nwa{$subnetID}{'Edges'};
    }

    my @a_combSubnetIDs = keys %h_combinedSnws;

    my $rh_newAvScore = getAverageScores(\@a_combSubnetIDs, \%h_combinedSnws, \@a_datasets);
    for my $combSnwID(@a_combSubnetIDs){
        if (scalar(@a_datasets) > 1) {
            $h_combinedSnws{$combSnwID}{'Average_score'} = $$rh_newAvScore{$combSnwID};
        }
    }

    print "Number of subnetworks after merging = " . scalar(@a_combSubnetIDs) . "\n\n";

    return \%h_combinedSnws, $rh_newAvScore, \%h_scores_nwa;
}

sub mergeSubnetworks {
    my $combSnwID = shift;
    my $rh_combinedSnws = shift;
    my $rh_comb2snws = shift;
    my $snw = shift;
    my $rh_snwsToMerge = shift;
    my $rh_subnets_nwa = shift;
    my $rh_avScore = shift;
    my $rh_alreadyMerged = shift;

    for my $node(keys %{$$rh_subnets_nwa{$snw}{'Nodes'}}){
        $$rh_combinedSnws{$combSnwID}{'Nodes'}{$node} = 1;
    }
    push(@{$$rh_combinedSnws{$combSnwID}{'Merged_snw_seeds'}}, $$rh_subnets_nwa{$snw}{'Seed'})
        unless(defined $$rh_combinedSnws{$combSnwID}{'Merged_snw_seeds'} and
               Utils::in_array($$rh_subnets_nwa{$snw}{'Seed'}, $$rh_combinedSnws{$combSnwID}{'Merged_snw_seeds'}));

    $$rh_alreadyMerged{$snw} = 1;
    push(@{$$rh_comb2snws{$combSnwID}}, $snw);

    for my $neighborSnw(keys %{$$rh_snwsToMerge{$snw}}){
        next if(defined $$rh_alreadyMerged{$neighborSnw});

        my $score = $$rh_avScore{$neighborSnw};
        $$rh_combinedSnws{$combSnwID}{'Merged_snw_best_score'} = $score
            if($score > $$rh_combinedSnws{$combSnwID}{'Merged_snw_best_score'});

        mergeSubnetworks($combSnwID, $rh_combinedSnws, $rh_comb2snws,
                         $neighborSnw, $rh_snwsToMerge, $rh_subnets_nwa,
                         $rh_avScore, $rh_alreadyMerged);
    }
}

sub overlap {
    my ($rh_subnetworkA, $rh_subnetworkB) = (shift, shift);

    my $overlap = 0;
    my @a_connectors;
    for my $node (keys %$rh_subnetworkB) {
        if (defined $$rh_subnetworkA{$node}) {
            $overlap++;
            push(@a_connectors, $node);
        }
    }
    $overlap /= scalar(keys %$rh_subnetworkB);

    return $overlap, \@a_connectors;
}
