#!/usr/bin/perl
#
# HTS-Net project
#
# Copyright Maxime Garcia, Ghislain Bidaut, Claire Rioualen, Quentin Da Costa
# (2009-2017)
#
# maxime.u.garcia@gmail.com
# ghislain.bidaut@inserm.fr
# claire.rioualen@inserm.fr
# quentin.da-costa@inserm.fr
#
# http://iti.sourceforge.net
#
# This software is a computer program whose purpose is to superimpose
# interactome over several gene expression datasets to find
# discriminative subnetworks predicting metastatic relapse in breast
# cancer.
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty and the software's author, the holder of the
# economic rights, and the successive licensors have only limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading, using, modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean that it is complicated to manipulate, and that also
# therefore means that it is reserved for developers and experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and, more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod

=head1 Name
subnetworksStatisticalValidation.pl

=head1 Synopsis

=head2 Multiple output directories
subnetworksStatisticalValidation.pl
        -od  | --outputDir <Output directory>

=head2 Output files in current directory
subnetworksStatisticalValidation.pl
        -so | --singleOutput

        -randnwaf | --randomNetworkAttributeFiles <Random attribute file 1 Random attribute file 2 Random attribute file 3>
        -realnwaf | --realNetworkAttributeFile    <Real attribute file>

=head2 Required
        -htsnet | --htsnetPath       <HTS-Net directory>

=head2 Optional
        -d   | --distributions <Distribution 1 Distribution 2 Distribution 3>
OR      -n   | --normal
OR      -g   | --gamma
OR      -gmd | --gaussianMixture
        -s   | --suffix        <Suffix string>

        -h | --help | -u | --usage

=head1 Examples
perl subnetworksStatisticalValidation.pl -htsnet ~/workspace/htsnet/ -od ~/workspace/output_data/ --distributions gamma normal gmd
perl subnetworksStatisticalValidation.pl --singleOutput -htsnet ~/workspace/local-htsnet/ --normal -randnwaf random1-subnet-scores-nwa.txt random2-subnet-scores-nwa.txt random3-subnet-scores-nwa.txt -realnwaf real-subnet-scores-nwa.txt

=head1 Description
=over

=item Calls the R scripts which create threshold files and plot histograms of real subnetwork scores;
=item Filters subnetwork files generated by detectSubnetworks.pl;
=item Counts kept subnetworks.

=back

=head1 Authors
Ghislain Bidaut     ghislain.bidaut@inserm.fr
Quentin Da Costa    quentin.da-costa@inserm.fr
Maxime Garcia       maxime.u.garcia@gmail.com
Claire Rioualen     claire.rioualen@inserm.fr

=cut


## Modules ##

use strict;
use warnings;
use Data::Dumper;

use Getopt::Long;
use File::Path;
use File::Copy;

use Subnetworks;
use Utils;


## Arguments management ##

my ($help,   $gamma,
    $normal, $gaussianMixture,
    $d_htsnet,  $d_output,
    $suffix, $outputInCurrentDir,
    $f_nwa,  $d_rscript);
my (@a_f_randomNwa, @a_distributions);

GetOptions(
    'help|usage'           => \$help,
    'gamma|g'              => \$gamma,
    'normal'               => \$normal,
    'gaussianMixture|gmd'  => \$gaussianMixture,
    'distributions|d=s{,}' => \@a_distributions,
    'htsnetPath|htsnet=s'  => \$d_htsnet,
    'rscriptDir|rsd=s'     => \$d_rscript,
    'outputDir|od=s'       => \$d_output,
    'suffix|s=s'           => \$suffix,

    'singleOutput|so'                           => \$outputInCurrentDir,
    'randomNetworkAttributeFiles|randnwaf=s{,}' => \@a_f_randomNwa,
    'realNetworkAttributeFile|realnwaf=s'       => \$f_nwa);

if(defined $d_rscript) {
    if($d_rscript !~ m/\/$/){$d_rscript .= "/";}
}
else {
    $d_rscript = "";
}

if (defined $gamma) {
    @a_distributions = ("gamma", "gamma", "gamma");
}

if (defined $gaussianMixture) {
    @a_distributions = ("gmd", "gmd", "gmd");
}

if (defined $normal || scalar(@a_distributions) != 3) {
    @a_distributions = ("normal", "normal", "normal");
}

my ($d_scores, $d_thresholds, $d_lists);

if(defined $outputInCurrentDir and
   scalar(@a_f_randomNwa) == 3 and
   defined $f_nwa and
   defined $d_htsnet){

    ($d_output, $d_scores, $d_thresholds, $d_lists) = ("", "", "", "");

    print "$0:

    HTS-Net directory                  = $d_htsnet

    Random network attribute files = @a_f_randomNwa
    Real network attribute file    = $f_nwa

    Distribution 1                 = $a_distributions[0]
    Distribution 2                 = $a_distributions[1]
    Distribution 3                 = $a_distributions[2]
\n\n";
}

elsif(defined $d_output and
      defined $d_htsnet){

    if($d_output !~ m/\/$/){$d_output .= "/";}
    $d_scores     = $d_output . "scores/";
    $d_thresholds = $d_output . "thresholds/";
    $d_lists      = $d_output . "lists/";

    # Retrieve full path of network attribute files.
    $f_nwa         = glob("$d_output" . "subnets/*nwa.txt");
    @a_f_randomNwa = glob("$d_output" . "random*/*nwa.txt");

    if (!defined $f_nwa || scalar(@a_f_randomNwa) == 0) {
        die "Network attribute files are missing";
    }

    rmtree([$d_scores,     "$d_output/pictures",
            $d_thresholds, $d_lists],
           0, 0);

    mkdir $d_scores;
    mkdir $d_thresholds;
    mkdir "$d_output/pictures";
    mkdir $d_lists;

    print "Directories scores/, thresholds/, pictures/ and lists/ created

$0:

    HTS-Net directory                  = $d_htsnet
    Output directory               = $d_output

    Random network attribute files = @a_f_randomNwa
    Real network attribute file    = $f_nwa

    Distribution 1                 = $a_distributions[0]
    Distribution 2                 = $a_distributions[1]
    Distribution 3                 = $a_distributions[2]
\n\n";
}

else{
    
    print "
NAME
subnetworksStatisticalValidation.pl

DESCRIPTION
- Calls the R scripts which create threshold files and plot histograms of real subnetwork scores;
- Filters subnetwork files generated by detectSubnetworks.pl;
- Counts kept subnetworks.

PARAMETERS

Multiple output directories
-od | --outputDir <Output directory>\n\tUse the given directory (full path) as output directory which must be the same than the one used in detectSubnetworks.pl

Output files in current directory
-so | --singleOutput
-randnwaf | --randomNetworkAttributeFiles <Random attribute file 1 Random attribute file 2 Random attribute file 3>\n\tScores of random subnetworks
-realnwaf | --realNetworkAttributeFile <Real attribute file>\n\tScores of real subnetworks

Required
-htsnet | --htsnetPath <HTS-Net directory>\n\tHTS-Net directory where pipeline/ and R/ are located

Optional
-d | --distributions <Distribution 1 Distribution 2 Distribution 3>\n\tUse the given distribution (\"normal\", \"gamma\" or \"gmd\") for fitting with scores from random mix methods 1, 2, 3 (not required if normal or gamma or gaussian mixture option used)
-n | --normal\n\tUse a normal distribution for every method (DEFAULT)
-g | --gamma\n\tUse a gamma distribution for every method
-gmd | --gaussianMixture\n\tUse a gaussian mixture distribution for every method
-s | --suffix <Suffix string>\n\tWrite the given string in created files so that you can identify from which run they come from
-rsd | --rscriptDir <Rscript directory>

-h | --help | -u | --usage\n\tDisplay this user guide

EXAMPLES
$0 --htsnetPath ~/workspace/htsnet/ -od ~/workspace/output_data/ -d gamma normal gmd
$0 --singleOutput -htsnet ~/workspace/local-htsnet/ --normal -randnwaf random1-subnet-scores-nwa.txt random2-subnet-scores-nwa.txt random3-subnet-scores-nwa.txt -realnwaf real-subnet-scores-nwa.txt

Please report problems at <ghislain.bidaut\@inserm.fr>

";
    exit(0);
}

$d_htsnet =~ s/\/+$//;


## Main program ##

my $oldTime = time;

# For each distribution, system call of the R script which computes the score threshold for each p-value and each dataset.
# Only one file of subnetwork scores for each random method is processed. Merge multiple files coming from
# parallelized multiple call of detectSubnetworks before subnetworksStatisticalValidation runing.
for my $f_randomNwa (@a_f_randomNwa){
    my @a_temp   = split(/\//, $f_randomNwa);
    my $fileName = $a_temp[-1];
    my @a_distNb = split(/-/, $fileName);
    @a_distNb    = grep(/^rand/, @a_distNb);

    if(scalar(@a_distNb) != 1){
        die "Error in input attribute file of random subnetworks $f_randomNwa";
    }
    my $distNb = substr($a_distNb[0], -1);

    # Less robust (if the name of network attribute file is changing):
    # my $fileName =~ s/-subnet-scores-nwa.txt//;
    # $distNb = substr($fileName, -1);

    #print "\nR system call: $d_htsnet/R/computeThresholds.R $f_randomNwa $a_distributions[$distNb-1] $distNb $d_output\n\n";
    system("$d_rscript"."Rscript $d_htsnet/R/computeThresholds.R $f_randomNwa $a_distributions[$distNb-1] $distNb $d_output");
}

copy($f_nwa, $d_scores . "scores-subnets-nwa.txt") || die "\nCopy failed:\n$!";

#print "\nR system call: $d_htsnet/R/plotSubnetworkScoreHist.R $f_nwa $d_output\n\n";
system("$d_rscript"."Rscript $d_htsnet/R/plotSubnetworkScoreHist.R $f_nwa $d_output");

# Retrieve network attribute hash:
# \%rh_networkAtts($subnetwork_ID => \%('Scores' => \%($datasetLabel => $subnetwork_score),
#                                       'CNV_scores' => \%($datasetLabel => $subnetwork_score))).
my($rh_networkAtts, $rh_nodeAtts,
   $rh_edgeAtts) = Subnetworks::readSubnets(undef, $f_nwa);

my @a_subnetIDs = keys %$rh_networkAtts;
my @a_scoreKeys = keys %{$$rh_networkAtts{$a_subnetIDs[0]}{'Scores'}};
my @a_cnvScoreKeys;
@a_cnvScoreKeys = keys %{$$rh_networkAtts{$a_subnetIDs[0]}{'CNV_scores'}}
    if(defined $$rh_networkAtts{$a_subnetIDs[0]}{'CNV_scores'});

my $nbOfDatasets = scalar(@a_scoreKeys) + scalar(@a_cnvScoreKeys);

#print "Number of datasets found = $nbOfDatasets\n";

# Retrieve full path of threshold files.
my @a_f_threshold = glob($d_thresholds . "thresholds*.txt");

#print "
#Writing subnets-kept files (which list subnets with a score exceeding the threshold) for each distribution and each p-value;
#Writing count-subnets-kept files (which give the number of subnetwork(s) with a score exceeding the threshold for each consensus number of dataset(s))
#...
#";

my $headFix = "";
if(defined $suffix){
    $headFix = "# $suffix\n";
}

# One threshold file by distribution.
for my $f_threshold (@a_f_threshold) {
    #print "\nCurrent threshold file:\n$f_threshold\n";
    open(THRESHOLDSFILE, "<$f_threshold") || die "Cannot open $f_threshold:\n$!";

    my (@a_row, %h_pValToTh);
    # Read the file line by line.
    while (<THRESHOLDSFILE>) {
        next if ($_ =~ m/^#/);
        $_ =~ s/[\r\n]+$//;
        @a_row = split(/\t/, $_);
        # Create \%(p-value => \%(datasetLabel => threshold)).
        $h_pValToTh{$a_row[1]}{$a_row[0]} = $a_row[2];
    }
    close THRESHOLDSFILE;

    # Retrieve the file name without path and extension.
    my $fileName = Utils::getName($f_threshold);
    $fileName =~ s/thresholds_//;
    my $date = localtime(time);

    # Print the count file of the kept subnets.
    #print "Count file = $d_thresholds" . "count-subnets-kept_$fileName.txt\n";
    open(COUNTFILE, ">$d_thresholds" . "count-subnets-kept_$fileName.txt")
      || die "Cannot open count-subnets-kept_$fileName.txt:\n$!";
    print COUNTFILE $headFix;
    print COUNTFILE "# $fileName\n";
    print COUNTFILE "# $date\n";
    print COUNTFILE "Kept_subnet_p-values\\Consensus";

    for (my $i = 1; $i <= $nbOfDatasets; $i++) {
        print COUNTFILE "\t$i";
    }
    print COUNTFILE "\n";

    open(KEPTSUBNETSFILE, ">$d_lists" . "subnets-kept_$fileName.txt")
      || die "Cannot open subnets-kept_$fileName.txt:\n$!";

    print KEPTSUBNETSFILE $headFix;
    print KEPTSUBNETSFILE "# $date
# Threshold file used:
# $f_threshold\n";

    for my $pVal (sort {$b <=> $a} keys(%h_pValToTh)) {
        # Retrieve \%(datasetLabel => threshold) for this p-value.
        my $rh_thresholds = $h_pValToTh{$pVal};
        #print "Current pValue = $pVal\n";

        my (@a_keptSnwStrings, @a_nbOfKeptSubnets);
        for my $subnetID (@a_subnetIDs) {
            my $nbOfExceededTh = 0;
            # Test if the subnetwork score is greater than or equal to the threshold for each dataset.
            for my $label (@a_scoreKeys) {
                if (defined $$rh_networkAtts{$subnetID}{'Scores'}{$label} and
                    $$rh_networkAtts{$subnetID}{'Scores'}{$label} >= $$rh_thresholds{$label}) {

                    # Add the subnetwork to the list.
                    $a_keptSnwStrings[$nbOfExceededTh] .= "$subnetID\n";
                    # Count the number of kept subnets for each possible consensus number of datasets ($nbOfExceededTh + 1).
                    $a_nbOfKeptSubnets[$nbOfExceededTh]++;
                    $nbOfExceededTh++;
                }
            }
            for my $label (@a_cnvScoreKeys) {
                if (defined $$rh_networkAtts{$subnetID}{'CNV_scores'}{$label} and
                    $$rh_networkAtts{$subnetID}{'CNV_scores'}{$label} >= $$rh_thresholds{"CNV_$label"}) {

                    # Add the subnetwork to the list.
                    $a_keptSnwStrings[$nbOfExceededTh] .= "$subnetID\n";
                    # Count the number of kept subnets for each possible consensus number of datasets ($nbOfExceededTh + 1).
                    $a_nbOfKeptSubnets[$nbOfExceededTh]++;
                    $nbOfExceededTh++;
                }
            }
        }

        print COUNTFILE "$pVal";

        my $length = scalar(@a_nbOfKeptSubnets);

        # Printing the kept subnet files for each consensus value, then the number of kept subnets by p-value and consensus value.
        for (my $i = 1; $i <= $length; $i++) {
            print KEPTSUBNETSFILE ">START|$pVal|$i\n";
            print KEPTSUBNETSFILE $a_keptSnwStrings[$i - 1];
            print COUNTFILE "\t" . $a_nbOfKeptSubnets[$i - 1];
        }

        if ($length < $nbOfDatasets) {
            foreach (($length + 1) .. $nbOfDatasets) {
                print COUNTFILE "\t0";
            }
        }
        print COUNTFILE "\n";
    }
    close COUNTFILE;
    close KEPTSUBNETSFILE;
}

# Time measurement system - ENDING  #TMS
my $date    = localtime(time);
my $time    = time;
my $period  = $time - $oldTime;

# Printing exit message
#print "\n$date: process complete\n";
#print "It took $period seconds to run\n";
#print "\n$0 done.\n\n";
